/// <reference path="mwg.d.ts" />
/// <reference path="mwg.structure.d.ts" />
/// <reference path="mwg.blas.ts" />

module org {
  export module mwg {
    export module ml {
      export interface AnomalyDetectionNode extends org.mwg.Node {
        learn(callback: org.mwg.Callback<boolean>): void;
        classify(callback: org.mwg.Callback<boolean>): void;
      }
      export abstract class BaseMLNode extends org.mwg.base.BaseNode {
        public static FROM_SEPARATOR: string = ";";
        public static FROM: string = "from";
        constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
          super(p_world, p_time, p_id, p_graph);
        }
        public static requireNotNull(obj: any, message: string): void {
          if (obj == null) {
            throw new Error(message);
          }
        }
        public illegalArgumentIfFalse(condition: boolean, errorMessage: string): void {
          if (!condition) {
            throw new Error(errorMessage);
          }
        }
        public extractFeatures(callback: org.mwg.Callback<Float64Array>): void {
          let query: string = <string>super.get(BaseMLNode.FROM);
          if (query != null) {
            let split: string[] = query.split(BaseMLNode.FROM_SEPARATOR);
            let tasks: org.mwg.task.Task[] = new Array<org.mwg.task.Task>(split.length);
            for (let i: number = 0; i < split.length; i++) {
              let t: org.mwg.task.Task = org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.travelInWorld("" + this.world()));
              t.then(org.mwg.internal.task.CoreActions.travelInTime(this.time() + ""));
              t.parse(split[i].trim(), this.graph());
              tasks[i] = t;
            }
            let result: Float64Array = new Float64Array(tasks.length);
            let waiter: org.mwg.DeferCounter = this.graph().newCounter(tasks.length);
            for (let i: number = 0; i < split.length; i++) {
              let taskIndex: number = i;
              let initial: org.mwg.task.TaskResult<any> = org.mwg.task.Tasks.emptyResult();
              initial.add(this);
              tasks[i].executeWith(this.graph(), initial, (currentResult : org.mwg.task.TaskResult<any>) => {
{
                  if (currentResult == null) {
                    result[taskIndex] = org.mwg.Constants.NULL_LONG;
                  } else {
                    result[taskIndex] = parseFloat(currentResult.get(0).toString());
                    currentResult.free();
                  }
                  waiter.count();
                }              });
            }
            waiter.then(() => {
{
                callback(result);
              }            });
          } else {
            callback(null);
          }
        }
      }
      export interface ClassificationNode extends org.mwg.Node {
        learn(expectedClass: number, callback: org.mwg.Callback<boolean>): void;
        classify(callback: org.mwg.Callback<number>): void;
      }
      export interface ClusteringNode extends org.mwg.Node {
        learn(callback: org.mwg.Callback<boolean>): void;
        inferCluster(callback: org.mwg.Callback<number>): void;
      }
      export class MLPlugin extends org.mwg.structure.StructurePlugin {
        public start(graph: org.mwg.Graph): void {
          super.start(graph);
          graph.actionRegistry().declaration(org.mwg.ml.algorithm.regression.actions.ReadContinuous.NAME).setParams(org.mwg.Type.STRING).setFactory((params : any[]) => {
{
              return new org.mwg.ml.algorithm.regression.actions.ReadContinuous(<string>params[0]);
            }          });
          graph.actionRegistry().declaration(org.mwg.ml.algorithm.regression.actions.SetContinuous.NAME).setParams(org.mwg.Type.STRING, org.mwg.Type.STRING).setFactory((params : any[]) => {
{
              return new org.mwg.ml.algorithm.regression.actions.SetContinuous(<string>params[0], <string>params[1]);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.regression.PolynomialNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.regression.PolynomialNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.neuralnet.FlatNeuralNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.neuralnet.FlatNeuralNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.profiling.GaussianNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.profiling.GaussianNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.profiling.GaussianSlotNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.profiling.GaussianSlotNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.profiling.GaussianMixtureNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.profiling.GaussianMixtureNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.regression.LiveLinearRegressionNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.regression.LiveLinearRegressionNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.algorithm.anomalydetector.InterquartileRangeOutlierDetectorNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.algorithm.anomalydetector.InterquartileRangeOutlierDetectorNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.neuralnet.NeuralNode.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.neuralnet.NeuralNode(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.ml.neuralnet.NeuralNodeEmpty.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.ml.neuralnet.NeuralNodeEmpty(world, time, id, graph);
            }          });
        }
      }
      export interface ProfilingNode extends org.mwg.Node {
        learn(callback: org.mwg.Callback<boolean>): void;
        learnWith(values: Float64Array): void;
        predict(callback: org.mwg.Callback<Float64Array>): void;
      }
      export interface ProgressReporter {
        updateProgress(value: number): void;
        isCancelled(): boolean;
        updateInformation(info: string): void;
      }
      export interface RegressionNode extends org.mwg.Node {
        learn(output: number, callback: org.mwg.Callback<boolean>): void;
        extrapolate(callback: org.mwg.Callback<number>): void;
      }
      export module actions {
        export class ActionTraverseOrKeep implements org.mwg.task.Action {
          public static NAME: string = "traverseOrKeep";
          private _name: string;
          constructor(p_name: string) {
            this._name = p_name;
          }
          public eval(ctx: org.mwg.task.TaskContext): void {
            let flatName: string = ctx.template(this._name);
            let previousResult: org.mwg.task.TaskResult<any> = ctx.result();
            if (previousResult != null) {
              let finalResult: org.mwg.task.TaskResult<any> = ctx.newResult();
              let previousSize: number = previousResult.size();
              let defer: org.mwg.DeferCounter = ctx.graph().newCounter(previousSize);
              for (let i: number = 0; i < previousSize; i++) {
                let loop: any = previousResult.get(i);
                if (loop instanceof org.mwg.base.BaseNode) {
                  let casted: org.mwg.Node = <org.mwg.Node>loop;
                  if (casted.type(flatName) == org.mwg.Type.RELATION) {
                    casted.relation(flatName, (result : org.mwg.Node[]) => {
{
                        if (result != null) {
                          for (let j: number = 0; j < result.length; j++) {
                            finalResult.add(result[j]);
                          }
                        }
                        defer.count();
                      }                    });
                  } else {
                    finalResult.add(casted.graph().cloneNode(casted));
                    defer.count();
                  }
                } else {
                  finalResult.add(loop);
                  defer.count();
                }
              }
              defer.then(() => {
{
                  ctx.continueWith(finalResult);
                }              });
            } else {
              ctx.continueTask();
            }
          }
          public serialize(builder: java.lang.StringBuilder): void {
            builder.append(ActionTraverseOrKeep.NAME);
            builder.append(org.mwg.Constants.TASK_PARAM_OPEN);
            org.mwg.internal.task.TaskHelper.serializeString(this._name, builder, true);
            builder.append(org.mwg.Constants.TASK_PARAM_CLOSE);
          }
          public toString(): string {
            let res: java.lang.StringBuilder = new java.lang.StringBuilder();
            this.serialize(res);
            return res.toString();
          }
        }
      }
      export module algorithm {
        export module anomalydetector {
          export class InterquartileRangeOutlierDetectorNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.AnomalyDetectionNode {
            public static MAX_BUFFER_LIMIT: number = 100000;
            public static NAME: string = "InterquartileRangeAnomalyDetection";
            private static INTERNAL_VALUE_BUFFER_KEY: string = "_valueBuffer";
            public static BUFFER_SIZE_KEY: string = "BufferSize";
            public static BUFFER_SIZE_DEF: number = 50;
            public static INPUT_DIM_KEY: string = "InputDimensions";
            public static INPUT_DIM_DEF: number = -1;
            public static UPPER_PERCENTILE: number = 0.75;
            public static LOWER_PERCENTILE: number = 0.25;
            public static RANGE_COEF: number = 1.5;
            public static UPPER_BOUND_KEY_PREFIX: string = "UpperBoundDimension";
            public static LOWER_BOUND_KEY_PREFIX: string = "LowerBoundDimension";
            private static enforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asIntWithin(InterquartileRangeOutlierDetectorNode.BUFFER_SIZE_KEY, 1, InterquartileRangeOutlierDetectorNode.MAX_BUFFER_LIMIT);
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public addValue(value: Float64Array): boolean {
              this.illegalArgumentIfFalse(value != null, "Value must be not null");
              let state: org.mwg.plugin.NodeState = this.unphasedState();
              let dimensions: number = state.getFromKeyWithDefault(InterquartileRangeOutlierDetectorNode.INPUT_DIM_KEY, InterquartileRangeOutlierDetectorNode.INPUT_DIM_DEF);
              if (dimensions < 0) {
                dimensions = value.length;
                state.setFromKey(InterquartileRangeOutlierDetectorNode.INPUT_DIM_KEY, org.mwg.Type.INT, value.length);
              }
              let buffer: Float64Array = state.getFromKeyWithDefault(InterquartileRangeOutlierDetectorNode.INTERNAL_VALUE_BUFFER_KEY, new Float64Array(0));
              let bufferLength: number = buffer.length / dimensions;
              let maxBufferLength: number = state.getFromKeyWithDefault(InterquartileRangeOutlierDetectorNode.BUFFER_SIZE_KEY, InterquartileRangeOutlierDetectorNode.BUFFER_SIZE_DEF);
              let numValuesToRemoveFromBeginning: number = Math.max(0, bufferLength + 1 - maxBufferLength);
              let newBufferLength: number = bufferLength + 1 - numValuesToRemoveFromBeginning;
              let newBuffer: Float64Array = new Float64Array(newBufferLength * dimensions);
              java.lang.System.arraycopy(buffer, numValuesToRemoveFromBeginning * dimensions, newBuffer, 0, newBuffer.length - dimensions);
              java.lang.System.arraycopy(value, 0, newBuffer, newBuffer.length - dimensions, dimensions);
              let upperBounds: Float64Array = new Float64Array(dimensions);
              let lowerBounds: Float64Array = new Float64Array(dimensions);
              for (let i: number = 0; i < dimensions; i++) {
                let column: Float64Array = new Float64Array(newBufferLength);
                let index: number = i;
                for (let j: number = 0; j < newBufferLength; j++) {
                  column[j] = newBuffer[index];
                  index += dimensions;
                }
                column.sort();
                let upperPercentile: number = column[Math.min(<number>Math.round(newBufferLength * InterquartileRangeOutlierDetectorNode.UPPER_PERCENTILE), newBufferLength - 1)];
                let lowerPercentile: number = column[Math.max(<number>Math.round(newBufferLength * InterquartileRangeOutlierDetectorNode.LOWER_PERCENTILE), 0)];
                let interquartileRange: number = upperPercentile - lowerPercentile;
                upperBounds[i] = upperPercentile + InterquartileRangeOutlierDetectorNode.RANGE_COEF * interquartileRange;
                lowerBounds[i] = lowerPercentile - InterquartileRangeOutlierDetectorNode.RANGE_COEF * interquartileRange;
                state.setFromKey(InterquartileRangeOutlierDetectorNode.UPPER_BOUND_KEY_PREFIX + i, org.mwg.Type.DOUBLE, upperBounds[i]);
                state.setFromKey(InterquartileRangeOutlierDetectorNode.LOWER_BOUND_KEY_PREFIX + i, org.mwg.Type.DOUBLE, lowerBounds[i]);
              }
              state.setFromKey(InterquartileRangeOutlierDetectorNode.INTERNAL_VALUE_BUFFER_KEY, org.mwg.Type.DOUBLE_ARRAY, newBuffer);
              return this.checkValue(value, upperBounds, lowerBounds);
            }
            public checkValue(value: Float64Array, lowerBounds: Float64Array, upperBounds: Float64Array): boolean {
              for (let i: number = 0; i < value.length; i++) {
                if (!((value[i] <= upperBounds[i]) && (value[i] >= lowerBounds[i]))) {
                  return true;
                }
              }
              return false;
            }
            public learn(callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((result : Float64Array) => {
{
                  let outcome: boolean = this.addValue(result);
                  callback(outcome);
                }              });
            }
            public classify(callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((result : Float64Array) => {
{
                  let lowerBounds: Float64Array = new Float64Array(result.length);
                  let upperBounds: Float64Array = new Float64Array(result.length);
                  let state: org.mwg.plugin.NodeState = this.unphasedState();
                  for (let i: number = 0; i < result.length; i++) {
                    lowerBounds[i] = <number>state.getFromKey(org.mwg.ml.algorithm.anomalydetector.InterquartileRangeOutlierDetectorNode.LOWER_BOUND_KEY_PREFIX + i);
                    upperBounds[i] = <number>state.getFromKey(org.mwg.ml.algorithm.anomalydetector.InterquartileRangeOutlierDetectorNode.UPPER_BOUND_KEY_PREFIX + i);
                  }
                  let isAnomaly: boolean = this.checkValue(result, lowerBounds, upperBounds);
                  callback(isAnomaly);
                }              });
            }
            public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
              if (InterquartileRangeOutlierDetectorNode.INTERNAL_VALUE_BUFFER_KEY === propertyName || InterquartileRangeOutlierDetectorNode.INPUT_DIM_KEY === propertyName) {
              } else {
                InterquartileRangeOutlierDetectorNode.enforcer.check(propertyName, propertyType, propertyValue);
                super.set(propertyName, propertyType, propertyValue);
              }
              return this;
            }
          }
        }
        export module preprocessing {
          export class PCA {
            private _data: org.mwg.struct.DMatrix;
            private _olddim: number;
            private _min: Float64Array;
            private _max: Float64Array;
            private _sigma: Float64Array;
            private _avg: Float64Array;
            private _information: Float64Array;
            private _processType: number;
            private _percentToRetain: number;
            private _bestDim: number;
            public _svdDecompose: org.mwg.ml.common.matrix.SVDDecompose;
            public static EPS: number = 1e-30;
            private static _matrixV: org.mwg.struct.DMatrix;
            public static NOPROCESS: number = 0;
            public static CENTER_ON_AVG: number = 1;
            public static NORMALIZE: number = 2;
            public get_bestDim(): number {
              return this._bestDim;
            }
            public get_data(): org.mwg.struct.DMatrix {
              return this._data;
            }
            private static clone(data: Float64Array): Float64Array {
              let c: Float64Array = new Float64Array(data.length);
              java.lang.System.arraycopy(data, 0, c, 0, data.length);
              return c;
            }
            public getDimensionInfo(): Float64Array {
              return org.mwg.ml.algorithm.preprocessing.PCA.clone(this._information);
            }
            public get_min(): Float64Array {
              return org.mwg.ml.algorithm.preprocessing.PCA.clone(this._min);
            }
            public get_max(): Float64Array {
              return org.mwg.ml.algorithm.preprocessing.PCA.clone(this._max);
            }
            public get_avg(): Float64Array {
              return org.mwg.ml.algorithm.preprocessing.PCA.clone(this._avg);
            }
            public get_sigma(): Float64Array {
              return org.mwg.ml.algorithm.preprocessing.PCA.clone(this._sigma);
            }
            public normalizeData(data: org.mwg.struct.DMatrix): void {
              let d: number = 1;
              for (let j: number = 0; j < data.columns(); j++) {
                if (this._sigma[j] < PCA.EPS) {
                  for (let i: number = 0; i < data.rows(); i++) {
                    data.set(i, j, 0);
                  }
                } else {
                  d = 1 / this._sigma[j];
                  for (let i: number = 0; i < data.rows(); i++) {
                    data.set(i, j, (data.get(i, j) - this._avg[j]) * d);
                  }
                }
              }
            }
            public convertVector(data: Float64Array): Float64Array {
              let v: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(org.mwg.ml.algorithm.preprocessing.PCA.clone(data), 1, data.length);
              if (this._processType == PCA.NORMALIZE) {
                this.normalizeData(v);
              }
              let res: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiply(v, PCA._matrixV);
              let result: Float64Array = new Float64Array(res.columns());
              for (let i: number = 0; i < res.columns(); i++) {
                result[i] = res.get(0, i);
              }
              return result;
            }
            public convertSpace(initial: org.mwg.ml.common.matrix.VolatileDMatrix): org.mwg.struct.DMatrix {
              if (this._processType == PCA.NORMALIZE) {
                this.normalizeData(initial);
              }
              return org.mwg.ml.common.matrix.MatrixOps.multiply(initial, PCA._matrixV);
            }
            public setDimension(dim: number): void {
              PCA._matrixV = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this._olddim, dim);
              let tempV: org.mwg.struct.DMatrix = this._svdDecompose.getVt();
              for (let i: number = 0; i < this._olddim; i++) {
                for (let j: number = 0; j < dim; j++) {
                  PCA._matrixV.set(i, j, tempV.get(j, i));
                }
              }
            }
            public inverseNormalizeData(data: org.mwg.struct.DMatrix): void {
              for (let j: number = 0; j < data.columns(); j++) {
                if ((this._sigma[j]) < PCA.EPS) {
                  for (let i: number = 0; i < data.rows(); i++) {
                    data.set(i, j, this._avg[j]);
                  }
                } else {
                  for (let i: number = 0; i < data.rows(); i++) {
                    data.set(i, j, data.get(i, j) * this._sigma[j] + this._avg[j]);
                  }
                }
              }
            }
            public inverseConvertVector(data: Float64Array): Float64Array {
              let v: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(org.mwg.ml.algorithm.preprocessing.PCA.clone(data), 1, data.length);
              let res: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiplyTranspose(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, v, org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, PCA._matrixV);
              if (this._processType == PCA.NORMALIZE) {
                this.inverseNormalizeData(res);
              }
              let result: Float64Array = new Float64Array(res.columns());
              for (let i: number = 0; i < res.columns(); i++) {
                result[i] = res.get(0, i);
              }
              return result;
            }
            public inverseConvertSpace(initial: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              let res: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiplyTranspose(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, initial, org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, PCA._matrixV);
              if (this._processType == PCA.NORMALIZE) {
                this.inverseNormalizeData(res);
              }
              return res;
            }
            public getTransformationVector(): org.mwg.struct.DMatrix {
              return PCA._matrixV;
            }
            private calculateMinMaxAvg(): void {
              this._min = new Float64Array(this._data.columns());
              this._max = new Float64Array(this._data.columns());
              this._avg = new Float64Array(this._data.columns());
              this._sigma = new Float64Array(this._data.columns());
              this._olddim = this._data.columns();
              for (let j: number = 0; j < this._data.columns(); j++) {
                this._min[j] = this._data.get(0, j);
                this._max[j] = this._min[j];
                this._avg[j] = this._min[j];
                this._sigma[j] = this._min[j] * this._min[j];
              }
              let d: number;
              for (let i: number = 1; i < this._data.rows(); i++) {
                for (let j: number = 0; j < this._data.columns(); j++) {
                  d = this._data.get(i, j);
                  this._avg[j] += d;
                  this._sigma[j] += d * d;
                  if (d < this._min[j]) {
                    this._min[j] = d;
                  } else if (d > this._max[j]) {
                    this._max[j] = d;
                  }

                }
              }
              for (let j: number = 0; j < this._data.columns(); j++) {
                this._avg[j] = this._avg[j] / this._data.rows();
                this._sigma[j] = Math.sqrt(this._sigma[j]);
              }
            }
            private static shiftColumn(data: org.mwg.struct.DMatrix, shift: Float64Array, workInPlace: boolean): org.mwg.struct.DMatrix {
              let temp: org.mwg.struct.DMatrix = data;
              if (!workInPlace) {
                temp = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(data);
              }
              for (let i: number = 0; i < temp.rows(); i++) {
                for (let j: number = 0; j < temp.columns(); j++) {
                  temp.set(i, j, temp.get(i, j) - shift[j]);
                }
              }
              return temp;
            }
            private static inverseShift(data: org.mwg.struct.DMatrix, shift: Float64Array, workInPlace: boolean): org.mwg.struct.DMatrix {
              let temp: org.mwg.struct.DMatrix = data;
              if (!workInPlace) {
                temp = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(data);
              }
              for (let i: number = 0; i < temp.rows(); i++) {
                for (let j: number = 0; j < temp.columns(); j++) {
                  temp.set(i, j, temp.get(i, j) + shift[j]);
                }
              }
              return temp;
            }
            public retainDynamic(svector: Float64Array): number {
              let d: number = 0;
              for (let i: number = 0; i < svector.length; i++) {
                d += svector[i] * svector[i];
              }
              let integrator: number = 0;
              let previoust: number = 1;
              let t: number = 1;
              let p: number;
              let tag: boolean = true;
              integrator = svector[0] * svector[0];
              let xi: number = 0;
              this._information = new Float64Array(svector.length + 1);
              for (let i: number = 1; i < svector.length; i++) {
                this._information[i] = integrator * 100 / d;
                previoust = t;
                t = svector[i] * svector[i] / (svector[i - 1] * svector[i - 1]);
                console.log(i + " , " + svector[i] + " , " + t / previoust + " , " + this._information[i] + "%");
                p = integrator * 100 / d;
                if (t / previoust < 0.85 && xi == 0 && i != 1 && p >= 85 && tag) {
                  tag = false;
                  this._percentToRetain = p;
                  xi = i;
                }
                integrator += svector[i] * svector[i];
              }
              this._information[svector.length] = 100;
              if (xi == 0) {
                this._percentToRetain = integrator * 100 / d;
                xi = svector.length;
              }
              console.log(svector.length + " , " + svector[svector.length - 1] + " , " + t / previoust + " , " + integrator * 100 / d + "%");
              console.log("");
              this._bestDim = xi;
              return xi;
            }
            public static retain(svector: Float64Array, percent: number): number {
              let d: number = 0;
              for (let i: number = 0; i < svector.length; i++) {
                d += svector[i] * svector[i];
              }
              d = d * percent;
              let t: number = 0;
              for (let i: number = 0; i < svector.length; i++) {
                t += svector[i] * svector[i];
                if (t > d) {
                  return i + 1;
                }
              }
              return svector.length;
            }
            constructor(data: org.mwg.struct.DMatrix, processType: number) {
              this._data = data;
              this._processType = processType;
              this.calculateMinMaxAvg();
              if (processType == PCA.CENTER_ON_AVG) {
                org.mwg.ml.algorithm.preprocessing.PCA.shiftColumn(this._data, this._avg, true);
              } else if (processType == PCA.NORMALIZE) {
                this.normalizeData(this._data);
              }

              this._svdDecompose = org.mwg.ml.common.matrix.MatrixOps.defaultEngine().decomposeSVD(this._data, false);
              let singularValues: Float64Array = this._svdDecompose.getS();
              console.log("Singular values");
              for (let i: number = 0; i < singularValues.length; i++) {
                console.log(singularValues[i]);
              }
              console.log("");
              console.log("Need to retain: " + this.retainDynamic(singularValues) + " / " + data.columns() + " dimensions");
              console.log("Energy retained: " + this._percentToRetain + " %");
            }
          }
        }
        export module profiling {
          export class GaussianMixtureNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.ProfilingNode {
            public static NAME: string = "GaussianMixtureNode";
            public static MIN: string = "min";
            public static MAX: string = "max";
            public static AVG: string = "avg";
            public static COV: string = "cov";
            public static LEVEL: string = "_level";
            public static LEVEL_DEF: number = 0;
            public static WIDTH: string = "_width";
            public static WIDTH_DEF: number = 10;
            public static COMPRESSION_FACTOR: string = "_compression";
            public static COMPRESSION_FACTOR_DEF: number = 2;
            public static COMPRESSION_ITER: string = "_compressioniter";
            public static COMPRESSION_ITER_DEF: number = 10;
            public static THRESHOLD: string = "_threshold";
            public static THRESHOLD_DEF: number = 3;
            public static RESOLUTION: string = "_resolution";
            public static INTERNAL_SUBGAUSSIAN: string = "_subGaussian";
            private static INTERNAL_SUM_KEY: string = "_sum";
            private static INTERNAL_SUMSQUARE_KEY: string = "_sumSquare";
            private static INTERNAL_TOTAL_KEY: string = "_total";
            private static INTERNAL_MIN_KEY: string = "_min";
            private static INTERNAL_MAX_KEY: string = "_max";
            private static enforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asIntWithin(GaussianMixtureNode.LEVEL, 0, 1000).asIntWithin(GaussianMixtureNode.WIDTH, 1, 1000).asPositiveDouble(GaussianMixtureNode.COMPRESSION_FACTOR).asPositiveDouble(GaussianMixtureNode.THRESHOLD).asDoubleArray(GaussianMixtureNode.RESOLUTION);
            private static traverseTask: org.mwg.task.Task = GaussianMixtureNode.getTraverseTask();
            private static selectTraverseTask: org.mwg.task.Task = GaussianMixtureNode.getSelectTraverseTask();
            private static deepTraverseTask: org.mwg.task.Task = GaussianMixtureNode.getDeepTraverseTask();
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
              GaussianMixtureNode.enforcer.check(propertyName, propertyType, propertyValue);
              return super.set(propertyName, propertyType, propertyValue);
            }
            public type(attributeName: string): number {
              if (attributeName === GaussianMixtureNode.AVG) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianMixtureNode.MIN) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianMixtureNode.MAX) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianMixtureNode.COV) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianMixtureNode.RESOLUTION) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else {
                return super.type(attributeName);
              }




            }
            public get(attributeName: string): any {
              if (attributeName === GaussianMixtureNode.AVG) {
                return this.getAvg();
              } else if (attributeName === GaussianMixtureNode.MIN) {
                return this.getMin();
              } else if (attributeName === GaussianMixtureNode.MAX) {
                return this.getMax();
              } else if (attributeName === GaussianMixtureNode.MAX) {
                return this.getMax();
              } else if (attributeName === GaussianMixtureNode.COV) {
                let resolved: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
                let initialResolution: Float64Array = <Float64Array>resolved.getFromKey(GaussianMixtureNode.RESOLUTION);
                let nbfeature: number = this.getNumberOfFeatures();
                if (initialResolution == null) {
                  initialResolution = new Float64Array(nbfeature);
                  for (let i: number = 0; i < nbfeature; i++) {
                    initialResolution[i] = 1;
                  }
                }
                return this.getCovariance(this.getAvg(), initialResolution);
              } else {
                return super.get(attributeName);
              }




            }
            public learn(callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((values : Float64Array) => {
{
                  this.learnWith(values);
                }              });
            }
            private static getTraverseTask(): org.mwg.task.Task {
              let traverseTask: org.mwg.task.Task = org.mwg.task.Tasks.newTask();
              traverseTask.then(org.mwg.internal.task.CoreActions.defineAsGlobalVar("parent")).then(org.mwg.internal.task.CoreActions.traverse(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN)).thenDo((ctx : org.mwg.task.TaskContext) => {
{
                  let width: number = <number>ctx.variable("width").get(0);
                  let compressionFactor: number = <number>ctx.variable("compressionFactor").get(0);
                  let compressionIter: number = <number>ctx.variable("compressionIter").get(0);
                  let resolution: Float64Array = <Float64Array>ctx.variable("resolution").get(0);
                  let threshold: number = <number>ctx.variable("threshold").get(0);
                  let values: Float64Array = <Float64Array>ctx.variable("values").get(0);
                  let root: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = <org.mwg.ml.algorithm.profiling.GaussianMixtureNode>ctx.variable("root").get(0);
                  let result: org.mwg.task.TaskResult<org.mwg.Node> = ctx.resultAsNodes();
                  let parent: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = <org.mwg.ml.algorithm.profiling.GaussianMixtureNode>ctx.variable("parent").get(0);
                  let resultChild: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = root.filter(result, values, resolution, threshold, parent.getLevel() - 1.0);
                  if (resultChild != null) {
                    parent.internallearn(values, width, compressionFactor, compressionIter, resolution, threshold, false);
                    ctx.continueWith(ctx.wrapClone(resultChild));
                  } else {
                    parent.internallearn(values, width, compressionFactor, compressionIter, resolution, threshold, true);
                    ctx.continueWith(null);
                  }
                }              }).ifThen((ctx : org.mwg.task.TaskContext) => {
{
                  return (ctx.result() != null);
                }              }, traverseTask);
              return traverseTask;
            }
            public learnWith(values: Float64Array): void {
              let resolved: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
              let width: number = resolved.getFromKeyWithDefault(GaussianMixtureNode.WIDTH, GaussianMixtureNode.WIDTH_DEF);
              let compressionFactor: number = resolved.getFromKeyWithDefault(GaussianMixtureNode.COMPRESSION_FACTOR, GaussianMixtureNode.COMPRESSION_FACTOR_DEF);
              let compressionIter: number = resolved.getFromKeyWithDefault(GaussianMixtureNode.COMPRESSION_ITER, GaussianMixtureNode.COMPRESSION_ITER_DEF);
              let resolution: Float64Array = <Float64Array>resolved.getFromKey(GaussianMixtureNode.RESOLUTION);
              if (resolution == null) {
                resolution = new Float64Array(values.length);
                for (let i: number = 0; i < values.length; i++) {
                  resolution[i] = 1;
                }
              }
              let threshold: number = resolved.getFromKeyWithDefault(GaussianMixtureNode.THRESHOLD, GaussianMixtureNode.THRESHOLD_DEF);
              let context: org.mwg.task.TaskContext = GaussianMixtureNode.traverseTask.prepare(this.graph(), this, (result : org.mwg.task.TaskResult<any>) => {
{
                  if (result != null) {
                    result.free();
                  }
                }              });
              let resolutionTR: org.mwg.task.TaskResult<any> = context.newResult().add(resolution);
              let valuesTR: org.mwg.task.TaskResult<any> = context.newResult().add(values);
              context.setGlobalVariable("width", width);
              context.setGlobalVariable("compressionFactor", compressionFactor);
              context.setGlobalVariable("compressionIter", compressionIter);
              context.setGlobalVariable("resolution", resolutionTR);
              context.setGlobalVariable("threshold", threshold);
              context.setGlobalVariable("values", valuesTR);
              context.setGlobalVariable("root", this);
              GaussianMixtureNode.traverseTask.executeUsing(context);
            }
            private checkInside(min: Float64Array, max: Float64Array, resolution: Float64Array, threshold: number, level: number): boolean {
              let threshold2: number = threshold + level * 0.707;
              let avg: Float64Array = this.getAvg();
              let result: boolean = true;
              let cov: Float64Array = this.getCovarianceArray(avg, resolution);
              for (let i: number = 0; i < min.length; i++) {
                cov[i] = Math.sqrt(cov[i]);
                if (((avg[i] + cov[i]) < (min[i] - threshold2 * resolution[i])) || ((avg[i] - cov[i]) > (max[i] + threshold2 * resolution[i]))) {
                  result = false;
break;
                }
              }
              return result;
            }
            private filter(result: org.mwg.task.TaskResult<org.mwg.Node>, features: Float64Array, resolution: Float64Array, threshold: number, level: number): org.mwg.ml.algorithm.profiling.GaussianMixtureNode {
              if (result == null || result.size() == 0) {
                return null;
              }
              let distances: Float64Array = new Float64Array(result.size());
              let min: number = java.lang.Double.MAX_VALUE;
              let index: number = 0;
              for (let i: number = 0; i < result.size(); i++) {
                let temp: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = (<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>result.get(i));
                let avg: Float64Array = temp.getAvg();
                distances[i] = org.mwg.ml.algorithm.profiling.GaussianMixtureNode.distance(features, avg, temp.getCovarianceArray(avg, resolution));
                if (distances[i] < min) {
                  min = distances[i];
                  index = i;
                }
              }
              if (min < threshold + level * 0.707) {
                return (<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>result.get(index));
              } else {
                return null;
              }
            }
            public predict(callback: org.mwg.Callback<Float64Array>): void {}
            public getLevel(): number {
              return this._resolver.resolveState(this).getFromKeyWithDefault(GaussianMixtureNode.LEVEL, GaussianMixtureNode.LEVEL_DEF);
            }
            public getWidth(): number {
              return this._resolver.resolveState(this).getFromKeyWithDefault(GaussianMixtureNode.WIDTH, GaussianMixtureNode.WIDTH_DEF);
            }
            public getCompressionFactor(): number {
              return this._resolver.resolveState(this).getFromKeyWithDefault(GaussianMixtureNode.COMPRESSION_FACTOR, GaussianMixtureNode.COMPRESSION_FACTOR_DEF);
            }
            public getCompressionIter(): number {
              return this._resolver.resolveState(this).getFromKeyWithDefault(GaussianMixtureNode.COMPRESSION_ITER, GaussianMixtureNode.COMPRESSION_ITER_DEF);
            }
            private createLevel(values: Float64Array, level: number, width: number, compressionFactor: number, compressionIter: number, resolution: Float64Array, threshold: number): org.mwg.ml.algorithm.profiling.GaussianMixtureNode {
              let g: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = <org.mwg.ml.algorithm.profiling.GaussianMixtureNode>this.graph().newTypedNode(this.world(), this.time(), GaussianMixtureNode.NAME);
              g.set(GaussianMixtureNode.LEVEL, org.mwg.Type.INT, level);
              g.internallearn(values, width, compressionFactor, compressionIter, resolution, threshold, false);
              super.addToRelation(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN, g);
              return g;
            }
            private checkAndCompress(width: number, compressionFactor: number, compressionIter: number, resolution: Float64Array, threshold: number): void {
              let selfPointer: org.mwg.Node = this;
              let subgaussians: org.mwg.struct.Relation = <org.mwg.struct.Relation>super.get(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN);
              if (subgaussians != null && subgaussians.size() != 0 && subgaussians.size() >= compressionFactor * width) {
                super.relation(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN, (result : org.mwg.Node[]) => {
{
                    let subgauss: org.mwg.ml.algorithm.profiling.GaussianMixtureNode[] = new Array<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>(result.length);
                    let data: Array<Float64Array> = new Array<Float64Array>(result.length);
                    for (let i: number = 0; i < result.length; i++) {
                      subgauss[i] = <org.mwg.ml.algorithm.profiling.GaussianMixtureNode>result[i];
                      data[i] = subgauss[i].getAvg();
                    }
                    let clusteringEngine: org.mwg.ml.algorithm.profiling.KMeans = new org.mwg.ml.algorithm.profiling.KMeans();
                    let clusters: Array<Int32Array> = clusteringEngine.getClusterIds(data, width, compressionIter, resolution);
                    let mainClusters: org.mwg.ml.algorithm.profiling.GaussianMixtureNode[] = new Array<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>(width);
                    for (let i: number = 0; i < width; i++) {
                      if (clusters[i] != null && clusters[i].length > 0) {
                        let max: number = 0;
                        let maxpos: number = 0;
                        for (let j: number = 0; j < clusters[i].length; j++) {
                          let x: number = subgauss[clusters[i][j]].getTotal();
                          if (x > max) {
                            max = x;
                            maxpos = clusters[i][j];
                          }
                        }
                        mainClusters[i] = subgauss[maxpos];
                      }
                    }
                    for (let i: number = 0; i < width; i++) {
                      if (clusters[i].length > 1 && mainClusters[i].getTotal() == 1 && mainClusters[i].getLevel() > 0) {
                        mainClusters[i].createLevel(mainClusters[i].getAvg(), mainClusters[i].getLevel() - 1, width, compressionFactor, compressionIter, resolution, threshold).free();
                      }
                      if (clusters[i] != null && clusters[i].length > 0) {
                        for (let j: number = 0; j < clusters[i].length; j++) {
                          let g: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = subgauss[clusters[i][j]];
                          if (g != mainClusters[i]) {
                            mainClusters[i].move(g);
                            selfPointer.removeFromRelation(org.mwg.ml.algorithm.profiling.GaussianMixtureNode.INTERNAL_SUBGAUSSIAN, g);
                            g.free();
                          }
                        }
                        mainClusters[i].checkAndCompress(width, compressionFactor, compressionIter, resolution, threshold);
                      }
                    }
                    for (let i: number = 0; i < result.length; i++) {
                      result[i].free();
                    }
                  }                });
              }
            }
            private move(subgaus: org.mwg.ml.algorithm.profiling.GaussianMixtureNode): void {
              let total: number = this.getTotal();
              let level: number = this.getLevel();
              let sum: Float64Array = this.getSum();
              let min: Float64Array = this.getMin();
              let max: Float64Array = this.getMax();
              let sumsquares: Float64Array = this.getSumSquares();
              total = total + subgaus.getTotal();
              let sum2: Float64Array = subgaus.getSum();
              let min2: Float64Array = subgaus.getMin();
              let max2: Float64Array = subgaus.getMax();
              let sumsquares2: Float64Array = subgaus.getSumSquares();
              for (let i: number = 0; i < sum.length; i++) {
                sum[i] = sum[i] + sum2[i];
                if (min2[i] < min[i]) {
                  min[i] = min2[i];
                }
                if (max2[i] > max[i]) {
                  max[i] = max2[i];
                }
              }
              for (let i: number = 0; i < sumsquares.length; i++) {
                sumsquares[i] = sumsquares[i] + sumsquares2[i];
              }
              this.set(GaussianMixtureNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, total);
              this.set(GaussianMixtureNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
              this.set(GaussianMixtureNode.INTERNAL_MIN_KEY, org.mwg.Type.DOUBLE_ARRAY, min);
              this.set(GaussianMixtureNode.INTERNAL_MAX_KEY, org.mwg.Type.DOUBLE_ARRAY, max);
              this.set(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY, org.mwg.Type.DOUBLE_ARRAY, sumsquares);
              if (level > 0) {
                let subrelations: org.mwg.struct.Relation = <org.mwg.struct.Relation>subgaus.get(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN);
                if (subrelations == null || subrelations.size() == 0) {
                  subgaus.set(GaussianMixtureNode.LEVEL, org.mwg.Type.INT, level - 1);
                  if (level - 1 == 0) {
                    subgaus.remove(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN);
                  }
                  super.addToRelation(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN, subgaus);
                } else {
                  let oldrel: org.mwg.struct.Relation = <org.mwg.struct.Relation>this.getOrCreate(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN, org.mwg.Type.RELATION);
                  for (let i: number = 0; i < subrelations.size(); i++) {
                    oldrel.add(subrelations.get(i));
                  }
                }
              }
            }
            private static getSelectTraverseTask(): org.mwg.task.Task {
              let deepTraverse: org.mwg.task.Task = org.mwg.task.Tasks.newTask().loop("0", "{{requestedLev}}", org.mwg.task.Tasks.newTask().then(new org.mwg.ml.actions.ActionTraverseOrKeep(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN)).select((node : org.mwg.Node, context : org.mwg.task.TaskContext) => {
{
                  let childlev: number = <number>context.variable("rootLevel").get(0);
                  let i: number = <number>context.variable("i").get(0);
                  let finalMin: Float64Array = <Float64Array>context.variable("finalMin").get(0);
                  let finalMax: Float64Array = <Float64Array>context.variable("finalMax").get(0);
                  let err: Float64Array = <Float64Array>context.variable("err").get(0);
                  let threshold: number = <number>context.variable("threshold").get(0);
                  return (<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>node).checkInside(finalMin, finalMax, err, threshold, childlev - i);
                }              }));
              return deepTraverse;
            }
            private static getDeepTraverseTask(): org.mwg.task.Task {
              let deepTraverse: org.mwg.task.Task = org.mwg.task.Tasks.newTask().loop("0", "{{requestedLev}}", org.mwg.task.Tasks.newTask().then(new org.mwg.ml.actions.ActionTraverseOrKeep(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN)));
              return deepTraverse;
            }
            public query(level: number, min: Float64Array, max: Float64Array, callback: org.mwg.Callback<org.mwg.ml.algorithm.profiling.ProbaDistribution>): void {
              let nbfeature: number = this.getNumberOfFeatures();
              if (nbfeature == 0) {
                callback(null);
                return;
              }
              let resolved: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
              let initialResolution: Float64Array = <Float64Array>resolved.getFromKey(GaussianMixtureNode.RESOLUTION);
              if (initialResolution == null) {
                initialResolution = new Float64Array(nbfeature);
                for (let i: number = 0; i < nbfeature; i++) {
                  initialResolution[i] = 1;
                }
              }
              if (min == null) {
                min = this.getMin();
              }
              if (max == null) {
                max = this.getMax();
              }
              for (let i: number = 0; i < nbfeature; i++) {
                if ((max[i] - min[i]) < initialResolution[i]) {
                  min[i] = min[i] - initialResolution[i];
                  max[i] = min[i] + 2 * initialResolution[i];
                }
              }
              let finalMin: Float64Array = min;
              let finalMax: Float64Array = max;
              let err: Float64Array = initialResolution;
              let threshold: number = resolved.getFromKeyWithDefault(GaussianMixtureNode.THRESHOLD, GaussianMixtureNode.THRESHOLD_DEF);
              let context: org.mwg.task.TaskContext = GaussianMixtureNode.selectTraverseTask.prepare(this.graph(), this, (result : org.mwg.task.TaskResult<any>) => {
{
                  let covBackup: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(nbfeature, nbfeature);
                  for (let i: number = 0; i < nbfeature; i++) {
                    covBackup.set(i, i, err[i]);
                  }
                  let mvnBackup: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(null, covBackup, false);
                  let totals: Int32Array = new Int32Array(result.size());
                  let globalTotal: number = 0;
                  let distributions: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution[] = new Array<org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution>(result.size());
                  for (let i: number = 0; i < result.size(); i++) {
                    let temp: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = (<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>result.get(i));
                    totals[i] = temp.getTotal();
                    globalTotal += totals[i];
                    let avg: Float64Array = temp.getAvg();
                    if (totals[i] > 2) {
                      distributions[i] = new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(avg, temp.getCovariance(avg, err), false);
                      distributions[i].setMin(temp.getMin());
                      distributions[i].setMax(temp.getMax());
                    } else {
                      distributions[i] = mvnBackup.clone(avg);
                    }
                  }
                  result.free();
                  callback(new org.mwg.ml.algorithm.profiling.ProbaDistribution(totals, distributions, globalTotal));
                }              });
              let resMin: org.mwg.task.TaskResult<any> = context.newResult();
              resMin.add(finalMin);
              let resMax: org.mwg.task.TaskResult<any> = context.newResult();
              resMax.add(finalMax);
              let resErr: org.mwg.task.TaskResult<any> = context.newResult();
              resErr.add(err);
              context.setGlobalVariable("requestedLev", this.getLevel() - level);
              context.setGlobalVariable("rootLevel", level);
              context.setGlobalVariable("finalMin", resMin);
              context.setGlobalVariable("finalMax", resMax);
              context.setGlobalVariable("err", resErr);
              context.setGlobalVariable("threshold", threshold);
              GaussianMixtureNode.selectTraverseTask.executeUsing(context);
            }
            public generateDistributions(level: number, callback: org.mwg.Callback<org.mwg.ml.algorithm.profiling.ProbaDistribution>): void {
              let nbfeature: number = this.getNumberOfFeatures();
              if (nbfeature == 0) {
                callback(null);
                return;
              }
              let resolved: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
              let initialResolution: Float64Array = <Float64Array>resolved.getFromKey(GaussianMixtureNode.RESOLUTION);
              if (initialResolution == null) {
                initialResolution = new Float64Array(nbfeature);
                for (let i: number = 0; i < nbfeature; i++) {
                  initialResolution[i] = 1;
                }
              }
              let err: Float64Array = initialResolution;
              let context: org.mwg.task.TaskContext = GaussianMixtureNode.deepTraverseTask.prepare(this.graph(), this, (leaves : org.mwg.task.TaskResult<any>) => {
{
                  let covBackup: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(nbfeature, nbfeature);
                  for (let i: number = 0; i < nbfeature; i++) {
                    covBackup.set(i, i, err[i]);
                  }
                  let mvnBackup: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(null, covBackup, false);
                  let totals: Int32Array = new Int32Array(leaves.size());
                  let globalTotal: number = 0;
                  let distributions: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution[] = new Array<org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution>(leaves.size());
                  for (let i: number = 0; i < leaves.size(); i++) {
                    let temp: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = (<org.mwg.ml.algorithm.profiling.GaussianMixtureNode>leaves.get(i));
                    totals[i] = temp.getTotal();
                    globalTotal += totals[i];
                    let avg: Float64Array = temp.getAvg();
                    if (totals[i] > 2) {
                      distributions[i] = new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(avg, temp.getCovariance(avg, err), false);
                      distributions[i].setMin(temp.getMin());
                      distributions[i].setMax(temp.getMax());
                    } else {
                      distributions[i] = mvnBackup.clone(avg);
                    }
                  }
                  leaves.free();
                  callback(new org.mwg.ml.algorithm.profiling.ProbaDistribution(totals, distributions, globalTotal));
                }              });
              context.setGlobalVariable("requestedLev", this.getLevel() - level);
              GaussianMixtureNode.deepTraverseTask.executeUsing(context);
            }
            public toString(): string {
              return GaussianMixtureNode.NAME;
            }
            private internallearn(values: Float64Array, width: number, compressionFactor: number, compressionIter: number, resolution: Float64Array, threshold: number, createNode: boolean): void {
              let features: number = values.length;
              let reccursive: boolean = false;
              let total: number = this.getTotal();
              let level: number = this.getLevel();
              if (total == 0) {
                let sum: Float64Array = new Float64Array(features);
                java.lang.System.arraycopy(values, 0, sum, 0, features);
                total = 1;
                this.set(GaussianMixtureNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, total);
                this.set(GaussianMixtureNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
              } else {
                let sum: Float64Array;
                let min: Float64Array;
                let max: Float64Array;
                let sumsquares: Float64Array;
                if (total == 1) {
                  sum = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                  min = new Float64Array(features);
                  max = new Float64Array(features);
                  java.lang.System.arraycopy(sum, 0, min, 0, features);
                  java.lang.System.arraycopy(sum, 0, max, 0, features);
                  sumsquares = new Float64Array(features * (features + 1) / 2);
                  let count: number = 0;
                  for (let i: number = 0; i < features; i++) {
                    for (let j: number = i; j < features; j++) {
                      sumsquares[count] = sum[i] * sum[j];
                      count++;
                    }
                  }
                  if (createNode && level > 0) {
                    let newLev: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = this.createLevel(sum, level - 1, width, compressionFactor, compressionIter, resolution, threshold);
                    let d: number = org.mwg.ml.algorithm.profiling.GaussianMixtureNode.distance(values, sum, resolution);
                    if (d < threshold) {
                      reccursive = true;
                      newLev.internallearn(values, width, compressionFactor, compressionIter, resolution, threshold, createNode);
                    }
                    newLev.free();
                  }
                } else {
                  sum = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                  min = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_MIN_KEY);
                  max = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_MAX_KEY);
                  sumsquares = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
                }
                for (let i: number = 0; i < features; i++) {
                  if (values[i] < min[i]) {
                    min[i] = values[i];
                  }
                  if (values[i] > max[i]) {
                    max[i] = values[i];
                  }
                  sum[i] += values[i];
                }
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    sumsquares[count] += values[i] * values[j];
                    count++;
                  }
                }
                total++;
                if (createNode && level > 0 && !reccursive) {
                  let newLev: org.mwg.ml.algorithm.profiling.GaussianMixtureNode = this.createLevel(values, level - 1, width, compressionFactor, compressionIter, resolution, threshold);
                  newLev.free();
                  this.checkAndCompress(width, compressionFactor, compressionIter, resolution, threshold);
                }
                this.set(GaussianMixtureNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, total);
                this.set(GaussianMixtureNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
                this.set(GaussianMixtureNode.INTERNAL_MIN_KEY, org.mwg.Type.DOUBLE_ARRAY, min);
                this.set(GaussianMixtureNode.INTERNAL_MAX_KEY, org.mwg.Type.DOUBLE_ARRAY, max);
                this.set(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY, org.mwg.Type.DOUBLE_ARRAY, sumsquares);
              }
            }
            public getNumberOfFeatures(): number {
              let total: number = this.getTotal();
              if (total == 0) {
                return 0;
              } else {
                let sum: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                return sum.length;
              }
            }
            public getSum(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              } else {
                return <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
              }
            }
            public getSumSquares(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let sum: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                let features: number = sum.length;
                let sumsquares: Float64Array = new Float64Array(features * (features + 1) / 2);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    sumsquares[count] = sum[i] * sum[j];
                    count++;
                  }
                }
                return sumsquares;
              } else {
                return <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
              }
            }
            public getProbability(featArray: Float64Array, err: Float64Array, normalizeOnAvg: boolean): number {
              let sum: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
              let sumsquares: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
              let mnd: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution.getDistribution(sum, sumsquares, this.getTotal(), false);
              if (mnd == null) {
                return 0;
              } else {
                return mnd.density(featArray, normalizeOnAvg);
              }
            }
            public getProbabilityArray(featArray: Array<Float64Array>, err: Float64Array, normalizeOnAvg: boolean): Float64Array {
              let res: Float64Array = new Float64Array(featArray.length);
              let sum: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
              let sumsquares: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
              let mnd: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution.getDistribution(sum, sumsquares, this.getTotal(), false);
              if (mnd == null) {
                return res;
              } else {
                for (let i: number = 0; i < res.length; i++) {
                  res[i] = mnd.density(featArray[i], normalizeOnAvg);
                }
                return res;
              }
            }
            public getTotal(): number {
              let x: number = <number>super.get(GaussianMixtureNode.INTERNAL_TOTAL_KEY);
              if (x == null) {
                return 0;
              } else {
                return x;
              }
            }
            public getAvg(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                return <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
              } else {
                let avg: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                for (let i: number = 0; i < avg.length; i++) {
                  avg[i] = avg[i] / total;
                }
                return avg;
              }
            }
            public getCovarianceArray(avg: Float64Array, err: Float64Array): Float64Array {
              if (avg == null) {
                let errClone: Float64Array = new Float64Array(err.length);
                java.lang.System.arraycopy(err, 0, errClone, 0, err.length);
                return errClone;
              }
              if (err == null) {
                err = new Float64Array(avg.length);
              }
              let features: number = avg.length;
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total > 1) {
                let covariances: Float64Array = new Float64Array(features);
                let sumsquares: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  covariances[i] = (sumsquares[count] / total - avg[i] * avg[i]) * correction;
                  if (covariances[i] < err[i]) {
                    covariances[i] = err[i];
                  }
                  count += features - i;
                }
                return covariances;
              } else {
                let errClone: Float64Array = new Float64Array(err.length);
                java.lang.System.arraycopy(err, 0, errClone, 0, err.length);
                return errClone;
              }
            }
            public getCovariance(avg: Float64Array, err: Float64Array): org.mwg.struct.DMatrix {
              let features: number = avg.length;
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (err == null) {
                err = new Float64Array(avg.length);
              }
              if (total > 1) {
                let covariances: Float64Array = new Float64Array(features * features);
                let sumsquares: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUMSQUARE_KEY);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    covariances[i * features + j] = (sumsquares[count] / total - avg[i] * avg[j]) * correction;
                    covariances[j * features + i] = covariances[i * features + j];
                    count++;
                    if (covariances[i * features + i] < err[i]) {
                      covariances[i * features + i] = err[i];
                    }
                  }
                }
                return org.mwg.ml.common.matrix.VolatileDMatrix.wrap(covariances, features, features);
              } else {
                return null;
              }
            }
            public getMin(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let min: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                return min;
              } else {
                let min: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_MIN_KEY);
                return min;
              }
            }
            public getMax(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let max: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_SUM_KEY);
                return max;
              } else {
                let max: Float64Array = <Float64Array>super.get(GaussianMixtureNode.INTERNAL_MAX_KEY);
                return max;
              }
            }
            public getSubGraph(): Float64Array {
              let res: org.mwg.struct.Relation = <org.mwg.struct.Relation>super.get(GaussianMixtureNode.INTERNAL_SUBGAUSSIAN);
              if (res == null) {
                return null;
              }
              let reslong: Float64Array = new Float64Array(res.size());
              for (let i: number = 0; i < res.size(); i++) {
                reslong[i] = res.get(i);
              }
              return reslong;
            }
            public static distance(features: Float64Array, avg: Float64Array, resolution: Float64Array): number {
              let max: number = 0;
              let temp: number;
              for (let i: number = 0; i < features.length; i++) {
                temp = (features[i] - avg[i]) * (features[i] - avg[i]) / resolution[i];
                if (temp > max) {
                  max = temp;
                }
              }
              return Math.sqrt(max);
            }
          }
          export class GaussianNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.ProfilingNode {
            public static NAME: string = "GaussianNode";
            public static MIN: string = "min";
            public static MAX: string = "max";
            public static AVG: string = "avg";
            public static COV: string = "cov";
            public static PRECISION: string = "_precision";
            private static INTERNAL_SUM_KEY: string = "_sum";
            private static INTERNAL_SUMSQUARE_KEY: string = "_sumSquare";
            private static INTERNAL_TOTAL_KEY: string = "_total";
            private static INTERNAL_MIN_KEY: string = "_min";
            private static INTERNAL_MAX_KEY: string = "_max";
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public type(attributeName: string): number {
              if (attributeName === GaussianNode.AVG) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianNode.MIN) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianNode.MAX) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else if (attributeName === GaussianNode.COV) {
                return org.mwg.Type.DOUBLE_ARRAY;
              } else {
                return super.type(attributeName);
              }



            }
            public get(attributeName: string): any {
              if (attributeName === GaussianNode.AVG) {
                return this.getAvg();
              } else if (attributeName === GaussianNode.MIN) {
                return this.getMin();
              } else if (attributeName === GaussianNode.MAX) {
                return this.getMax();
              } else if (attributeName === GaussianNode.MAX) {
                return this.getMax();
              } else if (attributeName === GaussianNode.COV) {
                let resolved: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
                let initialPrecision: Float64Array = <Float64Array>resolved.getFromKey(GaussianNode.PRECISION);
                let nbfeature: number = this.getNumberOfFeatures();
                if (initialPrecision == null) {
                  initialPrecision = new Float64Array(nbfeature);
                  for (let i: number = 0; i < nbfeature; i++) {
                    initialPrecision[i] = 1;
                  }
                }
                return this.getCovariance(this.getAvg(), initialPrecision);
              } else {
                return super.get(attributeName);
              }




            }
            public learn(callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((values : Float64Array) => {
{
                  this.learnWith(values);
                  callback(true);
                }              });
            }
            public learnWith(values: Float64Array): void {
              let features: number = values.length;
              let total: number = this.getTotal();
              if (total == 0) {
                let sum: Float64Array = new Float64Array(features);
                java.lang.System.arraycopy(values, 0, sum, 0, features);
                total = 1;
                this.set(GaussianNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, total);
                this.set(GaussianNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
              } else {
                let sum: Float64Array;
                let min: Float64Array;
                let max: Float64Array;
                let sumsquares: Float64Array;
                if (total == 1) {
                  sum = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                  min = new Float64Array(features);
                  max = new Float64Array(features);
                  java.lang.System.arraycopy(sum, 0, min, 0, features);
                  java.lang.System.arraycopy(sum, 0, max, 0, features);
                  sumsquares = new Float64Array(features * (features + 1) / 2);
                  let count: number = 0;
                  for (let i: number = 0; i < features; i++) {
                    for (let j: number = i; j < features; j++) {
                      sumsquares[count] = sum[i] * sum[j];
                      count++;
                    }
                  }
                } else {
                  sum = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                  min = <Float64Array>super.get(GaussianNode.INTERNAL_MIN_KEY);
                  max = <Float64Array>super.get(GaussianNode.INTERNAL_MAX_KEY);
                  sumsquares = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
                }
                for (let i: number = 0; i < features; i++) {
                  if (values[i] < min[i]) {
                    min[i] = values[i];
                  }
                  if (values[i] > max[i]) {
                    max[i] = values[i];
                  }
                  sum[i] += values[i];
                }
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    sumsquares[count] += values[i] * values[j];
                    count++;
                  }
                }
                total++;
                this.set(GaussianNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, total);
                this.set(GaussianNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
                this.set(GaussianNode.INTERNAL_MIN_KEY, org.mwg.Type.DOUBLE_ARRAY, min);
                this.set(GaussianNode.INTERNAL_MAX_KEY, org.mwg.Type.DOUBLE_ARRAY, max);
                this.set(GaussianNode.INTERNAL_SUMSQUARE_KEY, org.mwg.Type.DOUBLE_ARRAY, sumsquares);
              }
            }
            public predict(callback: org.mwg.Callback<Float64Array>): void {}
            public toString(): string {
              return GaussianNode.NAME;
            }
            public getNumberOfFeatures(): number {
              let total: number = this.getTotal();
              if (total == 0) {
                return 0;
              } else {
                let sum: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                return sum.length;
              }
            }
            public getSum(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              } else {
                return <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
              }
            }
            public getSumSquares(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let sum: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                let features: number = sum.length;
                let sumsquares: Float64Array = new Float64Array(features * (features + 1) / 2);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    sumsquares[count] = sum[i] * sum[j];
                    count++;
                  }
                }
                return sumsquares;
              } else {
                return <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
              }
            }
            public getProbability(featArray: Float64Array, err: Float64Array, normalizeOnAvg: boolean): number {
              let sum: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
              let sumsquares: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
              let mnd: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution.getDistribution(sum, sumsquares, this.getTotal(), false);
              if (mnd == null) {
                return 0;
              } else {
                return mnd.density(featArray, normalizeOnAvg);
              }
            }
            public getExponent(featArray: Float64Array): number {
              let sum: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
              let sumsquares: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
              let mnd: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution.getDistribution(sum, sumsquares, this.getTotal(), false);
              if (mnd == null) {
                return 0;
              } else {
                return mnd.densityExponent(featArray);
              }
            }
            public getProbabilityArray(featArray: Array<Float64Array>, err: Float64Array, normalizeOnAvg: boolean): Float64Array {
              let res: Float64Array = new Float64Array(featArray.length);
              let sum: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
              let sumsquares: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
              let mnd: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution.getDistribution(sum, sumsquares, this.getTotal(), false);
              if (mnd == null) {
                return res;
              } else {
                for (let i: number = 0; i < res.length; i++) {
                  res[i] = mnd.density(featArray[i], normalizeOnAvg);
                }
                return res;
              }
            }
            public getTotal(): number {
              let x: number = <number>super.get(GaussianNode.INTERNAL_TOTAL_KEY);
              if (x == null) {
                return 0;
              } else {
                return x;
              }
            }
            public getAvg(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                return <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
              } else {
                let avg: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                for (let i: number = 0; i < avg.length; i++) {
                  avg[i] = avg[i] / total;
                }
                return avg;
              }
            }
            public getCovarianceArray(avg: Float64Array, err: Float64Array): Float64Array {
              if (avg == null) {
                let errClone: Float64Array = new Float64Array(err.length);
                java.lang.System.arraycopy(err, 0, errClone, 0, err.length);
                return errClone;
              }
              if (err == null) {
                err = new Float64Array(avg.length);
              }
              let features: number = avg.length;
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total > 1) {
                let covariances: Float64Array = new Float64Array(features);
                let sumsquares: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  covariances[i] = (sumsquares[count] / total - avg[i] * avg[i]) * correction;
                  if (covariances[i] < err[i]) {
                    covariances[i] = err[i];
                  }
                  count += features - i;
                }
                return covariances;
              } else {
                let errClone: Float64Array = new Float64Array(err.length);
                java.lang.System.arraycopy(err, 0, errClone, 0, err.length);
                return errClone;
              }
            }
            public getCovariance(avg: Float64Array, err: Float64Array): org.mwg.struct.DMatrix {
              let features: number = avg.length;
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (err == null) {
                err = new Float64Array(avg.length);
              }
              if (total > 1) {
                let covariances: Float64Array = new Float64Array(features * features);
                let sumsquares: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUMSQUARE_KEY);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    covariances[i * features + j] = (sumsquares[count] / total - avg[i] * avg[j]) * correction;
                    covariances[j * features + i] = covariances[i * features + j];
                    count++;
                    if (covariances[i * features + i] < err[i]) {
                      covariances[i * features + i] = err[i];
                    }
                  }
                }
                return org.mwg.ml.common.matrix.VolatileDMatrix.wrap(covariances, features, features);
              } else {
                return null;
              }
            }
            public getMin(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let min: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                return min;
              } else {
                let min: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_MIN_KEY);
                return min;
              }
            }
            public getMax(): Float64Array {
              let total: number = this.getTotal();
              if (total == 0) {
                return null;
              }
              if (total == 1) {
                let max: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_SUM_KEY);
                return max;
              } else {
                let max: Float64Array = <Float64Array>super.get(GaussianNode.INTERNAL_MAX_KEY);
                return max;
              }
            }
            public static distance(features: Float64Array, avg: Float64Array, precisions: Float64Array): number {
              let max: number = 0;
              let temp: number;
              for (let i: number = 0; i < features.length; i++) {
                temp = (features[i] - avg[i]) * (features[i] - avg[i]) / precisions[i];
                if (temp > max) {
                  max = temp;
                }
              }
              return Math.sqrt(max);
            }
          }
          export class GaussianSlotNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.ProfilingNode {
            public static NAME: string = "GaussianSlotProfiling";
            public static SLOTS_NUMBER: string = "SLOTS_NUMBER";
            public static SLOTS_NUMBER_DEF: number = 1;
            public static PERIOD_SIZE: string = "PERIOD_SIZE";
            public static PERIOD_SIZE_DEF: number = 24 * 3600 * 1000;
            private static INTERNAL_FEATURES_NUMBER: string = "_featuresNb";
            private static INTERNAL_TOTAL_KEY: string = "_total";
            private static INTERNAL_MIN_KEY: string = "_min";
            private static INTERNAL_MAX_KEY: string = "_max";
            private static INTERNAL_SUM_KEY: string = "_sum";
            private static INTERNAL_SUMSQUARE_KEY: string = "_sumSquare";
            private static enforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asPositiveInt(GaussianSlotNode.SLOTS_NUMBER).asPositiveLong(GaussianSlotNode.PERIOD_SIZE);
            public learn(callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((values : Float64Array) => {
{
                  this.learnWith(values);
                  callback(true);
                }              });
            }
            public learnWith(values: Float64Array): void {
              let insTime: number = this.time();
              let resolved: org.mwg.plugin.NodeState = this.phasedState();
              let numOfSlot: number = resolved.getFromKeyWithDefault(GaussianSlotNode.SLOTS_NUMBER, GaussianSlotNode.SLOTS_NUMBER_DEF);
              let total: Int32Array;
              let min: Float64Array;
              let max: Float64Array;
              let sum: Float64Array;
              let sumSquare: Float64Array;
              let features: number = values.length;
              total = <Int32Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_TOTAL_KEY);
              if (numOfSlot == 1 || numOfSlot == 0) {
                if (total == null) {
                  resolved.setFromKey(GaussianSlotNode.INTERNAL_FEATURES_NUMBER, org.mwg.Type.INT, features);
                  total = new Int32Array(1);
                  min = new Float64Array(features);
                  max = new Float64Array(features);
                  sum = new Float64Array(features);
                  sumSquare = new Float64Array(features * (features + 1) / 2);
                } else {
                  min = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_MIN_KEY);
                  max = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_MAX_KEY);
                  sum = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUM_KEY);
                  sumSquare = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUMSQUARE_KEY);
                }
                this.update(total, min, max, sum, sumSquare, values, 0, features, 0, 0);
                return;
              } else {
                if (total == null) {
                  resolved.setFromKey(GaussianSlotNode.INTERNAL_FEATURES_NUMBER, org.mwg.Type.INT, features);
                  total = new Int32Array(numOfSlot + 1);
                  min = new Float64Array((numOfSlot + 1) * features);
                  max = new Float64Array((numOfSlot + 1) * features);
                  sum = new Float64Array((numOfSlot + 1) * features);
                  sumSquare = new Float64Array((numOfSlot + 1) * features * (features + 1) / 2);
                } else {
                  min = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_MIN_KEY);
                  max = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_MAX_KEY);
                  sum = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUM_KEY);
                  sumSquare = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUMSQUARE_KEY);
                }
                let periodSize: number = resolved.getFromKeyWithDefault(GaussianSlotNode.PERIOD_SIZE, GaussianSlotNode.PERIOD_SIZE_DEF);
                let slot: number = org.mwg.ml.algorithm.profiling.GaussianSlotNode.getIntTime(insTime, numOfSlot, periodSize);
                let index: number = slot * features;
                let indexSquare: number = slot * features * (features + 1) / 2;
                let indexTot: number = numOfSlot * features;
                let indexSquareTot: number = numOfSlot * features * (features + 1) / 2;
                this.update(total, min, max, sum, sumSquare, values, slot, features, index, indexSquare);
                this.update(total, min, max, sum, sumSquare, values, numOfSlot, features, indexTot, indexSquareTot);
              }
              resolved.setFromKey(GaussianSlotNode.INTERNAL_FEATURES_NUMBER, org.mwg.Type.INT, features);
              resolved.setFromKey(GaussianSlotNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT_ARRAY, total);
              resolved.setFromKey(GaussianSlotNode.INTERNAL_MIN_KEY, org.mwg.Type.DOUBLE_ARRAY, min);
              resolved.setFromKey(GaussianSlotNode.INTERNAL_MAX_KEY, org.mwg.Type.DOUBLE_ARRAY, max);
              resolved.setFromKey(GaussianSlotNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sum);
              resolved.setFromKey(GaussianSlotNode.INTERNAL_SUMSQUARE_KEY, org.mwg.Type.DOUBLE_ARRAY, sumSquare);
            }
            public predict(callback: org.mwg.Callback<Float64Array>): void {
              let resolved: org.mwg.plugin.NodeState = this.unphasedState();
              let features: number = resolved.getFromKeyWithDefault(GaussianSlotNode.INTERNAL_FEATURES_NUMBER, 0);
              if (features == 0) {
                callback(null);
                return;
              }
              let numOfSlot: number = resolved.getFromKeyWithDefault(GaussianSlotNode.SLOTS_NUMBER, GaussianSlotNode.SLOTS_NUMBER_DEF);
              let periodSize: number = resolved.getFromKeyWithDefault(GaussianSlotNode.PERIOD_SIZE, GaussianSlotNode.PERIOD_SIZE_DEF);
              let slot: number = org.mwg.ml.algorithm.profiling.GaussianSlotNode.getIntTime(this.time(), numOfSlot, periodSize);
              let index: number = slot * features;
              let total: Int32Array = <Int32Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_TOTAL_KEY);
              let sum: Float64Array = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUM_KEY);
              let result: Float64Array = new Float64Array(features);
              if (total != null) {
                if (total[slot] != 0) {
                  for (let j: number = 0; j < features; j++) {
                    result[j] = sum[j + index] / total[slot];
                  }
                }
              }
              callback(result);
            }
            public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
              GaussianSlotNode.enforcer.check(propertyName, propertyType, propertyValue);
              return super.set(propertyName, propertyType, propertyValue);
            }
            public get(attributeName: string): any {
              let state: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
              if (attributeName === GaussianSlotNode.SLOTS_NUMBER) {
                return state.getFromKeyWithDefault(GaussianSlotNode.SLOTS_NUMBER, GaussianSlotNode.SLOTS_NUMBER_DEF);
              } else if (attributeName === GaussianSlotNode.PERIOD_SIZE) {
                return state.getFromKeyWithDefault(GaussianSlotNode.PERIOD_SIZE, GaussianSlotNode.PERIOD_SIZE_DEF);
              } else {
                return super.get(attributeName);
              }

            }
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public static getIntTime(time: number, numOfSlot: number, periodSize: number): number {
              if (numOfSlot <= 1) {
                return 0;
              }
              let res: number = time % periodSize;
              res = res / (periodSize / numOfSlot);
              return <number>res;
            }
            private update(total: Int32Array, min: Float64Array, max: Float64Array, sum: Float64Array, sumSquare: Float64Array, values: Float64Array, slot: number, features: number, index: number, indexSquare: number): void {
              if (total[slot] == 1) {
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  min[index + i] = values[i];
                  max[index + i] = values[i];
                  sum[index + i] = values[i];
                  for (let j: number = i; j < features; j++) {
                    sumSquare[indexSquare + count] += values[i] * values[j];
                    count++;
                  }
                }
              } else {
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  if (values[i] < min[index + i]) {
                    min[index + i] = values[i];
                  }
                  if (values[i] > max[index + i]) {
                    max[index + i] = values[i];
                  }
                  sum[index + i] += values[i];
                  for (let j: number = i; j < features; j++) {
                    sumSquare[indexSquare + count] += values[i] * values[j];
                    count++;
                  }
                }
              }
              total[slot] += 1;
            }
            public getMin(): Float64Array {
              return <Float64Array>this.unphasedState().getFromKey(GaussianSlotNode.INTERNAL_MIN_KEY);
            }
            public getMax(): Float64Array {
              return <Float64Array>this.unphasedState().getFromKey(GaussianSlotNode.INTERNAL_MAX_KEY);
            }
            public getSum(): Float64Array {
              return <Float64Array>this.unphasedState().getFromKey(GaussianSlotNode.INTERNAL_SUM_KEY);
            }
            public getSumSquare(): Float64Array {
              return <Float64Array>this.unphasedState().getFromKey(GaussianSlotNode.INTERNAL_SUMSQUARE_KEY);
            }
            public getTotal(): Int32Array {
              return <Int32Array>this.unphasedState().getFromKey(GaussianSlotNode.INTERNAL_TOTAL_KEY);
            }
            public getAvg(): Float64Array {
              let resolved: org.mwg.plugin.NodeState = this.unphasedState();
              let numOfSlot: number = resolved.getFromKeyWithDefault(GaussianSlotNode.SLOTS_NUMBER, GaussianSlotNode.SLOTS_NUMBER_DEF);
              let features: number = resolved.getFromKeyWithDefault(GaussianSlotNode.INTERNAL_FEATURES_NUMBER, 0);
              if (features == 0) {
                return null;
              }
              let total: Int32Array = <Int32Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_TOTAL_KEY);
              let sum: Float64Array = <Float64Array>resolved.getFromKey(GaussianSlotNode.INTERNAL_SUM_KEY);
              let result: Float64Array = new Float64Array(sum.length);
              if (total != null) {
                if (numOfSlot > 1) {
                  let count: number = 0;
                  for (let i: number = 0; i < (numOfSlot + 1); i++) {
                    if (total[i] != 0) {
                      for (let j: number = 0; j < features; j++) {
                        result[count] = sum[count] / total[i];
                        count++;
                      }
                    } else {
                      count += features;
                    }
                  }
                } else {
                  if (total[0] != 0) {
                    for (let j: number = 0; j < features; j++) {
                      result[j] = sum[j] / total[0];
                    }
                  }
                }
              }
              return result;
            }
          }
          export class KMeans {
            public getClusterIds(data: Array<Float64Array>, numOfCluster: number, numOfIterations: number, div: Float64Array): Array<Int32Array> {
              let result: Array<Int32Array> = new Array<Int32Array>(numOfCluster);
              let features: number = data[0].length;
              let totals: Int32Array = new Int32Array(numOfCluster);
              let centroids: Array<Float64Array> = new Array<Float64Array>(numOfCluster);
              for (let i: number = 0; i < numOfCluster; i++) {
                centroids[i] = new Float64Array(features);
              }
              let categories: Int32Array = new Int32Array(data.length);
              for (let i: number = 0; i < numOfCluster; i++) {
                java.lang.System.arraycopy(data[i], 0, centroids[i], 0, features);
              }
              for (let iter: number = 0; iter < numOfIterations; iter++) {
                for (let i: number = 0; i < totals.length; i++) {
                  totals[i] = 0;
                }
                for (let i: number = 0; i < data.length; i++) {
                  categories[i] = this.calculateCategory(data[i], centroids, div);
                  totals[categories[i]]++;
                }
                for (let i: number = 0; i < centroids.length; i++) {
                  for (let j: number = 0; j < features; j++) {
                    centroids[i][j] = 0;
                  }
                }
                for (let i: number = 0; i < data.length; i++) {
                  for (let j: number = 0; j < features; j++) {
                    centroids[categories[i]][j] += data[i][j];
                  }
                }
                for (let i: number = 0; i < centroids.length; i++) {
                  if (totals[i] != 0) {
                    for (let j: number = 0; j < features; j++) {
                      centroids[i][j] = centroids[i][j] / totals[i];
                    }
                  } else {
                    let rand: java.util.Random = new java.util.Random();
                    let avg: Float64Array = data[rand.nextInt(data.length)];
                    java.lang.System.arraycopy(avg, 0, centroids[i], 0, features);
                  }
                }
              }
              for (let i: number = 0; i < numOfCluster; i++) {
                result[i] = new Int32Array(totals[i]);
                let k: number = 0;
                for (let j: number = 0; j < data.length; j++) {
                  if (categories[j] == i) {
                    result[i][k] = j;
                    k++;
                  }
                }
              }
              return result;
            }
            private calculateCategory(values: Float64Array, centroids: Array<Float64Array>, div: Float64Array): number {
              let min: number = java.lang.Double.MAX_VALUE;
              let pos: number = 0;
              for (let i: number = 0; i < centroids.length; i++) {
                let d: number = org.mwg.ml.algorithm.profiling.KMeans.distance(values, centroids[i], div);
                if (d < min) {
                  min = d;
                  pos = i;
                }
              }
              return pos;
            }
            public static distance(features: Float64Array, avg: Float64Array, precisions: Float64Array): number {
              let max: number = 0;
              let temp: number;
              for (let i: number = 0; i < features.length; i++) {
                temp = (features[i] - avg[i]) * (features[i] - avg[i]) / precisions[i];
                if (temp > max) {
                  max = temp;
                }
              }
              return Math.sqrt(max);
            }
          }
          export class ProbaDistribution {
            public distributions: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution[];
            public total: Int32Array;
            public global: number;
            constructor(total: Int32Array, distributions: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution[], global: number) {
              this.total = total;
              this.distributions = distributions;
              this.global = global;
            }
            public calculate(features: Float64Array): number {
              let result: number = 0;
              if (this.total != null) {
                for (let j: number = 0; j < this.distributions.length; j++) {
                  if (org.mwg.ml.algorithm.profiling.KMeans.distance(features, this.distributions[j].getAvg(), this.distributions[j].getCovDiag()) < 5) {
                    result += this.distributions[j].density(features, false) * this.total[j] / this.global;
                  }
                }
              } else {
                for (let j: number = 0; j < this.distributions.length; j++) {
                  if (org.mwg.ml.algorithm.profiling.KMeans.distance(features, this.distributions[j].getAvg(), this.distributions[j].getCovDiag()) < 5) {
                    result += this.distributions[j].density(features, false) / this.global;
                  }
                }
              }
              if (result > 1) {
                result = 1;
              }
              return result;
            }
            public calculateArray(features: Array<Float64Array>, reporter: org.mwg.ml.ProgressReporter): Float64Array {
              if (reporter != null) {
                reporter.updateInformation("Number of distributions: " + this.distributions.length + " , values: " + this.global);
              }
              let result: Float64Array = new Float64Array(features.length);
              let calibration: number = 0;
              for (let i: number = 0; i < features.length; i++) {
                result[i] = this.calculate(features[i]);
                calibration += result[i];
                if (reporter != null) {
                  let progress: number = i * (1.0 / (features.length));
                  progress = progress * 50 + 50;
                  reporter.updateProgress(<number>progress);
                  if (reporter.isCancelled()) {
                    return null;
                  }
                }
              }
              if (calibration != 0) {
                for (let i: number = 0; i < features.length; i++) {
                  result[i] = result[i] / calibration;
                }
              }
              return result;
            }
            public addUpProbabilities(features: Array<Float64Array>): number {
              let res: number = 0;
              for (let i: number = 0; i < features.length; i++) {
                res += this.calculate(features[i]);
              }
              return res;
            }
            public compareProbaDistribution(other: org.mwg.ml.algorithm.profiling.ProbaDistribution, features: Array<Float64Array>): Float64Array {
              let error: Float64Array = new Float64Array(2);
              let res1: Float64Array = this.calculateArray(features, null);
              let res2: Float64Array = other.calculateArray(features, null);
              let temp: number = 0;
              for (let i: number = 0; i < res1.length; i++) {
                temp = Math.abs(res1[i] - res2[i]);
                error[0] += temp;
                if (temp > error[1]) {
                  error[1] = temp;
                }
              }
              error[0] = error[0] / res1.length;
              return error;
            }
          }
        }
        export module regression {
          export class LiveLinearRegressionNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.RegressionNode {
            public static ALPHA_KEY: string = "ALPHA";
            public static ALPHA_DEF: number = 0.0001;
            public static LAMBDA_KEY: string = "LAMBDA";
            public static LAMBDA_DEF: number = 0.00001;
            public static ITERATION_KEY: string = "ITERATION";
            public static ITERATION_DEF: number = 5;
            public static THRESHOLD_KEY: string = "THRESHOLD";
            public static THRESHOLD_DEF: number = 0.01;
            public static LAST_ERR_KEY: string = "_ERR";
            public static WEIGHT_KEY: string = "_WEIGHT";
            private static INTERNAL_TOTAL_KEY: string = "_TOTAL_KEY";
            private static INTERNAL_WEIGHT_BACKUP_KEY: string = "_WEIGHTBACKUP";
            private static MISMATCH_MSG: string = "Different Imput lengths are not supported";
            public static NAME: string = "LiveLinearRegression";
            private static enforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asDouble(LiveLinearRegressionNode.ALPHA_KEY).asDouble(LiveLinearRegressionNode.LAMBDA_KEY).asInt(LiveLinearRegressionNode.ITERATION_KEY);
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public learn(output: number, callback: org.mwg.Callback<boolean>): void {
              this.extractFeatures((input : Float64Array) => {
{
                  this.internalLearn(input, output, callback);
                }              });
            }
            public internalLearn(input: Float64Array, output: number, callback: org.mwg.Callback<boolean>): void {
              let state: org.mwg.plugin.NodeState = this._resolver.alignState(this);
              let iterations: number = state.getFromKeyWithDefault(LiveLinearRegressionNode.ITERATION_KEY, LiveLinearRegressionNode.ITERATION_DEF);
              let alpha: number = state.getFromKeyWithDefault(LiveLinearRegressionNode.ALPHA_KEY, LiveLinearRegressionNode.ALPHA_DEF);
              let lambda: number = state.getFromKeyWithDefault(LiveLinearRegressionNode.LAMBDA_KEY, LiveLinearRegressionNode.LAMBDA_DEF);
              let weights: Float64Array = <Float64Array>state.getFromKey(LiveLinearRegressionNode.WEIGHT_KEY);
              if (weights == null) {
                weights = new Float64Array(input.length + 1);
                let random: java.util.Random = new java.util.Random();
                for (let i: number = 0; i < weights.length; i++) {
                  weights[i] = random.nextDouble() * 0.001;
                }
              }
              let prevErr: number = state.getFromKeyWithDefault(LiveLinearRegressionNode.LAST_ERR_KEY, 0.0);
              let currErr: number = this.calculate(weights, input) - output;
              state.setFromKey(LiveLinearRegressionNode.LAST_ERR_KEY, org.mwg.Type.DOUBLE, currErr);
              if (input == null || weights.length != (input.length + 1)) {
                throw new Error(LiveLinearRegressionNode.MISMATCH_MSG);
              }
              let featuresize: number = input.length;
              for (let j: number = 0; j < iterations; j++) {
                let h: number = this.calculate(weights, input) - output;
                for (let i: number = 0; i < featuresize; i++) {
                  weights[i] = weights[i] - alpha * (h * input[i] + lambda * weights[i]);
                }
                weights[featuresize] = weights[featuresize] - alpha * h;
              }
              let bckupWeight: Float64Array = <Float64Array>state.getFromKey(LiveLinearRegressionNode.INTERNAL_WEIGHT_BACKUP_KEY);
              if (bckupWeight == null) {
                state.setFromKey(LiveLinearRegressionNode.WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weights);
                state.setFromKey(LiveLinearRegressionNode.INTERNAL_WEIGHT_BACKUP_KEY, org.mwg.Type.DOUBLE_ARRAY, weights);
                state.setFromKey(LiveLinearRegressionNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, 1);
              } else {
                let diff: number = 0;
                for (let i: number = 0; i < weights.length; i++) {
                  diff = Math.max(diff, Math.abs(weights[i] - bckupWeight[i]));
                }
                let deviation: number = state.getFromKeyWithDefault(LiveLinearRegressionNode.THRESHOLD_KEY, LiveLinearRegressionNode.THRESHOLD_DEF);
                if (diff > deviation) {
                  state = this.phasedState();
                  state.setFromKey(LiveLinearRegressionNode.WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weights);
                  state.setFromKey(LiveLinearRegressionNode.INTERNAL_WEIGHT_BACKUP_KEY, org.mwg.Type.DOUBLE_ARRAY, weights);
                  state.setFromKey(LiveLinearRegressionNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, 1);
                } else {
                  state.setFromKey(LiveLinearRegressionNode.WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weights);
                  state.setFromKey(LiveLinearRegressionNode.INTERNAL_TOTAL_KEY, org.mwg.Type.INT, <number>state.getFromKey(LiveLinearRegressionNode.INTERNAL_TOTAL_KEY) + 1);
                }
              }
              if (callback != null) {
                callback(true);
              }
            }
            public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
              LiveLinearRegressionNode.enforcer.check(propertyName, propertyType, propertyValue);
              return super.set(propertyName, propertyType, propertyValue);
            }
            private calculate(weights: Float64Array, input: Float64Array): number {
              let h: number = 0;
              for (let j: number = 0; j < input.length; j++) {
                h += weights[j] * input[j];
              }
              h += weights[input.length];
              return h;
            }
            public extrapolate(callback: org.mwg.Callback<number>): void {
              let state: org.mwg.plugin.NodeState = this._resolver.resolveState(this);
              let weights: Float64Array = <Float64Array>state.getFromKey(LiveLinearRegressionNode.WEIGHT_KEY);
              if (weights == null) {
                if (callback != null) {
                  callback(0.0);
                }
              } else {
                this.extractFeatures((input : Float64Array) => {
{
                    if (input.length != weights.length - 1) {
                      throw new Error(org.mwg.ml.algorithm.regression.LiveLinearRegressionNode.MISMATCH_MSG);
                    } else if (callback != null) {
                      callback(this.calculate(weights, input));
                    }

                  }                });
              }
            }
          }
          export class PolynomialNode extends org.mwg.ml.BaseMLNode implements org.mwg.ml.RegressionNode {
            public static PRECISION: string = "precision";
            public static PRECISION_DEF: number = 1;
            public static VALUE: string = "value";
            public static NAME: string = "PolynomialNode";
            public static INTERNAL_WEIGHT_KEY: string = "weight";
            public static INTERNAL_STEP_KEY: string = "step";
            private static INTERNAL_TIME_BUFFER: string = "times";
            private static INTERNAL_VALUES_BUFFER: string = "values";
            private static INTERNAL_NB_PAST_KEY: string = "nb";
            private static INTERNAL_LAST_TIME_KEY: string = "lastTime";
            public static MAX_DEGREE: string = "maxdegree";
            public static MAX_DEGREE_DEF: number = 20;
            private static NOT_MANAGED_ATT_ERROR: string = "Polynomial node can only handle value attribute, please use a super node to store other data";
            private static enforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asPositiveDouble(PolynomialNode.PRECISION);
            private static degenforcer: org.mwg.utility.Enforcer = new org.mwg.utility.Enforcer().asPositiveInt(PolynomialNode.MAX_DEGREE);
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
              if (propertyName === PolynomialNode.VALUE) {
                this.learn(parseFloat(propertyValue.toString()), null);
              } else if (propertyName === PolynomialNode.PRECISION) {
                PolynomialNode.enforcer.check(propertyName, propertyType, propertyValue);
                super.set(propertyName, org.mwg.Type.DOUBLE, propertyValue);
              } else if (propertyName === PolynomialNode.MAX_DEGREE) {
                PolynomialNode.degenforcer.check(propertyName, propertyType, propertyValue);
                super.set(propertyName, org.mwg.Type.INT, <number>propertyValue);
              } else {
                throw new Error(PolynomialNode.NOT_MANAGED_ATT_ERROR);
              }


              return this;
            }
            public get(propertyName: string): any {
              if (propertyName === PolynomialNode.VALUE) {
                let res: number[] = [null];
                this.extrapolate((result : number) => {
{
                    res[0] = result;
                  }                });
                return res[0];
              } else {
                return super.get(propertyName);
              }
            }
            public learn(value: number, callback: org.mwg.Callback<boolean>): void {
              let previousState: org.mwg.plugin.NodeState = this.unphasedState();
              let timeOrigin: number = previousState.time();
              let nodeTime: number = this.time();
              let precision: number = previousState.getFromKeyWithDefault(PolynomialNode.PRECISION, PolynomialNode.PRECISION_DEF);
              let weight: Float64Array = <Float64Array>previousState.getFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY);
              if (weight == null) {
                weight = new Float64Array(1);
                weight[0] = value;
                previousState.setFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weight);
                previousState.setFromKey(PolynomialNode.INTERNAL_NB_PAST_KEY, org.mwg.Type.INT, 1);
                previousState.setFromKey(PolynomialNode.INTERNAL_STEP_KEY, org.mwg.Type.LONG, 0);
                previousState.setFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY, org.mwg.Type.LONG, 0);
                previousState.setFromKey(PolynomialNode.INTERNAL_TIME_BUFFER, org.mwg.Type.DOUBLE_ARRAY, new Float64Array([0]));
                previousState.setFromKey(PolynomialNode.INTERNAL_VALUES_BUFFER, org.mwg.Type.DOUBLE_ARRAY, new Float64Array([value]));
                if (callback != null) {
                  callback(true);
                }
                return;
              }
              let previousTime: number = timeOrigin + <number>previousState.getFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY);
              if (nodeTime > previousTime) {
                let stp: number = <number>previousState.getFromKey(PolynomialNode.INTERNAL_STEP_KEY);
                let lastTime: number = nodeTime - timeOrigin;
                if (stp == null || stp == 0) {
                  if (lastTime == 0) {
                    weight = new Float64Array(1);
                    weight[0] = value;
                    previousState.setFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weight);
                    previousState.setFromKey(PolynomialNode.INTERNAL_TIME_BUFFER, org.mwg.Type.DOUBLE_ARRAY, new Float64Array([0]));
                    previousState.setFromKey(PolynomialNode.INTERNAL_VALUES_BUFFER, org.mwg.Type.DOUBLE_ARRAY, new Float64Array([value]));
                    if (callback != null) {
                      callback(true);
                    }
                    return;
                  } else {
                    stp = lastTime;
                    previousState.setFromKey(PolynomialNode.INTERNAL_STEP_KEY, org.mwg.Type.LONG, stp);
                  }
                }
                let deg: number = weight.length - 1;
                let num: number = <number>previousState.getFromKey(PolynomialNode.INTERNAL_NB_PAST_KEY);
                let t: number = (nodeTime - timeOrigin);
                t = t / stp;
                let maxError: number = this.maxErr(precision, deg);
                let maxd: number = previousState.getFromKeyWithDefault(PolynomialNode.MAX_DEGREE, PolynomialNode.MAX_DEGREE_DEF);
                let times: Float64Array = org.mwg.ml.algorithm.regression.PolynomialNode.updateBuffer(previousState, t, maxd, PolynomialNode.INTERNAL_TIME_BUFFER);
                let values: Float64Array = org.mwg.ml.algorithm.regression.PolynomialNode.updateBuffer(previousState, value, maxd, PolynomialNode.INTERNAL_VALUES_BUFFER);
                if (Math.abs(org.mwg.ml.common.matrix.operation.PolynomialFit.extrapolate(t, weight) - value) <= maxError) {
                  previousState.setFromKey(PolynomialNode.INTERNAL_NB_PAST_KEY, org.mwg.Type.INT, num + 1);
                  previousState.setFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY, org.mwg.Type.LONG, lastTime);
                  if (callback != null) {
                    callback(true);
                  }
                  return;
                }
                let newdeg: number = Math.min(times.length, maxd);
                while (deg < newdeg && times.length < maxd * 4) {
                  maxError = this.maxErr(precision, deg);
                  let pf: org.mwg.ml.common.matrix.operation.PolynomialFit = new org.mwg.ml.common.matrix.operation.PolynomialFit(deg);
                  pf.fit(times, values);
                  if (this.tempError(pf.getCoef(), times, values) <= maxError) {
                    weight = pf.getCoef();
                    previousState.setFromKey(PolynomialNode.INTERNAL_NB_PAST_KEY, org.mwg.Type.INT, num + 1);
                    previousState.setFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weight);
                    previousState.setFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY, org.mwg.Type.LONG, lastTime);
                    if (callback != null) {
                      callback(true);
                    }
                    return;
                  }
                  deg++;
                }
                let newstep: number = nodeTime - previousTime;
                let phasedState: org.mwg.plugin.NodeState = this.newState(previousTime);
                let nvalues: Float64Array = new Float64Array(2);
                let ntimes: Float64Array = new Float64Array(2);
                ntimes[0] = 0;
                ntimes[1] = 1;
                nvalues[0] = values[values.length - 2];
                nvalues[1] = value;
                maxError = this.maxErr(precision, 0);
                if (Math.abs(nvalues[1] - nvalues[0]) <= maxError) {
                  weight = new Float64Array(1);
                  weight[0] = nvalues[0];
                } else {
                  weight = new Float64Array(2);
                  weight[0] = nvalues[0];
                  weight[1] = nvalues[1] - nvalues[0];
                }
                previousState.setFromKey(PolynomialNode.INTERNAL_TIME_BUFFER, org.mwg.Type.DOUBLE_ARRAY, null);
                previousState.setFromKey(PolynomialNode.INTERNAL_VALUES_BUFFER, org.mwg.Type.DOUBLE_ARRAY, null);
                phasedState.setFromKey(PolynomialNode.INTERNAL_TIME_BUFFER, org.mwg.Type.DOUBLE_ARRAY, ntimes);
                phasedState.setFromKey(PolynomialNode.INTERNAL_VALUES_BUFFER, org.mwg.Type.DOUBLE_ARRAY, nvalues);
                phasedState.setFromKey(PolynomialNode.PRECISION, org.mwg.Type.DOUBLE, precision);
                phasedState.setFromKey(PolynomialNode.MAX_DEGREE, org.mwg.Type.INT, maxd);
                phasedState.setFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY, org.mwg.Type.DOUBLE_ARRAY, weight);
                phasedState.setFromKey(PolynomialNode.INTERNAL_NB_PAST_KEY, org.mwg.Type.INT, 2);
                phasedState.setFromKey(PolynomialNode.INTERNAL_STEP_KEY, org.mwg.Type.LONG, newstep);
                phasedState.setFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY, org.mwg.Type.LONG, newstep);
                if (callback != null) {
                  callback(true);
                }
                return;
              } else {
              }
              if (callback != null) {
                callback(false);
              }
            }
            private static updateBuffer(state: org.mwg.plugin.NodeState, t: number, maxdeg: number, key: string): Float64Array {
              let ts: Float64Array = <Float64Array>state.getFromKey(key);
              if (ts == null) {
                ts = new Float64Array(1);
                ts[0] = t;
                state.setFromKey(key, org.mwg.Type.DOUBLE_ARRAY, ts);
                return ts;
              } else if (ts.length < maxdeg * 4) {
                let nts: Float64Array = new Float64Array(ts.length + 1);
                java.lang.System.arraycopy(ts, 0, nts, 0, ts.length);
                nts[ts.length] = t;
                state.setFromKey(key, org.mwg.Type.DOUBLE_ARRAY, nts);
                return nts;
              } else {
                let nts: Float64Array = new Float64Array(ts.length);
                java.lang.System.arraycopy(ts, 1, nts, 0, ts.length - 1);
                nts[ts.length - 1] = t;
                state.setFromKey(key, org.mwg.Type.DOUBLE_ARRAY, nts);
                return nts;
              }

            }
            public extrapolate(callback: org.mwg.Callback<number>): void {
              let time: number = this.time();
              let state: org.mwg.plugin.NodeState = this.unphasedState();
              let timeOrigin: number = state.time();
              let weight: Float64Array = <Float64Array>state.getFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY);
              if (weight == null) {
                if (callback != null) {
                  callback(0.0);
                }
                return;
              }
              let inferSTEP: number = <number>state.getFromKey(PolynomialNode.INTERNAL_STEP_KEY);
              if (inferSTEP == null || inferSTEP == 0) {
                if (callback != null) {
                  callback(weight[0]);
                }
                return;
              }
              let t: number = (time - timeOrigin);
              let lastTime: number = <number>state.getFromKey(PolynomialNode.INTERNAL_LAST_TIME_KEY);
              if (t > lastTime) {
                t = <number>lastTime;
              }
              t = t / inferSTEP;
              if (callback != null) {
                callback(org.mwg.ml.common.matrix.operation.PolynomialFit.extrapolate(t, weight));
              }
            }
            private maxErr(precision: number, degree: number): number {
              return precision / Math.pow(2, degree + 1);
            }
            private tempError(computedWeights: Float64Array, times: Float64Array, values: Float64Array): number {
              let maxErr: number = 0;
              let temp: number;
              for (let i: number = 0; i < times.length; i++) {
                temp = Math.abs(values[i] - org.mwg.ml.common.matrix.operation.PolynomialFit.extrapolate(times[i], computedWeights));
                if (temp > maxErr) {
                  maxErr = temp;
                }
              }
              return maxErr;
            }
            public toString(): string {
              let builder: java.lang.StringBuilder = new java.lang.StringBuilder();
              builder.append("{\"world\":");
              builder.append(this.world());
              builder.append(",\"time\":");
              builder.append(this.time());
              builder.append(",\"id\":");
              builder.append(this.id());
              let state: org.mwg.plugin.NodeState = this.unphasedState();
              let timeOrigin: number = state.time();
              let step: number = state.getFromKeyWithDefault(PolynomialNode.INTERNAL_STEP_KEY, 1);
              if (state != null) {
                let weight: Float64Array = <Float64Array>state.getFromKey(PolynomialNode.INTERNAL_WEIGHT_KEY);
                if (weight != null) {
                  builder.append("\"polynomial\":\"");
                  for (let i: number = 0; i < weight.length; i++) {
                    if (i != 0) {
                      builder.append("+(");
                    }
                    builder.append(weight[i]);
                    if (i == 1) {
                      if (timeOrigin == 0) {
                        if (step == 1) {
                          builder.append("*t");
                        } else {
                          builder.append("*t/" + step);
                        }
                      } else {
                        if (step == 1) {
                          builder.append("*(t-" + timeOrigin + ")");
                        } else {
                          builder.append("*(t-" + timeOrigin + ")/" + step);
                        }
                      }
                    } else if (i > 1) {
                      if (timeOrigin == 0) {
                        if (step == 1) {
                          builder.append("*t^");
                          builder.append(i);
                        } else {
                          builder.append("*(t/" + step + ")^");
                          builder.append(i);
                        }
                      } else {
                        if (step == 1) {
                          builder.append("*(t-" + timeOrigin + ")^");
                          builder.append(i);
                        } else {
                          builder.append("*((t-" + timeOrigin + ")/" + step + ")^");
                          builder.append(i);
                        }
                      }
                    }

                    if (i != 0) {
                      builder.append(")");
                    }
                  }
                  builder.append("\"");
                }
                builder.append("}");
              }
              return builder.toString();
            }
          }
          export module actions {
            export class ReadContinuous implements org.mwg.task.Action {
              public static NAME: string = "readContinuous";
              private polyTask: org.mwg.task.Task;
              private _relName: string;
              constructor(relName: string) {
                if (relName == null) {
                  throw new Error("name should not be null");
                }
                this._relName = relName;
                this.polyTask = org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.traverse(relName)).then(org.mwg.internal.task.CoreActions.attribute(org.mwg.ml.algorithm.regression.PolynomialNode.VALUE));
              }
              public eval(context: org.mwg.task.TaskContext): void {
                this.polyTask.executeWith(context.graph(), context.result(), (result : org.mwg.task.TaskResult<any>) => {
{
                    context.continueWith(result);
                  }                });
              }
              public serialize(builder: java.lang.StringBuilder): void {
                builder.append(ReadContinuous.NAME);
                builder.append(org.mwg.Constants.TASK_PARAM_OPEN);
                org.mwg.internal.task.TaskHelper.serializeString(this._relName, builder, true);
                builder.append(org.mwg.Constants.TASK_PARAM_CLOSE);
              }
              public toString(): string {
                let res: java.lang.StringBuilder = new java.lang.StringBuilder();
                this.serialize(res);
                return res.toString();
              }
            }
            export class SetContinuous implements org.mwg.task.Action {
              public static NAME: string = "setContinuous";
              private polyTask: org.mwg.task.Task;
              private _relName: string;
              private _value: string;
              constructor(relName: string, c_value: string) {
                if (relName == null || c_value == null) {
                  throw new Error("name or value should not be null");
                }
                this._relName = relName;
                this._value = c_value;
                this.polyTask = org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.defineAsVar("origin")).then(org.mwg.internal.task.CoreActions.traverse(this._relName)).then(org.mwg.internal.task.CoreActions.setAttribute(org.mwg.ml.algorithm.regression.PolynomialNode.VALUE, org.mwg.Type.DOUBLE, this._value)).then(org.mwg.internal.task.CoreActions.readVar("origin"));
              }
              public eval(context: org.mwg.task.TaskContext): void {
                this.polyTask.executeWith(context.graph(), context.result(), (result : org.mwg.task.TaskResult<any>) => {
{
                    context.continueWith(result);
                  }                });
              }
              public serialize(builder: java.lang.StringBuilder): void {
                builder.append(SetContinuous.NAME);
                builder.append(org.mwg.Constants.TASK_PARAM_OPEN);
                org.mwg.internal.task.TaskHelper.serializeString(this._relName, builder, true);
                builder.append(org.mwg.Constants.TASK_PARAM_SEP);
                org.mwg.internal.task.TaskHelper.serializeString(this._value, builder, true);
                builder.append(org.mwg.Constants.TASK_PARAM_CLOSE);
              }
              public toString(): string {
                let res: java.lang.StringBuilder = new java.lang.StringBuilder();
                this.serialize(res);
                return res.toString();
              }
            }
            export class SetPrecision implements org.mwg.task.Action {
              public static NAME: string = "setPrecision";
              private polyTask: org.mwg.task.Task;
              private _relName: string;
              private _value: string;
              constructor(relName: string, c_value: string) {
                if (relName == null || c_value == null) {
                  throw new Error("name or value should not be null");
                }
                this._relName = relName;
                this._value = c_value;
                this.polyTask = org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.defineAsVar("origin")).then(org.mwg.internal.task.CoreActions.traverse(this._relName)).then(org.mwg.internal.task.CoreActions.setAttribute(org.mwg.ml.algorithm.regression.PolynomialNode.PRECISION, org.mwg.Type.DOUBLE, this._value)).then(org.mwg.internal.task.CoreActions.readVar("origin"));
              }
              public eval(context: org.mwg.task.TaskContext): void {
                this.polyTask.executeWith(context.graph(), context.result(), (result : org.mwg.task.TaskResult<any>) => {
{
                    context.continueWith(result);
                  }                });
              }
              public serialize(builder: java.lang.StringBuilder): void {
                builder.append(SetPrecision.NAME);
                builder.append(org.mwg.Constants.TASK_PARAM_OPEN);
                org.mwg.internal.task.TaskHelper.serializeString(this._relName, builder, true);
                builder.append(org.mwg.Constants.TASK_PARAM_SEP);
                org.mwg.internal.task.TaskHelper.serializeString(this._value, builder, true);
                builder.append(org.mwg.Constants.TASK_PARAM_CLOSE);
              }
              public toString(): string {
                let res: java.lang.StringBuilder = new java.lang.StringBuilder();
                this.serialize(res);
                return res.toString();
              }
            }
          }
        }
      }
      export module analytics {
        export class AttributeNode extends org.mwg.ml.BaseMLNode {
          public static NAME: string = "AttributeNode";
          public static VALUE: string = "value";
          public static VALID_VALUE: string = "valid_value";
          public static IS_VALID: string = "is_valid";
          public static EXTRAPOLATE: string = "extrapolate";
          public static MIN_TOLERATED: string = "min_tol";
          public static MAX_TOLERATED: string = "max_tol";
          public static MIN_OBSERVED: string = "min_obs";
          public static MAX_OBSERVED: string = "max_obs";
          public static MIN_VALID: string = "min_val";
          public static MAX_VALID: string = "max_val";
          public static AVG: string = "avg";
          public static SIGMA: string = "sigma";
          public static TOTAL: string = "total";
          public static TOTAL_VALID: string = "total_val";
          private static INTERNAL_SUM_KEY: string = "_sum";
          private static INTERNAL_SUMSQUARE_KEY: string = "_sumSquare";
          private static NOT_MANAGED_ATT_ERROR: string = "Attribute node can only handle value attribute, please use a super node to store other data";
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          public set(propertyName: string, propertyType: number, propertyValue: any): org.mwg.Node {
            if (propertyName === AttributeNode.VALUE) {
              this.internalSet(parseFloat(propertyValue.toString()), null);
            } else if (propertyName === AttributeNode.MIN_TOLERATED) {
              super.set(AttributeNode.MIN_TOLERATED, org.mwg.Type.DOUBLE, propertyValue);
            } else if (propertyName === AttributeNode.MAX_TOLERATED) {
              super.set(AttributeNode.MAX_TOLERATED, org.mwg.Type.DOUBLE, propertyValue);
            } else if (propertyName === AttributeNode.EXTRAPOLATE) {
              super.set(AttributeNode.EXTRAPOLATE, org.mwg.Type.BOOL, propertyValue);
            } else {
              throw new Error(AttributeNode.NOT_MANAGED_ATT_ERROR);
            }



            return this;
          }
          public get(propertyName: string): any {
            switch (propertyName) {
              case AttributeNode.VALUE:
                return this.getValue();
              case AttributeNode.VALID_VALUE:
                return this.getValidValue();
              case AttributeNode.IS_VALID:
                return super.get(propertyName);
              case AttributeNode.AVG:
                return this.getAvg();
              case AttributeNode.SIGMA:
                return this.getSigma();
              default: 

                return super.get(propertyName);
            }
          }
          private getValue(): number {
            return 0.0;
          }
          private getValidValue(): number {
            return 0.0;
          }
          private getAvg(): number {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let totalVal: number = state.getFromKeyWithDefault(AttributeNode.TOTAL_VALID, 0);
            if (totalVal == 0) {
              return null;
            } else {
              let v: number = state.getFromKeyWithDefault(AttributeNode.INTERNAL_SUM_KEY, 0.0);
              return v / totalVal;
            }
          }
          private getSigma(): number {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let totalVal: number = state.getFromKeyWithDefault(AttributeNode.TOTAL_VALID, 0);
            if (totalVal < 1) {
              return null;
            } else {
              let avg: number = state.getFromKeyWithDefault(AttributeNode.INTERNAL_SUM_KEY, 0);
              avg = avg / totalVal;
              let correction: number = totalVal;
              correction = correction / (totalVal - 1);
              let sumsq: number = state.getFromKeyWithDefault(AttributeNode.INTERNAL_SUMSQUARE_KEY, 0.0);
              let cov: number = (sumsq / totalVal - avg * avg) * correction;
              return Math.sqrt(cov);
            }
          }
          private checkValid(v: number, min: number, max: number): boolean {
            let res: boolean = true;
            if (min != null) {
              res = res && (v >= min);
            }
            if (max != null) {
              res = res && (v <= max);
            }
            return res;
          }
          private internalSet(v: number, callback: org.mwg.Callback<boolean>): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            state.setFromKey(AttributeNode.VALUE, org.mwg.Type.DOUBLE, v);
            let total: number = state.getFromKeyWithDefault(AttributeNode.TOTAL, 0);
            if (total == 0) {
              state.setFromKey(AttributeNode.MIN_OBSERVED, org.mwg.Type.DOUBLE, v);
              state.setFromKey(AttributeNode.MAX_OBSERVED, org.mwg.Type.DOUBLE, v);
            }
            state.setFromKey(AttributeNode.TOTAL, org.mwg.Type.LONG, total + 1);
            let mintol: number = state.getFromKeyWithDefault(AttributeNode.MIN_TOLERATED, null);
            let maxtol: number = state.getFromKeyWithDefault(AttributeNode.MAX_TOLERATED, null);
            let minval: number = state.getFromKeyWithDefault(AttributeNode.MIN_VALID, null);
            let maxval: number = state.getFromKeyWithDefault(AttributeNode.MAX_VALID, null);
            let valid: boolean = this.checkValid(v, mintol, maxtol);
            state.setFromKey(AttributeNode.IS_VALID, org.mwg.Type.BOOL, valid);
            if (valid) {
              if (minval == null || v < minval) {
                state.setFromKey(AttributeNode.MIN_VALID, org.mwg.Type.DOUBLE, v);
              }
              if (maxval == null || v > maxval) {
                state.setFromKey(AttributeNode.MAX_VALID, org.mwg.Type.DOUBLE, v);
              }
              let internalSum: number = state.getFromKeyWithDefault(AttributeNode.INTERNAL_SUM_KEY, 0);
              internalSum += v;
              state.setFromKey(AttributeNode.INTERNAL_SUM_KEY, org.mwg.Type.DOUBLE, internalSum);
              let internalSumSq: number = state.getFromKeyWithDefault(AttributeNode.INTERNAL_SUMSQUARE_KEY, 0);
              internalSumSq += v * v;
              state.setFromKey(AttributeNode.INTERNAL_SUMSQUARE_KEY, org.mwg.Type.DOUBLE, internalSumSq);
              let totalval: number = state.getFromKeyWithDefault(AttributeNode.TOTAL_VALID, 0);
              totalval++;
              state.setFromKey(AttributeNode.TOTAL_VALID, org.mwg.Type.LONG, totalval);
            }
            if (callback != null) {
              callback(valid);
            }
          }
        }
      }
      export module common {
        export class DecisionTreeNode {
          public boundary: number;
          public classNum: number;
          public left: org.mwg.ml.common.DecisionTreeNode;
          public right: org.mwg.ml.common.DecisionTreeNode;
          public featureNum: number;
          private static ELEMENTS_PER_NODE: number = 5;
          public static deserializeFromDoubleArray(decisionTreeArray: Float64Array): org.mwg.ml.common.DecisionTreeNode {
            return org.mwg.ml.common.DecisionTreeNode.deserializeFromDoubleArrayWithRoot(decisionTreeArray, 0);
          }
          public static deserializeFromDoubleArrayWithRoot(decisionTreeArray: Float64Array, rootIndex: number): org.mwg.ml.common.DecisionTreeNode {
            let dtr: org.mwg.ml.common.DecisionTreeNode = new org.mwg.ml.common.DecisionTreeNode();
            dtr.boundary = decisionTreeArray[rootIndex * DecisionTreeNode.ELEMENTS_PER_NODE];
            dtr.classNum = <number>decisionTreeArray[rootIndex * DecisionTreeNode.ELEMENTS_PER_NODE + 1];
            let leftNum: number = <number>decisionTreeArray[rootIndex * DecisionTreeNode.ELEMENTS_PER_NODE + 2];
            dtr.left = (leftNum > 0) ? org.mwg.ml.common.DecisionTreeNode.deserializeFromDoubleArrayWithRoot(decisionTreeArray, leftNum) : null;
            let rightNum: number = <number>decisionTreeArray[rootIndex * DecisionTreeNode.ELEMENTS_PER_NODE + 3];
            dtr.right = (rightNum > 0) ? org.mwg.ml.common.DecisionTreeNode.deserializeFromDoubleArrayWithRoot(decisionTreeArray, rightNum) : null;
            dtr.featureNum = <number>decisionTreeArray[rootIndex * DecisionTreeNode.ELEMENTS_PER_NODE + 4];
            return dtr;
          }
          public serializeToDoubleArray(): Float64Array {
            let resultObjects: java.util.List<number> = this.serializeToDoubleObjectArray();
            let result: Float64Array = new Float64Array(resultObjects.size());
            for (let i: number = 0; i < result.length; i++) {
              result[i] = resultObjects.get(i);
            }
            return result;
          }
          private serializeToDoubleObjectArray(): java.util.List<number> {
            let result: java.util.List<number> = new java.util.ArrayList<number>();
            let frontier: java.util.List<org.mwg.ml.common.DecisionTreeNode> = new java.util.ArrayList<org.mwg.ml.common.DecisionTreeNode>();
            frontier.add(this);
            let currentNumberToAdd: number = 0;
            while (frontier.size() > 0) {
              let currentNode: org.mwg.ml.common.DecisionTreeNode = frontier.get(0);
              frontier.remove(0);
              result.add(currentNode.boundary);
              result.add(<number>currentNode.classNum);
              if (currentNode.left != null) {
                currentNumberToAdd++;
                result.add(<number>currentNumberToAdd);
                frontier.add(currentNode.left);
              } else {
                result.add(-1.0);
              }
              if (currentNode.right != null) {
                currentNumberToAdd++;
                result.add(<number>currentNumberToAdd);
                frontier.add(currentNode.right);
              } else {
                result.add(-1.0);
              }
              result.add(<number>currentNode.featureNum);
            }
            return result;
          }
        }
        export module distance {
          export class GaussianDistance implements org.mwg.structure.distance.Distance {
            public err: Float64Array;
            constructor(covariance: Float64Array) {
              this.err = covariance;
            }
            public measure(x: Float64Array, y: Float64Array): number {
              let max: number = 0;
              let temp: number;
              for (let i: number = 0; i < x.length; i++) {
                temp = (x[i] - y[i]) * (x[i] - y[i]) / this.err[i];
                if (temp > max) {
                  max = temp;
                }
              }
              return Math.sqrt(max);
            }
            public compare(x: number, y: number): boolean {
              return x < y;
            }
            public getMinValue(): number {
              return 0;
            }
            public getMaxValue(): number {
              return java.lang.Double.MAX_VALUE;
            }
          }
        }
        export module matrix {
          export class HybridMatrixEngine implements org.mwg.ml.common.matrix.MatrixEngine {
            private blas: org.mwg.ml.common.matrix.MatrixEngine;
            private jama: org.mwg.ml.common.matrix.MatrixEngine;
            private static MULT_LIMIT: number=12;
            private static INVERT_LIMIT: number=6;
            private static PINV_LIMIT: number=5;
            private static SOLVELU_LIMIT: number=6;
            private static SOLVEQR_LIMIT: number=10;
            private static SOLVESVD_LIMIT: number=8;
            constructor() {
              this.blas = new org.mwg.ml.common.matrix.blassolver.BlasMatrixEngine();
              this.jama = new org.mwg.ml.common.matrix.jamasolver.JamaMatrixEngine();
            }
            public multiplyTransposeAlphaBeta(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix, beta: number, matC: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              if (matA.leadingDimension() < HybridMatrixEngine.MULT_LIMIT && matB.leadingDimension() < HybridMatrixEngine.MULT_LIMIT) {
                return this.jama.multiplyTransposeAlphaBeta(transA, alpha, matA, transB, matB, beta, matC);
              } else {
                return this.blas.multiplyTransposeAlphaBeta(transA, alpha, matA, transB, matB, beta, matC);
              }
            }
            public invert(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
              if (mat.rows() < HybridMatrixEngine.INVERT_LIMIT) {
                return this.jama.invert(mat, invertInPlace);
              } else {
                return this.blas.invert(mat, invertInPlace);
              }
            }
            public pinv(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
              let res: org.mwg.ml.common.matrix.operation.PInvSVD = new org.mwg.ml.common.matrix.operation.PInvSVD();
              res.factor(mat, invertInPlace);
              return res.getPInv();
            }
            public solveLU(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
              if (matA.leadingDimension() < HybridMatrixEngine.SOLVELU_LIMIT && matB.leadingDimension() < HybridMatrixEngine.SOLVELU_LIMIT) {
                return this.jama.solveLU(matA, matB, workInPlace, transB);
              } else {
                return this.blas.solveLU(matA, matB, workInPlace, transB);
              }
            }
            public solveQR(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
              return this.blas.solveQR(matA, matB, workInPlace, transB);
            }
            public decomposeSVD(matA: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.SVDDecompose {
              return this.jama.decomposeSVD(matA, workInPlace);
            }
            public solve(A: org.mwg.struct.DMatrix, B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              return this.blas.solve(A, B);
            }
          }
          export interface MatrixEngine {
            multiplyTransposeAlphaBeta(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix, beta: number, matC: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix;
            invert(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix;
            pinv(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix;
            solveLU(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix;
            solveQR(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix;
            decomposeSVD(matA: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.SVDDecompose;
            solve(A: org.mwg.struct.DMatrix, B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix;
          }
          export class MatrixOps {
            private static _defaultEngine: org.mwg.ml.common.matrix.MatrixEngine = null;
            public static defaultEngine(): org.mwg.ml.common.matrix.MatrixEngine {
              if (MatrixOps._defaultEngine == null) {
                MatrixOps._defaultEngine = new org.mwg.ml.common.matrix.HybridMatrixEngine();
              }
              return MatrixOps._defaultEngine;
            }
            public static setDefaultEngine(engine: org.mwg.ml.common.matrix.MatrixEngine): void {
              MatrixOps._defaultEngine = engine;
            }
            public static copyMatrix(source: org.mwg.struct.DMatrix, destination: org.mwg.struct.DMatrix): void {
              for (let row: number = 0; row < source.rows(); row++) {
                for (let col: number = 0; col < source.columns(); col++) {
                  destination.set(row, col, source.get(row, col));
                }
              }
            }
            public static multiply(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().multiplyTransposeAlphaBeta(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, 1, matA, org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, matB, 0, null);
            }
            public static multiplyTranspose(transA: org.mwg.ml.common.matrix.TransposeType, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().multiplyTransposeAlphaBeta(transA, 1.0, matA, transB, matB, 0, null);
            }
            public static multiplyTransposeAlpha(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().multiplyTransposeAlphaBeta(transA, alpha, matA, transB, matB, 0, null);
            }
            public static multiplyTransposeAlphaBeta(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix, beta: number, matC: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().multiplyTransposeAlphaBeta(transA, alpha, matA, transB, matB, beta, matC);
            }
            public static invert(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().invert(mat, invertInPlace);
            }
            public static pinv(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
              return org.mwg.ml.common.matrix.MatrixOps.defaultEngine().pinv(mat, invertInPlace);
            }
            public static scale(alpha: number, matA: org.mwg.ml.common.matrix.VolatileDMatrix): void {
              if (alpha == 0) {
                matA.fill(0);
                return;
              }
              for (let i: number = 0; i < matA.rows() * matA.columns(); i++) {
                matA.unsafeSet(i, alpha * matA.unsafeGet(i));
              }
            }
            public static transpose(matA: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matA.columns(), matA.rows());
              let TRANSPOSE_SWITCH: number = 375;
              if (matA.columns() == matA.rows()) {
                org.mwg.ml.common.matrix.MatrixOps.transposeSquare(matA, result);
              } else if (matA.columns() > TRANSPOSE_SWITCH && matA.rows() > TRANSPOSE_SWITCH) {
                org.mwg.ml.common.matrix.MatrixOps.transposeBlock(matA, result);
              } else {
                org.mwg.ml.common.matrix.MatrixOps.transposeStandard(matA, result);
              }

              return result;
            }
            private static transposeSquare(matA: org.mwg.struct.DMatrix, result: org.mwg.struct.DMatrix): void {
              let index: number = 1;
              let indexEnd: number = matA.columns();
              for (let i: number = 0; i < matA.rows(); i++) {
                let indexOther: number = (i + 1) * matA.columns() + i;
                let n: number = i * (matA.columns() + 1);
                result.unsafeSet(n, matA.unsafeGet(n));
                for (; index < indexEnd; index++) {
                  result.unsafeSet(index, matA.unsafeGet(indexOther));
                  result.unsafeSet(indexOther, matA.unsafeGet(index));
                  indexOther += matA.columns();
                }
                index += i + 2;
                indexEnd += matA.columns();
              }
            }
            private static transposeStandard(matA: org.mwg.struct.DMatrix, result: org.mwg.struct.DMatrix): void {
              let index: number = 0;
              for (let i: number = 0; i < result.columns(); i++) {
                let index2: number = i;
                let end: number = index + result.rows();
                while (index < end) {
                  result.unsafeSet(index++, matA.unsafeGet(index2));
                  index2 += matA.rows();
                }
              }
            }
            private static transposeBlock(matA: org.mwg.struct.DMatrix, result: org.mwg.struct.DMatrix): void {
              let BLOCK_WIDTH: number = 60;
              for (let j: number = 0; j < matA.columns(); j += BLOCK_WIDTH) {
                let blockWidth: number = Math.min(BLOCK_WIDTH, matA.columns() - j);
                let indexSrc: number = j * matA.rows();
                let indexDst: number = j;
                for (let i: number = 0; i < matA.rows(); i += BLOCK_WIDTH) {
                  let blockHeight: number = Math.min(BLOCK_WIDTH, matA.rows() - i);
                  let indexSrcEnd: number = indexSrc + blockHeight;
                  for (; indexSrc < indexSrcEnd; indexSrc++) {
                    let colSrc: number = indexSrc;
                    let colDst: number = indexDst;
                    let end: number = colDst + blockWidth;
                    for (; colDst < end; colDst++) {
                      result.unsafeSet(colDst, matA.unsafeGet(colSrc));
                      colSrc += matA.rows();
                    }
                    indexDst += result.rows();
                  }
                }
              }
            }
            public static testDimensionsAB(transA: org.mwg.ml.common.matrix.TransposeType, transB: org.mwg.ml.common.matrix.TransposeType, matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix): boolean {
              if (transA.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                  return (matA.columns() == matB.rows());
                } else {
                  return (matA.columns() == matB.columns());
                }
              } else {
                if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                  return (matA.rows() == matB.rows());
                } else {
                  return (matA.rows() == matB.columns());
                }
              }
            }
            public static sub(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              if (matA.rows() != matB.rows() || matA.columns() != matB.columns()) {
                throw new Error("Matrices A and B have different dimensions for the substract operation");
              }
              let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matA.rows(), matA.columns());
              let total: number = matA.rows() * matA.columns();
              for (let i: number = 0; i < total; i++) {
                result.unsafeSet(i, matA.unsafeGet(i) - matB.unsafeGet(i));
              }
              return result;
            }
            public static add(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              if (matA.rows() != matB.rows() || matA.columns() != matB.columns()) {
                throw new Error("Matrices A and B have different dimensions for the add operation");
              }
              let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matA.rows(), matA.columns());
              let total: number = matA.rows() * matA.columns();
              for (let i: number = 0; i < total; i++) {
                result.unsafeSet(i, matA.unsafeGet(i) + matB.unsafeGet(i));
              }
              return result;
            }
          }
          export interface SVDDecompose {
            factor(A: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.SVDDecompose;
            getU(): org.mwg.struct.DMatrix;
            getVt(): org.mwg.struct.DMatrix;
            getS(): Float64Array;
            getSMatrix(): org.mwg.struct.DMatrix;
          }
          export class TransposeType {
            public static NOTRANSPOSE: TransposeType = new TransposeType();
            public static TRANSPOSE: TransposeType = new TransposeType();
            public equals(other: any): boolean {
              return this == other;
            }
            public static _TransposeTypeVALUES : TransposeType[] = [
              TransposeType.NOTRANSPOSE
              ,TransposeType.TRANSPOSE
            ];
            public static values():TransposeType[]{
              return TransposeType._TransposeTypeVALUES;
            }
          }
          export class VolatileDMatrix implements org.mwg.struct.DMatrix {
            private _data: Float64Array;
            private _nbRows: number;
            private _nbColumns: number;
            private _nbMaxColumn: number;
            constructor(backend: Float64Array, p_nbRows: number, p_nbColumns: number) {
              this._nbRows = p_nbRows;
              this._nbColumns = p_nbColumns;
              if (backend != null) {
                this._data = backend;
              } else {
                this._data = new Float64Array(this._nbRows * this._nbColumns);
              }
            }
            public init(rows: number, columns: number): org.mwg.struct.DMatrix {
              if (rows != this._nbRows && columns != this._nbColumns) {
                throw new Error("Bad API usage !");
              }
              this._nbMaxColumn = columns;
              this._data = new Float64Array(this._nbRows * this._nbColumns);
              return this;
            }
            public data(): Float64Array {
              return this._data;
            }
            public rows(): number {
              return this._nbRows;
            }
            public columns(): number {
              return this._nbColumns;
            }
            public column(index: number): Float64Array {
              let result: Float64Array = new Float64Array(this._nbRows);
              java.lang.System.arraycopy(this._data, (index * this._nbRows), result, 0, this._nbRows);
              return result;
            }
            public get(rowIndex: number, columnIndex: number): number {
              return this._data[rowIndex + columnIndex * this._nbRows];
            }
            public set(rowIndex: number, columnIndex: number, value: number): org.mwg.struct.DMatrix {
              this._data[rowIndex + columnIndex * this._nbRows] = value;
              return this;
            }
            public add(rowIndex: number, columnIndex: number, value: number): org.mwg.struct.DMatrix {
              let raw_index: number = rowIndex + columnIndex * this._nbRows;
              this._data[raw_index] = value + this._data[raw_index];
              return this;
            }
            public appendColumn(newColumn: Float64Array): org.mwg.struct.DMatrix {
              if (this._data == null) {
                this._nbRows = newColumn.length;
                this._nbColumns = org.mwg.Constants.MAP_INITIAL_CAPACITY;
                this._nbMaxColumn = 0;
                this._data = new Float64Array(this._nbRows * this._nbColumns);
              }
              if (this._nbMaxColumn == this._nbColumns) {
                this._nbColumns = this._nbColumns * 2;
                let newLength: number = this._nbColumns * this._nbRows;
                let next_backend: Float64Array = new Float64Array(newLength);
                java.lang.System.arraycopy(this._data, 0, next_backend, 0, this._data.length);
                this._data = next_backend;
              }
              java.lang.System.arraycopy(newColumn, 0, this._data, this._nbColumns * this._nbRows, newColumn.length);
              this._nbMaxColumn = this._nbMaxColumn + 1;
              return null;
            }
            public fill(value: number): org.mwg.struct.DMatrix {
              for (let i: number = 0; i < this._nbColumns * this._nbRows; i++) {
                this._data[i] = value;
              }
              return this;
            }
            public fillWith(values: Float64Array): org.mwg.struct.DMatrix {
              this._data = values;
              return this;
            }
            public fillWithRandom(random: java.util.Random, min: number, max: number): org.mwg.struct.DMatrix {
              for (let i: number = 0; i < this._nbRows * this._nbColumns; i++) {
                this._data[i] = random.nextDouble() * (max - min) + min;
              }
              return this;
            }
            public fillWithRandomStd(random: java.util.Random, std: number): org.mwg.struct.DMatrix {
              for (let i: number = 0; i < this._nbRows * this._nbColumns; i++) {
                this._data[i] = random.nextGaussian() * std;
              }
              return this;
            }
            public leadingDimension(): number {
              return Math.max(this._nbColumns, this._nbRows);
            }
            public unsafeGet(index: number): number {
              return this._data[index];
            }
            public unsafeSet(index: number, value: number): org.mwg.struct.DMatrix {
              this._data[index] = value;
              return this;
            }
            public static compare(a: Float64Array, b: Float64Array, eps: number): boolean {
              if (a == null || b == null) {
                return false;
              }
              for (let i: number = 0; i < a.length; i++) {
                if (Math.abs(a[i] - b[i]) > eps) {
                  return false;
                }
              }
              return true;
            }
            public static compareArray(a: Array<Float64Array>, b: Array<Float64Array>, eps: number): boolean {
              if (a == null || b == null) {
                return false;
              }
              for (let i: number = 0; i < a.length; i++) {
                if (!org.mwg.ml.common.matrix.VolatileDMatrix.compare(a[i], b[i], eps)) {
                  return false;
                }
              }
              return true;
            }
            public exportRowMatrix(): Float64Array {
              let res: Float64Array = new Float64Array(this._data.length);
              let k: number = 0;
              for (let i: number = 0; i < this._nbRows; i++) {
                for (let j: number = 0; j < this._nbColumns; j++) {
                  res[k] = this.get(i, j);
                  k++;
                }
              }
              return res;
            }
            public importRowMatrix(rowdata: Float64Array, rows: number, columns: number): org.mwg.ml.common.matrix.VolatileDMatrix {
              let res: org.mwg.ml.common.matrix.VolatileDMatrix = new org.mwg.ml.common.matrix.VolatileDMatrix(null, rows, columns);
              let k: number = 0;
              for (let i: number = 0; i < this._nbRows; i++) {
                for (let j: number = 0; j < this._nbColumns; j++) {
                  res.set(i, j, rowdata[k]);
                  k++;
                }
              }
              return res;
            }
            public addAtIndex(index: number, value: number): number {
              this._data[index] += value;
              return this._data[index];
            }
            public static random(rows: number, columns: number, min: number, max: number): org.mwg.struct.DMatrix {
              let res: org.mwg.ml.common.matrix.VolatileDMatrix = new org.mwg.ml.common.matrix.VolatileDMatrix(null, rows, columns);
              let rand: java.util.Random = new java.util.Random();
              for (let i: number = 0; i < rows * columns; i++) {
                res.unsafeSet(i, rand.nextDouble() * (max - min) + min);
              }
              return res;
            }
            public static compareMatrix(matA: org.mwg.ml.common.matrix.VolatileDMatrix, matB: org.mwg.ml.common.matrix.VolatileDMatrix): number {
              let err: number = 0;
              for (let i: number = 0; i < matA.rows(); i++) {
                for (let j: number = 0; j < matA.columns(); j++) {
                  if (err < Math.abs(matA.get(i, j) - matB.get(i, j))) {
                    err = Math.abs(matA.get(i, j) - matB.get(i, j));
                  }
                }
              }
              return err;
            }
            public static identity(rows: number, columns: number): org.mwg.struct.DMatrix {
              let res: org.mwg.struct.DMatrix = new org.mwg.ml.common.matrix.VolatileDMatrix(null, rows, columns);
              for (let i: number = 0; i < Math.max(rows, columns); i++) {
                res.set(i, i, 1.0);
              }
              return res;
            }
            public static empty(rows: number, columns: number): org.mwg.struct.DMatrix {
              return new org.mwg.ml.common.matrix.VolatileDMatrix(null, rows, columns);
            }
            public static wrap(data: Float64Array, rows: number, columns: number): org.mwg.struct.DMatrix {
              return new org.mwg.ml.common.matrix.VolatileDMatrix(data, rows, columns);
            }
            public static cloneFrom(origin: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
              let prev: Float64Array = origin.data();
              let copy: Float64Array = new Float64Array(prev.length);
              java.lang.System.arraycopy(prev, 0, copy, 0, copy.length);
              return new org.mwg.ml.common.matrix.VolatileDMatrix(copy, origin.rows(), origin.columns());
            }
          }
          export module blassolver {
            export class BlasMatrixEngine implements org.mwg.ml.common.matrix.MatrixEngine {
              private _blas: org.mwg.ml.common.matrix.blassolver.blas.Blas;
              constructor() {
                this._blas = new org.mwg.ml.common.matrix.blassolver.blas.JSBlas();
              }
              public setBlas(p_blas: org.mwg.ml.common.matrix.blassolver.blas.Blas): void {
                this._blas = p_blas;
              }
              public getBlas(): org.mwg.ml.common.matrix.blassolver.blas.Blas {
                return this._blas;
              }
              public multiplyTransposeAlphaBeta(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix, beta: number, matC: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                if (org.mwg.ml.common.matrix.MatrixOps.testDimensionsAB(transA, transB, matA, matB)) {
                  let k: number = 0;
                  let dimC: Int32Array = new Int32Array(2);
                  if (transA.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                    k = matA.columns();
                    if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                      dimC[0] = matA.rows();
                      dimC[1] = matB.columns();
                    } else {
                      dimC[0] = matA.rows();
                      dimC[1] = matB.rows();
                    }
                  } else {
                    k = matA.rows();
                    if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                      dimC[0] = matA.columns();
                      dimC[1] = matB.columns();
                    } else {
                      dimC[0] = matA.columns();
                      dimC[1] = matB.rows();
                    }
                  }
                  if (beta == 0 || matC == null) {
                    matC = org.mwg.ml.common.matrix.VolatileDMatrix.empty(dimC[0], dimC[1]);
                  }
                  this._blas.dgemm(transA, transB, matC.rows(), matC.columns(), k, alpha, matA.data(), 0, matA.rows(), matB.data(), 0, matB.rows(), beta, matC.data(), 0, matC.rows());
                  return matC;
                } else {
                  throw new Error("Dimensions mismatch between A,B and C");
                }
              }
              public invert(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
                if (mat.rows() != mat.columns()) {
                  return null;
                }
                if (invertInPlace) {
                  let dlu: org.mwg.ml.common.matrix.blassolver.LU = new org.mwg.ml.common.matrix.blassolver.LU(mat.rows(), mat.columns(), this._blas);
                  if (dlu.invert(mat)) {
                    return mat;
                  } else {
                    return null;
                  }
                } else {
                  let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(mat.rows(), mat.columns());
                  let A_temp: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(mat.rows(), mat.columns());
                  java.lang.System.arraycopy(mat.data(), 0, A_temp.data(), 0, mat.columns() * mat.rows());
                  let dlu: org.mwg.ml.common.matrix.blassolver.LU = new org.mwg.ml.common.matrix.blassolver.LU(A_temp.rows(), A_temp.columns(), this._blas);
                  if (dlu.invert(A_temp)) {
                    result.fillWith(A_temp.data());
                    return result;
                  } else {
                    return null;
                  }
                }
              }
              public pinv(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
                return this.solve(mat, org.mwg.ml.common.matrix.VolatileDMatrix.identity(mat.rows(), mat.rows()));
              }
              public solveQR(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
                let solver: org.mwg.ml.common.matrix.blassolver.QR = org.mwg.ml.common.matrix.blassolver.QR.factorize(matA, workInPlace, this._blas);
                let coef: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matA.columns(), matB.columns());
                if (transB != org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE) {
                  matB = org.mwg.ml.common.matrix.MatrixOps.transpose(matB);
                }
                solver.solve(matB, coef);
                return coef;
              }
              public decomposeSVD(matA: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.SVDDecompose {
                let svd: org.mwg.ml.common.matrix.blassolver.SVD = new org.mwg.ml.common.matrix.blassolver.SVD(matA.rows(), matA.columns(), this._blas);
                svd.factor(matA, workInPlace);
                return svd;
              }
              public solveLU(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
                if (!workInPlace) {
                  let A_temp: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matA.rows(), matA.columns());
                  java.lang.System.arraycopy(matA.data(), 0, A_temp.data(), 0, matA.columns() * matA.rows());
                  let dlu: org.mwg.ml.common.matrix.blassolver.LU = new org.mwg.ml.common.matrix.blassolver.LU(A_temp.rows(), A_temp.columns(), this._blas);
                  dlu.factor(A_temp, true);
                  if (dlu.isSingular()) {
                    return null;
                  }
                  let B_temp: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(matB.rows(), matB.columns());
                  java.lang.System.arraycopy(matB.data(), 0, B_temp.data(), 0, matB.columns() * matB.rows());
                  dlu.transSolve(B_temp, transB);
                  return B_temp;
                } else {
                  let dlu: org.mwg.ml.common.matrix.blassolver.LU = new org.mwg.ml.common.matrix.blassolver.LU(matA.rows(), matA.columns(), this._blas);
                  dlu.factor(matA, true);
                  if (dlu.isSingular()) {
                    return null;
                  }
                  dlu.transSolve(matB, transB);
                  return matB;
                }
              }
              public solve(A: org.mwg.struct.DMatrix, B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                return (A.rows() == A.columns() ? (new org.mwg.ml.common.matrix.blassolver.LU(A.rows(), A.columns(), this._blas).factor(A, false)).solve(B) : this.solveQR(A, B, false, org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE));
              }
            }
            export class JobSVD {
              public static All: JobSVD = new JobSVD();
              public static None: JobSVD = new JobSVD();
              public static Overwrite: JobSVD = new JobSVD();
              public static Part: JobSVD = new JobSVD();
              public netlib(): string {
                switch (this) {
                  case JobSVD.All:
                    return "A";
                  case JobSVD.Part:
                    return "S";
                  case JobSVD.Overwrite:
                    return "O";
                  default: 

                    return "N";
                }
              }
              public equals(other: any): boolean {
                return this == other;
              }
              public static _JobSVDVALUES : JobSVD[] = [
                JobSVD.All
                ,JobSVD.None
                ,JobSVD.Overwrite
                ,JobSVD.Part
              ];
              public static values():JobSVD[]{
                return JobSVD._JobSVDVALUES;
              }
            }
            export class LU {
              private LU: org.mwg.struct.DMatrix;
              private _blas: org.mwg.ml.common.matrix.blassolver.blas.Blas;
              private piv: Int32Array;
              private singular: boolean;
              public getLU(): org.mwg.struct.DMatrix {
                return this.LU;
              }
              constructor(m: number, n: number, blas: org.mwg.ml.common.matrix.blassolver.blas.Blas) {
                this._blas = blas;
                this.LU = org.mwg.ml.common.matrix.VolatileDMatrix.empty(m, n);
                this.piv = new Int32Array(Math.min(m, n));
              }
              public static factorize(A: org.mwg.struct.DMatrix, blas: org.mwg.ml.common.matrix.blassolver.blas.Blas): org.mwg.ml.common.matrix.blassolver.LU {
                return new org.mwg.ml.common.matrix.blassolver.LU(A.rows(), A.columns(), blas).factor(A, false);
              }
              public factor(A: org.mwg.struct.DMatrix, factorInPlace: boolean): org.mwg.ml.common.matrix.blassolver.LU {
                if (factorInPlace) {
                  this.singular = false;
                  let info: Int32Array = new Int32Array(1);
                  info[0] = 0;
                  this._blas.dgetrf(A.rows(), A.columns(), A.data(), 0, A.rows(), this.piv, 0, info);
                  if (info[0] > 0)
                    this.singular = true;
                  else if (info[0] < 0)
                    throw new Error();


                  this.LU.fillWith(A.data());
                  return this;
                } else {
                  this.singular = false;
                  let B: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(A);
                  let info: Int32Array = new Int32Array(1);
                  info[0] = 0;
                  this._blas.dgetrf(B.rows(), B.columns(), B.data(), 0, B.rows(), this.piv, 0, info);
                  if (info[0] > 0)
                    this.singular = true;
                  else if (info[0] < 0)
                    throw new Error();


                  this.LU.fillWith(B.data());
                  return this;
                }
              }
              public getL(): org.mwg.struct.DMatrix {
                let numRows: number = this.LU.rows();
                let numCols: number = this.LU.rows() < this.LU.columns() ? this.LU.rows() : this.LU.columns();
                let lower: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(numRows, numCols);
                for (let i: number = 0; i < numCols; i++) {
                  lower.set(i, i, 1.0);
                  for (let j: number = 0; j < i; j++) {
                    lower.set(i, j, this.LU.get(i, j));
                  }
                }
                if (numRows > numCols) {
                  for (let i: number = numCols; i < numRows; i++) {
                    for (let j: number = 0; j < numCols; j++) {
                      lower.set(i, j, this.LU.get(i, j));
                    }
                  }
                }
                return lower;
              }
              public getU(): org.mwg.struct.DMatrix {
                let numRows: number = this.LU.rows() < this.LU.columns() ? this.LU.rows() : this.LU.columns();
                let numCols: number = this.LU.columns();
                let upper: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(numRows, numCols);
                for (let i: number = 0; i < numRows; i++) {
                  for (let j: number = i; j < numCols; j++) {
                    upper.set(i, j, this.LU.get(i, j));
                  }
                }
                return upper;
              }
              public getPivots(): Int32Array {
                return this.piv;
              }
              public isSingular(): boolean {
                return this.singular;
              }
              public solve(B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                return this.transSolve(B, org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE);
              }
              public transSolve(B: org.mwg.struct.DMatrix, trans: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
                if (B.rows() != this.LU.rows())
                  throw new Error("B.numRows() != LU.numRows()");

                let info: Int32Array = new Int32Array(1);
                this._blas.dgetrs(trans, this.LU.rows(), B.columns(), this.LU.data(), 0, this.LU.rows(), this.piv, 0, B.data(), 0, B.rows(), info);
                if (info[0] < 0)
                  throw new Error();

                return B;
              }
              public invert(A: org.mwg.struct.DMatrix): boolean {
                let info: Int32Array = new Int32Array(1);
                info[0] = 0;
                this._blas.dgetrf(A.rows(), A.columns(), A.data(), 0, A.rows(), this.piv, 0, info);
                if (info[0] > 0) {
                  this.singular = true;
                  return false;
                } else if (info[0] < 0)
                  throw new Error();


                let lwork: number = A.rows() * A.rows();
                let work: Float64Array = new Float64Array(lwork);
                for (let i: number = 0; i < lwork; i++) {
                  work[i] = 0;
                }
                this._blas.dgetri(A.rows(), A.data(), 0, A.rows(), this.piv, 0, work, 0, lwork, info);
                if (info[0] != 0) {
                  return false;
                } else {
                  return true;
                }
              }
            }
            export class QR {
              private Q: org.mwg.struct.DMatrix;
              private R: org.mwg.struct.DMatrix;
              private _blas: org.mwg.ml.common.matrix.blassolver.blas.Blas;
              public m: number;
              public n: number;
              public k: number;
              public work: Float64Array;
              public workGen: Float64Array;
              public tau: Float64Array;
              constructor(rows: number, columns: number, blas: org.mwg.ml.common.matrix.blassolver.blas.Blas) {
                this._blas = blas;
                if (columns > rows)
                  throw new Error("n > m");

                this.m = rows;
                this.n = columns;
                this.k = Math.min(this.m, this.n);
                this.tau = new Float64Array(this.k);
                this.R = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.n, this.n);
              }
              public static factorize(A: org.mwg.struct.DMatrix, workInPlace: boolean, blas: org.mwg.ml.common.matrix.blassolver.blas.Blas): org.mwg.ml.common.matrix.blassolver.QR {
                return new org.mwg.ml.common.matrix.blassolver.QR(A.rows(), A.columns(), blas).factor(A, workInPlace);
              }
              public factor(matA: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.blassolver.QR {
                let A: org.mwg.struct.DMatrix;
                if (!workInPlace) {
                  A = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(matA);
                } else {
                  A = matA;
                }
                let lwork: number;
                this.work = new Float64Array(1);
                let info: Int32Array = new Int32Array(1);
                info[0] = 0;
                this._blas.dgeqrf(this.m, this.n, new Float64Array(0), 0, this.m, new Float64Array(0), 0, this.work, 0, -1, info);
                if (info[0] != 0)
                  lwork = this.n;
                else
                  lwork = <number>this.work[0];

                lwork = Math.max(1, lwork);
                this.work = new Float64Array(lwork);
                this.workGen = new Float64Array(1);
                info[0] = 0;
                this._blas.dorgqr(this.m, this.n, this.k, new Float64Array(0), 0, this.m, new Float64Array(0), 0, this.workGen, 0, -1, info);
                if (info[0] != 0)
                  lwork = this.n;
                else
                  lwork = <number>this.workGen[0];

                lwork = Math.max(1, lwork);
                this.workGen = new Float64Array(lwork);
                info[0] = 0;
                this._blas.dgeqrf(this.m, this.n, A.data(), 0, this.m, this.tau, 0, this.work, 0, this.work.length, info);
                if (info[0] < 0)
                  throw new Error("" + info[0]);

                for (let col: number = 0; col < A.columns(); col++) {
                  for (let row: number = 0; row <= col; row++) {
                    this.R.set(row, col, A.get(row, col));
                  }
                }
                info[0] = 0;
                this._blas.dorgqr(this.m, this.n, this.k, A.data(), 0, this.m, this.tau, 0, this.workGen, 0, this.workGen.length, info);
                if (info[0] < 0)
                  throw new Error();

                this.Q = A;
                return this;
              }
              public solve(B: org.mwg.struct.DMatrix, X: org.mwg.struct.DMatrix): void {
                let BnumCols: number = B.columns();
                let Y: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, 1);
                let Z: org.mwg.struct.DMatrix;
                for (let colB: number = 0; colB < BnumCols; colB++) {
                  for (let i: number = 0; i < this.m; i++) {
                    Y.unsafeSet(i, B.get(i, colB));
                  }
                  Z = org.mwg.ml.common.matrix.MatrixOps.multiplyTranspose(org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, this.Q, org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, Y);
                  this.solveU(this.R, Z.data(), this.n, this.m);
                  for (let i: number = 0; i < this.n; i++) {
                    X.set(i, colB, Z.unsafeGet(i));
                  }
                }
              }
              private solveU(U: org.mwg.struct.DMatrix, b: Float64Array, n: number, m: number): void {
                for (let i: number = n - 1; i >= 0; i--) {
                  let sum: number = b[i];
                  for (let j: number = i + 1; j < n; j++) {
                    sum -= U.get(i, j) * b[j];
                  }
                  b[i] = sum / U.get(i, i);
                }
              }
              public getR(): org.mwg.struct.DMatrix {
                return this.R;
              }
              public getQ(): org.mwg.struct.DMatrix {
                return this.Q;
              }
            }
            export class SVD implements org.mwg.ml.common.matrix.SVDDecompose {
              private work: Float64Array;
              private iwork: Int32Array;
              private m: number;
              private n: number;
              private vectors: boolean;
              private job: org.mwg.ml.common.matrix.blassolver.JobSVD;
              private S: Float64Array;
              private U: org.mwg.struct.DMatrix;
              private Vt: org.mwg.struct.DMatrix;
              private _blas: org.mwg.ml.common.matrix.blassolver.blas.Blas;
              constructor(m: number, n: number, blas: org.mwg.ml.common.matrix.blassolver.blas.Blas) {
                this.m = m;
                this.n = n;
                this._blas = blas;
                this.vectors = true;
                this.S = new Float64Array(Math.min(m, n));
                if (this.vectors) {
                  this.U = org.mwg.ml.common.matrix.VolatileDMatrix.empty(m, m);
                  this.Vt = org.mwg.ml.common.matrix.VolatileDMatrix.empty(n, n);
                } else
                  this.U = this.Vt = null;

                this.job = this.vectors ? org.mwg.ml.common.matrix.blassolver.JobSVD.All : org.mwg.ml.common.matrix.blassolver.JobSVD.None;
                this.iwork = new Int32Array(8 * Math.min(m, n));
                let worksize: Float64Array = new Float64Array(1);
                let info: Int32Array = new Int32Array(1);
                this._blas.dgesdd(this.job.netlib(), m, n, new Float64Array(0), Math.max(1, m), new Float64Array(0), new Float64Array(0), Math.max(1, m), new Float64Array(0), Math.max(1, n), worksize, -1, this.iwork, info);
                let lwork: number = -1;
                if (info[0] != 0) {
                  if (this.vectors)
                    lwork = 3 * Math.min(m, n) * Math.min(m, n) + Math.max(Math.max(m, n), 4 * Math.min(m, n) * Math.min(m, n) + 4 * Math.min(m, n));
                  else
                    lwork = 3 * Math.min(m, n) * Math.min(m, n) + Math.max(Math.max(m, n), 5 * Math.min(m, n) * Math.min(m, n) + 4 * Math.min(m, n));

                } else
                  lwork = <number>worksize[0];

                lwork = Math.max(lwork, 1);
                this.work = new Float64Array(lwork);
              }
              public factor(A: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.blassolver.SVD {
                if (A.rows() != this.m)
                  throw new Error("A.numRows() != m");
                else if (A.columns() != this.n)
                  throw new Error("A.numColumns() != n");


                let info: Int32Array = new Int32Array(1);
                info[0] = 0;
                if (workInPlace) {
                  this._blas.dgesdd(this.job.netlib(), this.m, this.n, A.data(), Math.max(1, this.m), this.S, this.vectors ? this.U.data() : new Float64Array(0), Math.max(1, this.m), this.vectors ? this.Vt.data() : new Float64Array(0), Math.max(1, this.n), this.work, this.work.length, this.iwork, info);
                } else {
                  let Adata: Float64Array = A.data();
                  let cloned: Float64Array = new Float64Array(Adata.length);
                  java.lang.System.arraycopy(Adata, 0, cloned, 0, Adata.length);
                  this._blas.dgesdd(this.job.netlib(), this.m, this.n, cloned, Math.max(1, this.m), this.S, this.vectors ? this.U.data() : new Float64Array(0), Math.max(1, this.m), this.vectors ? this.Vt.data() : new Float64Array(0), Math.max(1, this.n), this.work, this.work.length, this.iwork, info);
                }
                if (info[0] > 0)
                  throw new Error("NotConvergedException.Reason.Iterations");
                else if (info[0] < 0)
                  throw new Error();


                return this;
              }
              public hasSingularVectors(): boolean {
                return this.U != null;
              }
              public getU(): org.mwg.struct.DMatrix {
                return this.U;
              }
              public getVt(): org.mwg.struct.DMatrix {
                return this.Vt;
              }
              public getS(): Float64Array {
                return this.S;
              }
              public getSMatrix(): org.mwg.struct.DMatrix {
                let matS: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, this.n);
                for (let i: number = 0; i < this.S.length; i++) {
                  matS.set(i, i, this.S[i]);
                }
                return matS;
              }
            }
            export module blas {
              export interface Blas {
                dgemm(transA: org.mwg.ml.common.matrix.TransposeType, transB: org.mwg.ml.common.matrix.TransposeType, m: number, n: number, k: number, alpha: number, matA: Float64Array, offsetA: number, ldA: number, matB: Float64Array, offsetB: number, ldB: number, beta: number, matC: Float64Array, offsetC: number, ldC: number): void;
                dgetrs(transA: org.mwg.ml.common.matrix.TransposeType, dim: number, nrhs: number, matA: Float64Array, offsetA: number, ldA: number, ipiv: Int32Array, offsetIpiv: number, matB: Float64Array, offsetB: number, ldB: number, info: Int32Array): void;
                dgetri(dim: number, matA: Float64Array, offsetA: number, ldA: number, ipiv: Int32Array, offsetIpiv: number, work: Float64Array, offsetWork: number, ldWork: number, info: Int32Array): void;
                dgetrf(rows: number, columns: number, matA: Float64Array, offsetA: number, ldA: number, ipiv: Int32Array, offsetIpiv: number, info: Int32Array): void;
                dorgqr(m: number, n: number, k: number, matA: Float64Array, offsetA: number, ldA: number, taw: Float64Array, offsetTaw: number, work: Float64Array, offsetWork: number, lWork: number, info: Int32Array): void;
                dgeqrf(m: number, n: number, matA: Float64Array, offsetA: number, ldA: number, taw: Float64Array, offsetTaw: number, work: Float64Array, offsetwork: number, lWork: number, info: Int32Array): void;
                dgesdd(jobz: string, m: number, n: number, data: Float64Array, lda: number, s: Float64Array, u: Float64Array, ldu: number, vt: Float64Array, ldvt: number, work: Float64Array, length: number, iwork: Int32Array, info: Int32Array): void;
                connect(): void;
                disconnect(): void;
              }
              export class BlasHelper {
                private static TRANSPOSE_TYPE_CONJUCATE: string = "c";
                private static TRANSPOSE_TYPE_NOTRANSPOSE: string = "n";
                private static TRANSPOSE_TYPE_TRANSPOSE: string = "t";
                public static transTypeToChar(type: org.mwg.ml.common.matrix.TransposeType): string {
                  if (type.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                    return BlasHelper.TRANSPOSE_TYPE_NOTRANSPOSE;
                  } else if (type.equals(org.mwg.ml.common.matrix.TransposeType.TRANSPOSE)) {
                    return BlasHelper.TRANSPOSE_TYPE_TRANSPOSE;
                  }

                  return null;
                }
              }
            }
          }
          export module jamasolver {
            export class JamaMatrixEngine implements org.mwg.ml.common.matrix.MatrixEngine {
              public multiplyTransposeAlphaBeta(transA: org.mwg.ml.common.matrix.TransposeType, alpha: number, matA: org.mwg.struct.DMatrix, transB: org.mwg.ml.common.matrix.TransposeType, matB: org.mwg.struct.DMatrix, beta: number, matC: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                if (org.mwg.ml.common.matrix.MatrixOps.testDimensionsAB(transA, transB, matA, matB)) {
                  let dimC: Int32Array = new Int32Array(3);
                  if (transA.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                    if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                      dimC[0] = matA.rows();
                      dimC[1] = matB.columns();
                      dimC[2] = matA.columns();
                    } else {
                      dimC[0] = matA.rows();
                      dimC[1] = matB.rows();
                      dimC[2] = matA.columns();
                    }
                  } else {
                    if (transB.equals(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE)) {
                      dimC[0] = matA.columns();
                      dimC[1] = matB.columns();
                      dimC[2] = matA.rows();
                    } else {
                      dimC[0] = matA.columns();
                      dimC[1] = matB.rows();
                      dimC[2] = matA.rows();
                    }
                  }
                  if (beta == 0 || matC == null) {
                    matC = org.mwg.ml.common.matrix.VolatileDMatrix.empty(dimC[0], dimC[1]);
                  }
                  let temp: number = 0;
                  if (transA == org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE && transB == org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE) {
                    for (let i: number = 0; i < dimC[0]; i++) {
                      for (let j: number = 0; j < dimC[1]; j++) {
                        temp = 0;
                        for (let k: number = 0; k < dimC[2]; k++) {
                          temp += alpha * matA.get(i, k) * matB.get(k, j);
                        }
                        if (beta != 0) {
                          temp = temp + beta * matC.get(i, j);
                        }
                        matC.set(i, j, temp);
                      }
                    }
                  } else if (transA == org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE && transB == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE) {
                    for (let i: number = 0; i < dimC[0]; i++) {
                      for (let j: number = 0; j < dimC[1]; j++) {
                        temp = 0;
                        for (let k: number = 0; k < dimC[2]; k++) {
                          temp += alpha * matA.get(i, k) * matB.get(j, k);
                        }
                        if (beta != 0) {
                          temp = temp + beta * matC.get(i, j);
                        }
                        matC.set(i, j, temp);
                      }
                    }
                  } else if (transA == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE && transB == org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE) {
                    for (let i: number = 0; i < dimC[0]; i++) {
                      for (let j: number = 0; j < dimC[1]; j++) {
                        temp = 0;
                        for (let k: number = 0; k < dimC[2]; k++) {
                          temp += alpha * matA.get(k, i) * matB.get(k, j);
                        }
                        if (beta != 0) {
                          temp = temp + beta * matC.get(i, j);
                        }
                        matC.set(i, j, temp);
                      }
                    }
                  } else if (transA == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE && transB == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE) {
                    for (let i: number = 0; i < dimC[0]; i++) {
                      for (let j: number = 0; j < dimC[1]; j++) {
                        temp = 0;
                        for (let k: number = 0; k < dimC[2]; k++) {
                          temp += alpha * matA.get(k, i) * matB.get(j, k);
                        }
                        if (beta != 0) {
                          temp = temp + beta * matC.get(i, j);
                        }
                        matC.set(i, j, temp);
                      }
                    }
                  }



                  return matC;
                } else {
                  throw new Error("Dimensions mismatch between A,B and C");
                }
              }
              public invert(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
                return this.solve(mat, org.mwg.ml.common.matrix.VolatileDMatrix.identity(mat.rows(), mat.rows()));
              }
              public pinv(mat: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.struct.DMatrix {
                return this.solve(mat, org.mwg.ml.common.matrix.VolatileDMatrix.identity(mat.rows(), mat.rows()));
              }
              public solveLU(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
                let btem: org.mwg.struct.DMatrix;
                if (transB == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE) {
                  btem = org.mwg.ml.common.matrix.MatrixOps.transpose(matB);
                } else {
                  btem = matB;
                }
                return (new org.mwg.ml.common.matrix.jamasolver.LU(matA)).solve(btem);
              }
              public solveQR(matA: org.mwg.struct.DMatrix, matB: org.mwg.struct.DMatrix, workInPlace: boolean, transB: org.mwg.ml.common.matrix.TransposeType): org.mwg.struct.DMatrix {
                let btem: org.mwg.struct.DMatrix;
                if (transB == org.mwg.ml.common.matrix.TransposeType.TRANSPOSE) {
                  btem = org.mwg.ml.common.matrix.MatrixOps.transpose(matB);
                } else {
                  btem = matB;
                }
                return (new org.mwg.ml.common.matrix.jamasolver.QR(matA)).solve(btem);
              }
              public decomposeSVD(matA: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.SVDDecompose {
                return new org.mwg.ml.common.matrix.jamasolver.SVD(matA);
              }
              public solve(A: org.mwg.struct.DMatrix, B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                return (A.rows() == A.columns() ? (new org.mwg.ml.common.matrix.jamasolver.LU(A)).solve(B) : (new org.mwg.ml.common.matrix.jamasolver.QR(A)).solve(B));
              }
            }
            export class LU {
              private LU: org.mwg.struct.DMatrix;
              private m: number;
              private n: number;
              private pivsign: number;
              private piv: Int32Array;
              constructor(A: org.mwg.struct.DMatrix) {
                this.LU = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(A);
                this.m = A.rows();
                this.n = A.columns();
                this.piv = new Int32Array(this.m);
                for (let i: number = 0; i < this.m; i++) {
                  this.piv[i] = i;
                }
                this.pivsign = 1;
                let LUcolj: Float64Array = new Float64Array(this.m);
                for (let j: number = 0; j < this.n; j++) {
                  for (let i: number = 0; i < this.m; i++) {
                    LUcolj[i] = this.LU.get(i, j);
                  }
                  for (let i: number = 0; i < this.m; i++) {
                    let kmax: number = Math.min(i, j);
                    let s: number = 0.0;
                    for (let k: number = 0; k < kmax; k++) {
                      s += this.LU.get(i, k) * LUcolj[k];
                    }
                    LUcolj[i] -= s;
                    this.LU.set(i, j, LUcolj[i]);
                  }
                  let p: number = j;
                  for (let i: number = j + 1; i < this.m; i++) {
                    if (Math.abs(LUcolj[i]) > Math.abs(LUcolj[p])) {
                      p = i;
                    }
                  }
                  if (p != j) {
                    for (let k: number = 0; k < this.n; k++) {
                      let t: number = this.LU.get(p, k);
                      this.LU.set(p, k, this.LU.get(j, k));
                      this.LU.set(j, k, t);
                    }
                    let k: number = this.piv[p];
                    this.piv[p] = this.piv[j];
                    this.piv[j] = k;
                    this.pivsign = -this.pivsign;
                  }
                  if (j < this.m && this.LU.get(j, j) != 0.0) {
                    for (let i: number = j + 1; i < this.m; i++) {
                      this.LU.set(i, j, this.LU.get(i, j) / this.LU.get(j, j));
                    }
                  }
                }
              }
              public isNonsingular(): boolean {
                for (let j: number = 0; j < this.n; j++) {
                  if (this.LU.get(j, j) == 0)
                    return false;

                }
                return true;
              }
              public getL(): org.mwg.struct.DMatrix {
                let L: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, this.n);
                for (let i: number = 0; i < this.m; i++) {
                  for (let j: number = 0; j < this.n; j++) {
                    if (i > j) {
                      L.set(i, j, this.LU.get(i, j));
                    } else if (i == j) {
                      L.set(i, j, 1.0);
                    } else {
                      L.set(i, j, 0.0);
                    }

                  }
                }
                return L;
              }
              public getU(): org.mwg.struct.DMatrix {
                let U: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.n, this.n);
                for (let i: number = 0; i < this.n; i++) {
                  for (let j: number = 0; j < this.n; j++) {
                    if (i <= j) {
                      U.set(i, j, this.LU.get(i, j));
                    } else {
                      U.set(i, j, 0.0);
                    }
                  }
                }
                return U;
              }
              public getPivot(): Int32Array {
                let p: Int32Array = new Int32Array(this.m);
                for (let i: number = 0; i < this.m; i++) {
                  p[i] = this.piv[i];
                }
                return p;
              }
              public getDoublePivot(): Float64Array {
                let vals: Float64Array = new Float64Array(this.m);
                for (let i: number = 0; i < this.m; i++) {
                  vals[i] = <number>this.piv[i];
                }
                return vals;
              }
              public det(): number {
                if (this.m != this.n) {
                  throw new Error("DMatrix must be square.");
                }
                let d: number = <number>this.pivsign;
                for (let j: number = 0; j < this.n; j++) {
                  d *= this.LU.get(j, j);
                }
                return d;
              }
              public solve(B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                if (B.rows() != this.m) {
                  throw new Error("DMatrix row dimensions must agree.");
                }
                if (!this.isNonsingular()) {
                  throw new Error("DMatrix is singular.");
                }
                let nx: number = B.columns();
                let X: org.mwg.struct.DMatrix = this.getMatrix(B, this.piv, 0, nx - 1);
                for (let k: number = 0; k < this.n; k++) {
                  for (let i: number = k + 1; i < this.n; i++) {
                    for (let j: number = 0; j < nx; j++) {
                      X.add(i, j, -X.get(k, j) * this.LU.get(i, k));
                    }
                  }
                }
                for (let k: number = this.n - 1; k >= 0; k--) {
                  for (let j: number = 0; j < nx; j++) {
                    X.set(k, j, X.get(k, j) / this.LU.get(k, k));
                  }
                  for (let i: number = 0; i < k; i++) {
                    for (let j: number = 0; j < nx; j++) {
                      X.add(i, j, -X.get(k, j) * this.LU.get(i, k));
                    }
                  }
                }
                return X;
              }
              private getMatrix(A: org.mwg.struct.DMatrix, r: Int32Array, j0: number, j1: number): org.mwg.struct.DMatrix {
                let B: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(r.length, j1 - j0 + 1);
                try {
                  for (let i: number = 0; i < r.length; i++) {
                    for (let j: number = j0; j <= j1; j++) {
                      B.set(i, j - j0, A.get(r[i], j));
                    }
                  }
                } catch ($ex$) {
                  if ($ex$ instanceof Error) {
                    var e: Error = <Error>$ex$;
{
                      throw new Error("Submatrix indices");
                    }                  } else {
                    throw $ex$;
                  }
                }
                return B;
              }
            }
            export class QR {
              private QR: org.mwg.struct.DMatrix;
              private m: number;
              private n: number;
              private Rdiag: Float64Array;
              constructor(A: org.mwg.struct.DMatrix) {
                this.QR = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(A);
                this.m = A.rows();
                this.n = A.columns();
                this.Rdiag = new Float64Array(this.n);
                for (let k: number = 0; k < this.n; k++) {
                  let nrm: number = 0;
                  for (let i: number = k; i < this.m; i++) {
                    nrm = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(nrm, this.QR.get(i, k));
                  }
                  if (nrm != 0.0) {
                    if (this.QR.get(k, k) < 0) {
                      nrm = -nrm;
                    }
                    for (let i: number = k; i < this.m; i++) {
                      this.QR.set(i, k, this.QR.get(i, k) / nrm);
                    }
                    this.QR.add(k, k, 1.0);
                    for (let j: number = k + 1; j < this.n; j++) {
                      let s: number = 0.0;
                      for (let i: number = k; i < this.m; i++) {
                        s += this.QR.get(i, k) * this.QR.get(i, j);
                      }
                      s = -s / this.QR.get(k, k);
                      for (let i: number = k; i < this.m; i++) {
                        this.QR.add(i, j, s * this.QR.get(i, k));
                      }
                    }
                  }
                  this.Rdiag[k] = -nrm;
                }
              }
              public isFullRank(): boolean {
                for (let j: number = 0; j < this.n; j++) {
                  if (this.Rdiag[j] == 0)
                    return false;

                }
                return true;
              }
              public getH(): org.mwg.struct.DMatrix {
                let H: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, this.n);
                for (let i: number = 0; i < this.m; i++) {
                  for (let j: number = 0; j < this.n; j++) {
                    if (i >= j) {
                      H.set(i, j, this.QR.get(i, j));
                    } else {
                      H.set(i, j, 0.0);
                    }
                  }
                }
                return H;
              }
              public getR(): org.mwg.struct.DMatrix {
                let R: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.n, this.n);
                for (let i: number = 0; i < this.n; i++) {
                  for (let j: number = 0; j < this.n; j++) {
                    if (i < j) {
                      R.set(i, j, this.QR.get(i, j));
                    } else if (i == j) {
                      R.set(i, j, this.Rdiag[i]);
                    } else {
                      R.set(i, j, 0.0);
                    }

                  }
                }
                return R;
              }
              public getQ(): org.mwg.struct.DMatrix {
                let Q: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, this.n);
                for (let k: number = this.n - 1; k >= 0; k--) {
                  for (let i: number = 0; i < this.m; i++) {
                    Q.set(i, k, 0.0);
                  }
                  Q.set(k, k, 1.0);
                  for (let j: number = k; j < this.n; j++) {
                    if (this.QR.get(k, k) != 0) {
                      let s: number = 0.0;
                      for (let i: number = k; i < this.m; i++) {
                        s += this.QR.get(i, k) * Q.get(i, j);
                      }
                      s = -s / this.QR.get(k, k);
                      for (let i: number = k; i < this.m; i++) {
                        Q.add(i, j, s * this.QR.get(i, k));
                      }
                    }
                  }
                }
                return Q;
              }
              public solve(B: org.mwg.struct.DMatrix): org.mwg.struct.DMatrix {
                if (B.rows() != this.m) {
                  throw new Error("DMatrix row dimensions must agree.");
                }
                if (!this.isFullRank()) {
                  throw new Error("DMatrix is rank deficient.");
                }
                let nx: number = B.columns();
                let X: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(B);
                for (let k: number = 0; k < this.n; k++) {
                  for (let j: number = 0; j < nx; j++) {
                    let s: number = 0.0;
                    for (let i: number = k; i < this.m; i++) {
                      s += this.QR.get(i, k) * X.get(i, j);
                    }
                    s = -s / this.QR.get(k, k);
                    for (let i: number = k; i < this.m; i++) {
                      X.add(i, j, s * this.QR.get(i, k));
                    }
                  }
                }
                for (let k: number = this.n - 1; k >= 0; k--) {
                  for (let j: number = 0; j < nx; j++) {
                    X.set(k, j, X.get(k, j) / this.Rdiag[k]);
                  }
                  for (let i: number = 0; i < k; i++) {
                    for (let j: number = 0; j < nx; j++) {
                      X.add(i, j, -X.get(k, j) * this.QR.get(i, k));
                    }
                  }
                }
                return (org.mwg.ml.common.matrix.jamasolver.QR.getMatrix(X, 0, this.n - 1, 0, nx - 1));
              }
              private static getMatrix(B: org.mwg.struct.DMatrix, i0: number, i1: number, j0: number, j1: number): org.mwg.struct.DMatrix {
                let X: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(i1 - i0 + 1, j1 - j0 + 1);
                try {
                  for (let i: number = i0; i <= i1; i++) {
                    for (let j: number = j0; j <= j1; j++) {
                      X.set(i - i0, j - j0, B.get(i, j));
                    }
                  }
                } catch ($ex$) {
                  if ($ex$ instanceof Error) {
                    var e: Error = <Error>$ex$;
{
                      throw new Error("Submatrix indices");
                    }                  } else {
                    throw $ex$;
                  }
                }
                return X;
              }
            }
            export class SVD implements org.mwg.ml.common.matrix.SVDDecompose {
              private U: org.mwg.struct.DMatrix;
              private V: org.mwg.struct.DMatrix;
              private s: Float64Array;
              private m: number;
              private n: number;
              constructor(Arg: org.mwg.struct.DMatrix) {
                let A: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(Arg);
                this.m = Arg.rows();
                this.n = Arg.columns();
                let nu: number = Math.min(this.m, this.n);
                this.s = new Float64Array(Math.min(this.m + 1, this.n));
                this.U = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.m, nu);
                this.V = org.mwg.ml.common.matrix.VolatileDMatrix.empty(this.n, this.n);
                let e: Float64Array = new Float64Array(this.n);
                let work: Float64Array = new Float64Array(this.m);
                let wantu: boolean = true;
                let wantv: boolean = true;
                let nct: number = Math.min(this.m - 1, this.n);
                let nrt: number = Math.max(0, Math.min(this.n - 2, this.m));
                for (let k: number = 0; k < Math.max(nct, nrt); k++) {
                  if (k < nct) {
                    this.s[k] = 0;
                    for (let i: number = k; i < this.m; i++) {
                      this.s[k] = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(this.s[k], A.get(i, k));
                    }
                    if (this.s[k] != 0.0) {
                      if (A.get(k, k) < 0.0) {
                        this.s[k] = -this.s[k];
                      }
                      for (let i: number = k; i < this.m; i++) {
                        A.set(i, k, A.get(i, k) / this.s[k]);
                      }
                      A.add(k, k, 1.0);
                    }
                    this.s[k] = -this.s[k];
                  }
                  for (let j: number = k + 1; j < this.n; j++) {
                    if ((k < nct) && (this.s[k] != 0.0)) {
                      let t: number = 0;
                      for (let i: number = k; i < this.m; i++) {
                        t += A.get(i, k) * A.get(i, j);
                      }
                      t = -t / A.get(k, k);
                      for (let i: number = k; i < this.m; i++) {
                        A.add(i, j, +t * A.get(i, k));
                      }
                    }
                    e[j] = A.get(k, j);
                  }
                  if (wantu && (k < nct)) {
                    for (let i: number = k; i < this.m; i++) {
                      this.U.set(i, k, A.get(i, k));
                    }
                  }
                  if (k < nrt) {
                    e[k] = 0;
                    for (let i: number = k + 1; i < this.n; i++) {
                      e[k] = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(e[k], e[i]);
                    }
                    if (e[k] != 0.0) {
                      if (e[k + 1] < 0.0) {
                        e[k] = -e[k];
                      }
                      for (let i: number = k + 1; i < this.n; i++) {
                        e[i] /= e[k];
                      }
                      e[k + 1] += 1.0;
                    }
                    e[k] = -e[k];
                    if ((k + 1 < this.m) && (e[k] != 0.0)) {
                      for (let i: number = k + 1; i < this.m; i++) {
                        work[i] = 0.0;
                      }
                      for (let j: number = k + 1; j < this.n; j++) {
                        for (let i: number = k + 1; i < this.m; i++) {
                          work[i] += e[j] * A.get(i, j);
                        }
                      }
                      for (let j: number = k + 1; j < this.n; j++) {
                        let t: number = -e[j] / e[k + 1];
                        for (let i: number = k + 1; i < this.m; i++) {
                          A.add(i, j, t * work[i]);
                        }
                      }
                    }
                    if (wantv) {
                      for (let i: number = k + 1; i < this.n; i++) {
                        this.V.set(i, k, e[i]);
                      }
                    }
                  }
                }
                let p: number = Math.min(this.n, this.m + 1);
                if (nct < this.n) {
                  this.s[nct] = A.get(nct, nct);
                }
                if (this.m < p) {
                  this.s[p - 1] = 0.0;
                }
                if (nrt + 1 < p) {
                  e[nrt] = A.get(nrt, p - 1);
                }
                e[p - 1] = 0.0;
                if (wantu) {
                  for (let j: number = nct; j < nu; j++) {
                    for (let i: number = 0; i < this.m; i++) {
                      this.U.set(i, j, 0.0);
                    }
                    this.U.set(j, j, 1.0);
                  }
                  for (let k: number = nct - 1; k >= 0; k--) {
                    if (this.s[k] != 0.0) {
                      for (let j: number = k + 1; j < nu; j++) {
                        let t: number = 0;
                        for (let i: number = k; i < this.m; i++) {
                          t += this.U.get(i, k) * this.U.get(i, j);
                        }
                        t = -t / this.U.get(k, k);
                        for (let i: number = k; i < this.m; i++) {
                          this.U.add(i, j, t * this.U.get(i, k));
                        }
                      }
                      for (let i: number = k; i < this.m; i++) {
                        this.U.set(i, k, -this.U.get(i, k));
                      }
                      this.U.set(k, k, 1.0 + this.U.get(k, k));
                      for (let i: number = 0; i < k - 1; i++) {
                        this.U.set(i, k, 0.0);
                      }
                    } else {
                      for (let i: number = 0; i < this.m; i++) {
                        this.U.set(i, k, 0.0);
                      }
                      this.U.set(k, k, 1.0);
                    }
                  }
                }
                if (wantv) {
                  for (let k: number = this.n - 1; k >= 0; k--) {
                    if ((k < nrt) && (e[k] != 0.0)) {
                      for (let j: number = k + 1; j < nu; j++) {
                        let t: number = 0;
                        for (let i: number = k + 1; i < this.n; i++) {
                          t += this.V.get(i, k) * this.V.get(i, j);
                        }
                        t = -t / this.V.get(k + 1, k);
                        for (let i: number = k + 1; i < this.n; i++) {
                          this.V.add(i, j, t * this.V.get(i, k));
                        }
                      }
                    }
                    for (let i: number = 0; i < this.n; i++) {
                      this.V.set(i, k, 0.0);
                    }
                    this.V.set(k, k, 1.0);
                  }
                }
                let pp: number = p - 1;
                let iter: number = 0;
                let eps: number = Math.pow(2.0, -52.0);
                let tiny: number = Math.pow(2.0, -966.0);
                while (p > 0) {
                  let k: number, kase: number;
                  for (k = p - 2; k >= -1; k--) {
                    if (k == -1) {
break;
                    }
                    if (Math.abs(e[k]) <= tiny + eps * (Math.abs(this.s[k]) + Math.abs(this.s[k + 1]))) {
                      e[k] = 0.0;
break;
                    }
                  }
                  if (k == p - 2) {
                    kase = 4;
                  } else {
                    let ks: number;
                    for (ks = p - 1; ks >= k; ks--) {
                      if (ks == k) {
break;
                      }
                      let t: number = (ks != p ? Math.abs(e[ks]) : 0.) + (ks != k + 1 ? Math.abs(e[ks - 1]) : 0.);
                      if (Math.abs(this.s[ks]) <= tiny + eps * t) {
                        this.s[ks] = 0.0;
break;
                      }
                    }
                    if (ks == k) {
                      kase = 3;
                    } else if (ks == p - 1) {
                      kase = 1;
                    } else {
                      kase = 2;
                      k = ks;
                    }

                  }
                  k++;
                  switch (kase) {
                    case 1: {
                      let f: number = e[p - 2];
                      e[p - 2] = 0.0;
                      for (let j: number = p - 2; j >= k; j--) {
                        let t: number = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(this.s[j], f);
                        let cs: number = this.s[j] / t;
                        let sn: number = f / t;
                        this.s[j] = t;
                        if (j != k) {
                          f = -sn * e[j - 1];
                          e[j - 1] = cs * e[j - 1];
                        }
                        if (wantv) {
                          for (let i: number = 0; i < this.n; i++) {
                            t = cs * this.V.get(i, j) + sn * this.V.get(i, p - 1);
                            this.V.set(i, p - 1, -sn * this.V.get(i, j) + cs * this.V.get(i, p - 1));
                            this.V.set(i, j, t);
                          }
                        }
                      }
                    }break;
                    case 2: {
                      let f: number = e[k - 1];
                      e[k - 1] = 0.0;
                      for (let j: number = k; j < p; j++) {
                        let t: number = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(this.s[j], f);
                        let cs: number = this.s[j] / t;
                        let sn: number = f / t;
                        this.s[j] = t;
                        f = -sn * e[j];
                        e[j] = cs * e[j];
                        if (wantu) {
                          for (let i: number = 0; i < this.m; i++) {
                            t = cs * this.U.get(i, j) + sn * this.U.get(i, k - 1);
                            this.U.set(i, k - 1, -sn * this.U.get(i, j) + cs * this.U.get(i, k - 1));
                            this.U.set(i, j, t);
                          }
                        }
                      }
                    }break;
                    case 3: {
                      let scale: number = Math.max(Math.max(Math.max(Math.max(Math.abs(this.s[p - 1]), Math.abs(this.s[p - 2])), Math.abs(e[p - 2])), Math.abs(this.s[k])), Math.abs(e[k]));
                      let sp: number = this.s[p - 1] / scale;
                      let spm1: number = this.s[p - 2] / scale;
                      let epm1: number = e[p - 2] / scale;
                      let sk: number = this.s[k] / scale;
                      let ek: number = e[k] / scale;
                      let b: number = ((spm1 + sp) * (spm1 - sp) + epm1 * epm1) / 2.0;
                      let c: number = (sp * epm1) * (sp * epm1);
                      let shift: number = 0.0;
                      if ((b != 0.0) || (c != 0.0)) {
                        shift = Math.sqrt(b * b + c);
                        if (b < 0.0) {
                          shift = -shift;
                        }
                        shift = c / (b + shift);
                      }
                      let f: number = (sk + sp) * (sk - sp) + shift;
                      let g: number = sk * ek;
                      for (let j: number = k; j < p - 1; j++) {
                        let t: number = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(f, g);
                        let cs: number = f / t;
                        let sn: number = g / t;
                        if (j != k) {
                          e[j - 1] = t;
                        }
                        f = cs * this.s[j] + sn * e[j];
                        e[j] = cs * e[j] - sn * this.s[j];
                        g = sn * this.s[j + 1];
                        this.s[j + 1] = cs * this.s[j + 1];
                        if (wantv) {
                          for (let i: number = 0; i < this.n; i++) {
                            t = cs * this.V.get(i, j) + sn * this.V.get(i, j + 1);
                            this.V.set(i, j + 1, -sn * this.V.get(i, j) + cs * this.V.get(i, j + 1));
                            this.V.set(i, j, t);
                          }
                        }
                        t = org.mwg.ml.common.matrix.jamasolver.Utils.hypot(f, g);
                        cs = f / t;
                        sn = g / t;
                        this.s[j] = t;
                        f = cs * e[j] + sn * this.s[j + 1];
                        this.s[j + 1] = -sn * e[j] + cs * this.s[j + 1];
                        g = sn * e[j + 1];
                        e[j + 1] = cs * e[j + 1];
                        if (wantu && (j < this.m - 1)) {
                          for (let i: number = 0; i < this.m; i++) {
                            t = cs * this.U.get(i, j) + sn * this.U.get(i, j + 1);
                            this.U.set(i, j + 1, -sn * this.U.get(i, j) + cs * this.U.get(i, j + 1));
                            this.U.set(i, j, t);
                          }
                        }
                      }
                      e[p - 2] = f;
                      iter = iter + 1;
                    }break;
                    case 4: {
                      if (this.s[k] <= 0.0) {
                        this.s[k] = (this.s[k] < 0.0 ? -this.s[k] : 0.0);
                        if (wantv) {
                          for (let i: number = 0; i <= pp; i++) {
                            this.V.set(i, k, -this.V.get(i, k));
                          }
                        }
                      }
                      while (k < pp) {
                        if (this.s[k] >= this.s[k + 1]) {
break;
                        }
                        let t: number = this.s[k];
                        this.s[k] = this.s[k + 1];
                        this.s[k + 1] = t;
                        if (wantv && (k < this.n - 1)) {
                          for (let i: number = 0; i < this.n; i++) {
                            t = this.V.get(i, k + 1);
                            this.V.set(i, k + 1, this.V.get(i, k));
                            this.V.set(i, k, t);
                          }
                        }
                        if (wantu && (k < this.m - 1)) {
                          for (let i: number = 0; i < this.m; i++) {
                            t = this.U.get(i, k + 1);
                            this.U.set(i, k + 1, this.U.get(i, k));
                            this.U.set(i, k, t);
                          }
                        }
                        k++;
                      }
                      iter = 0;
                      p--;
                    }break;
                  }
                }
              }
              public factor(A: org.mwg.struct.DMatrix, workInPlace: boolean): org.mwg.ml.common.matrix.jamasolver.SVD {
                return new org.mwg.ml.common.matrix.jamasolver.SVD(A);
              }
              public getU(): org.mwg.struct.DMatrix {
                return this.U;
              }
              public getVt(): org.mwg.struct.DMatrix {
                return org.mwg.ml.common.matrix.MatrixOps.transpose(this.getV());
              }
              public getV(): org.mwg.struct.DMatrix {
                return this.V;
              }
              public getSingularValues(): Float64Array {
                return this.s;
              }
              public getSMatrix(): org.mwg.struct.DMatrix {
                let X: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(Math.min(this.m, this.n), this.n);
                for (let i: number = 0; i < this.s.length; i++) {
                  if (i < this.m && i < this.n) {
                    X.set(i, i, this.s[i]);
                  }
                }
                return X;
              }
              public getS(): Float64Array {
                return this.s;
              }
              public norm2(): number {
                return this.s[0];
              }
              public cond(): number {
                return this.s[0] / this.s[Math.min(this.m, this.n) - 1];
              }
              public rank(): number {
                let eps: number = Math.pow(2.0, -52.0);
                let tol: number = Math.max(this.m, this.n) * this.s[0] * eps;
                let r: number = 0;
                for (let i: number = 0; i < this.s.length; i++) {
                  if (this.s[i] > tol) {
                    r++;
                  }
                }
                return r;
              }
            }
            export class Utils {
              public static hypot(a: number, b: number): number {
                let r: number;
                if (Math.abs(a) > Math.abs(b)) {
                  r = b / a;
                  r = Math.abs(a) * Math.sqrt(1 + r * r);
                } else if (b != 0) {
                  r = a / b;
                  r = Math.abs(b) * Math.sqrt(1 + r * r);
                } else {
                  r = 0.0;
                }

                return r;
              }
            }
          }
          export module operation {
            export class Gaussian1D {
              public static getCovariance(sum: number, sumSq: number, total: number): number {
                return (sumSq - (sum * sum) / total) / (total - 1);
              }
              public static getDensity(sum: number, sumSq: number, total: number, feature: number): number {
                if (total < 2) {
                  return 0;
                }
                let avg: number = sum / total;
                let cov: number = org.mwg.ml.common.matrix.operation.Gaussian1D.getCovariance(sum, sumSq, total);
                return 1 / Math.sqrt(2 * Math.PI * cov) * Math.exp(-(feature - avg) * (feature - avg) / (2 * cov));
              }
              public static getDensityArray(sum: number, sumSq: number, total: number, feature: Float64Array): Float64Array {
                if (total < 2) {
                  return null;
                }
                let avg: number = sum / total;
                let cov: number = org.mwg.ml.common.matrix.operation.Gaussian1D.getCovariance(sum, sumSq, total);
                let exp: number = 1 / Math.sqrt(2 * Math.PI * cov);
                let proba: Float64Array = new Float64Array(feature.length);
                for (let i: number = 0; i < feature.length; i++) {
                  proba[i] = exp * Math.exp(-(feature[i] - avg) * (feature[i] - avg) / (2 * cov));
                }
                return proba;
              }
            }
            export class MultivariateNormalDistribution {
              public min: Float64Array;
              public max: Float64Array;
              public means: Float64Array;
              public covDiag: Float64Array;
              public inv: org.mwg.struct.DMatrix;
              public covariance: org.mwg.struct.DMatrix;
              public pinvsvd: org.mwg.ml.common.matrix.operation.PInvSVD;
              public rank: number;
              public det: number;
              constructor(means: Float64Array, cov: org.mwg.struct.DMatrix, allowSingular: boolean) {
                this.means = means;
                if (cov != null) {
                  this.covariance = cov;
                  this.covDiag = new Float64Array(cov.rows());
                  for (let i: number = 0; i < this.covDiag.length; i++) {
                    this.covDiag[i] = cov.get(i, i);
                  }
                  this.pinvsvd = new org.mwg.ml.common.matrix.operation.PInvSVD();
                  this.pinvsvd.factor(this.covariance, false);
                  this.inv = this.pinvsvd.getPInv();
                  this.det = this.pinvsvd.getDeterminant();
                  this.rank = this.pinvsvd.getRank();
                  if (!allowSingular && this.rank < cov.rows()) {
                    this.covariance = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(cov);
                    let temp: Float64Array = new Float64Array(this.covDiag.length);
                    for (let i: number = 0; i < this.covDiag.length; i++) {
                      temp[i] = Math.sqrt(this.covDiag[i]);
                    }
                    for (let i: number = 0; i < this.covDiag.length; i++) {
                      for (let j: number = i + 1; j < this.covDiag.length; j++) {
                        let d: number = this.covariance.get(i, j) - 0.0001 * temp[i] * temp[j];
                        this.covariance.set(i, j, d);
                        this.covariance.set(j, i, d);
                      }
                    }
                    this.pinvsvd = new org.mwg.ml.common.matrix.operation.PInvSVD();
                    this.pinvsvd.factor(this.covariance, false);
                    this.inv = this.pinvsvd.getPInv();
                    this.det = this.pinvsvd.getDeterminant();
                    this.rank = this.pinvsvd.getRank();
                  }
                }
              }
              public getMin(): Float64Array {
                return this.min;
              }
              public getMax(): Float64Array {
                return this.max;
              }
              public getAvg(): Float64Array {
                return this.means;
              }
              public getCovDiag(): Float64Array {
                return this.covDiag;
              }
              public setMin(min: Float64Array): void {
                this.min = min;
              }
              public setMax(max: Float64Array): void {
                this.max = max;
              }
              public static getCovariance(sum: Float64Array, sumsquares: Float64Array, total: number): org.mwg.struct.DMatrix {
                if (total < 2) {
                  return null;
                }
                let features: number = sum.length;
                let avg: Float64Array = new Float64Array(features);
                for (let i: number = 0; i < features; i++) {
                  avg[i] = sum[i] / total;
                }
                let covariances: Float64Array = new Float64Array(features * features);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    covariances[i * features + j] = (sumsquares[count] / total - avg[i] * avg[j]) * correction;
                    covariances[j * features + i] = covariances[i * features + j];
                    count++;
                  }
                }
                return org.mwg.ml.common.matrix.VolatileDMatrix.wrap(covariances, features, features);
              }
              public static getDistribution(sum: Float64Array, sumsquares: Float64Array, total: number, allowSingular: boolean): org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution {
                if (total < 2) {
                  return null;
                }
                let features: number = sum.length;
                let avg: Float64Array = new Float64Array(features);
                for (let i: number = 0; i < features; i++) {
                  avg[i] = sum[i] / total;
                }
                let covariances: Float64Array = new Float64Array(features * features);
                let correction: number = total;
                correction = correction / (total - 1);
                let count: number = 0;
                for (let i: number = 0; i < features; i++) {
                  for (let j: number = i; j < features; j++) {
                    covariances[i * features + j] = (sumsquares[count] / total - avg[i] * avg[j]) * correction;
                    covariances[j * features + i] = covariances[i * features + j];
                    count++;
                  }
                }
                let cov: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(covariances, features, features);
                return new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(avg, cov, allowSingular);
              }
              public density(features: Float64Array, normalizeOnAvg: boolean): number {
                if (normalizeOnAvg) {
                  return this.getExponentTerm(features);
                } else {
                  return Math.pow(2 * Math.PI, -0.5 * this.rank) * Math.pow(this.det, -0.5) * this.getExponentTerm(features);
                }
              }
              private getExponentTerm(features: Float64Array): number {
                let f: Float64Array = new Float64Array(features.length);
                java.lang.System.arraycopy(features, 0, f, 0, features.length);
                for (let i: number = 0; i < features.length; i++) {
                  f[i] = f[i] - this.means[i];
                }
                let ft: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(f, 1, f.length);
                let ftt: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(f, f.length, 1);
                let res: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiply(ft, this.inv);
                let res2: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiply(res, ftt);
                let d: number = Math.exp(-0.5 * res2.get(0, 0));
                return d;
              }
              public clone(avg: Float64Array): org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution {
                let res: org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution = new org.mwg.ml.common.matrix.operation.MultivariateNormalDistribution(avg, null, false);
                res.pinvsvd = this.pinvsvd;
                res.inv = this.inv;
                res.det = this.det;
                res.rank = this.rank;
                res.covDiag = this.covDiag;
                return res;
              }
              public densityExponent(features: Float64Array): number {
                let f: Float64Array = new Float64Array(features.length);
                java.lang.System.arraycopy(features, 0, f, 0, features.length);
                for (let i: number = 0; i < features.length; i++) {
                  f[i] = f[i] - this.means[i];
                }
                let ft: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(f, 1, f.length);
                let ftt: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(f, f.length, 1);
                let res: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiply(ft, this.inv);
                let res2: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiply(res, ftt);
                return -0.5 * res2.get(0, 0);
              }
            }
            export class PInvSVD {
              private _svd: org.mwg.ml.common.matrix.SVDDecompose;
              private pinv: org.mwg.struct.DMatrix;
              private S: org.mwg.struct.DMatrix;
              private rank: number;
              private det: number;
              public getRank(): number {
                return this.rank;
              }
              public getDeterminant(): number {
                return this.det;
              }
              constructor() {}
              public factor(A: org.mwg.struct.DMatrix, invertInPlace: boolean): org.mwg.ml.common.matrix.operation.PInvSVD {
                this._svd = org.mwg.ml.common.matrix.MatrixOps.defaultEngine().decomposeSVD(A, invertInPlace);
                let svd: org.mwg.struct.DMatrix[] = new Array<org.mwg.ml.common.matrix.VolatileDMatrix>(3);
                svd[0] = this._svd.getU();
                svd[1] = this._svd.getSMatrix();
                svd[2] = this._svd.getVt();
                let V: org.mwg.struct.DMatrix = this._svd.getVt();
                this.S = org.mwg.ml.common.matrix.VolatileDMatrix.cloneFrom(this._svd.getSMatrix());
                let maxSingular: number = 0;
                let dim: number = Math.min(this.S.columns(), this.S.rows());
                for (let i: number = 0; i < dim; i++) {
                  if (this.S.get(i, i) > maxSingular)
                    maxSingular = this.S.get(i, i);

                }
                let tau: number = Math.pow(2, -46) * Math.max(A.columns(), A.rows()) * maxSingular;
                this.rank = 0;
                this.det = 1;
                if (maxSingular != 0.0) {
                  for (let i: number = 0; i < dim; i++) {
                    let s: number = this.S.get(i, i);
                    if (s < tau)
                      this.S.set(i, i, 0);
                    else {
                      this.S.set(i, i, 1 / s);
                      this.det = this.det * s;
                      this.rank++;
                    }
                  }
                }
                let temp: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.MatrixOps.multiplyTranspose(org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, V, org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, this.S);
                this.pinv = org.mwg.ml.common.matrix.MatrixOps.multiplyTranspose(org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE, temp, org.mwg.ml.common.matrix.TransposeType.TRANSPOSE, this._svd.getU());
                return this;
              }
              public getSvd(): org.mwg.ml.common.matrix.SVDDecompose {
                return this._svd;
              }
              public getInvDeterminant(): org.mwg.struct.DMatrix {
                return this.S;
              }
              public getPInv(): org.mwg.struct.DMatrix {
                return this.pinv;
              }
            }
            export class PolynomialFit {
              private coef: org.mwg.struct.DMatrix;
              private degree: number = 0;
              constructor(degree: number) {
                this.degree = degree;
              }
              public getCoef(): Float64Array {
                return this.coef.data();
              }
              public fit(samplePoints: Float64Array, observations: Float64Array): void {
                let y: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.wrap(observations, observations.length, 1);
                let a: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(y.rows(), this.degree + 1);
                for (let i: number = 0; i < observations.length; i++) {
                  let obs: number = 1;
                  for (let j: number = 0; j < this.degree + 1; j++) {
                    a.set(i, j, obs);
                    obs *= samplePoints[i];
                  }
                }
                this.coef = org.mwg.ml.common.matrix.MatrixOps.defaultEngine().solveQR(a, y, true, org.mwg.ml.common.matrix.TransposeType.NOTRANSPOSE);
              }
              public static extrapolate(time: number, weights: Float64Array): number {
                let result: number = 0;
                let power: number = 1;
                for (let j: number = 0; j < weights.length; j++) {
                  result += weights[j] * power;
                  power = power * time;
                }
                return result;
              }
            }
          }
        }
      }
      export module neuralnet {
        export class FlatNeuralNode extends org.mwg.base.BaseNode {
          public static NAME: string = "FlatNeuralNode";
          private static NB_INPUTS: number = 1;
          private static NB_OUTPUTS: number = 2;
          private static NB_LAYERS: number = 3;
          private static NB_PER_LAYER: number = 4;
          private static LEARNING_RATE: number = 5;
          private static MATRICES_OFFSET: number = 10;
          private static layerWeights(state: org.mwg.plugin.NodeState, layer: number): org.mwg.struct.DMatrix {
            return <org.mwg.struct.DMatrix>state.getOrCreate(FlatNeuralNode.MATRICES_OFFSET + layer * 2, org.mwg.Type.DMATRIX);
          }
          private static layerBias(state: org.mwg.plugin.NodeState, layer: number): org.mwg.struct.DMatrix {
            return <org.mwg.struct.DMatrix>state.getOrCreate(FlatNeuralNode.MATRICES_OFFSET + layer * 2 + 1, org.mwg.Type.DMATRIX);
          }
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          public configure(inputs: number, outputs: number, hiddenlayers: number, nodesPerLayer: number, learningRate: number): org.mwg.ml.neuralnet.FlatNeuralNode {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            state.set(FlatNeuralNode.NB_INPUTS, org.mwg.Type.INT, inputs);
            state.set(FlatNeuralNode.NB_OUTPUTS, org.mwg.Type.INT, outputs);
            state.set(FlatNeuralNode.NB_LAYERS, org.mwg.Type.INT, hiddenlayers);
            state.set(FlatNeuralNode.NB_PER_LAYER, org.mwg.Type.INT, nodesPerLayer);
            state.set(FlatNeuralNode.LEARNING_RATE, org.mwg.Type.DOUBLE, learningRate);
            let previousDim: number = inputs;
            let nextDim: number = nodesPerLayer;
            let rand: java.util.Random = new java.util.Random();
            rand.setSeed(1234);
            for (let i: number = 0; i < hiddenlayers; i++) {
              let weights: org.mwg.struct.DMatrix = org.mwg.ml.neuralnet.FlatNeuralNode.layerWeights(state, i);
              let biases: org.mwg.struct.DMatrix = org.mwg.ml.neuralnet.FlatNeuralNode.layerBias(state, i);
              weights.init(previousDim, nextDim);
              biases.init(1, nextDim);
              weights.fillWithRandom(rand, -1.0, 1.0);
              biases.fillWithRandom(rand, -1.0, 1.0);
              previousDim = nodesPerLayer;
            }
            nextDim = outputs;
            let weights: org.mwg.struct.DMatrix = org.mwg.ml.neuralnet.FlatNeuralNode.layerWeights(state, hiddenlayers);
            let biases: org.mwg.struct.DMatrix = org.mwg.ml.neuralnet.FlatNeuralNode.layerBias(state, hiddenlayers);
            weights.init(previousDim, nextDim);
            biases.init(1, nextDim);
            weights.fillWithRandom(rand, -1.0, 1.0);
            biases.fillWithRandom(rand, -1.0, 1.0);
            return this;
          }
          public predict(inputVec: Float64Array): Float64Array {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let nbInput: number = <number>state.get(FlatNeuralNode.NB_INPUTS);
            if (inputVec.length != nbInput) {
              throw new Error("Please reconfigure the neuralnet before changing input or output dimensions");
            }
            let nbHiddenLayers: number = <number>state.get(FlatNeuralNode.NB_LAYERS);
            let integration: org.mwg.struct.DMatrix;
            let activation: org.mwg.struct.DMatrix;
            let input: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(1, nbInput);
            let weights: org.mwg.struct.DMatrix;
            let biases: org.mwg.struct.DMatrix;
            for (let i: number = 0; i < nbInput; i++) {
              input.set(0, i, inputVec[i]);
            }
            for (let layer: number = 0; layer < nbHiddenLayers + 1; layer++) {
              weights = org.mwg.ml.neuralnet.FlatNeuralNode.layerWeights(state, layer);
              biases = org.mwg.ml.neuralnet.FlatNeuralNode.layerBias(state, layer);
              integration = org.mwg.ml.common.matrix.MatrixOps.add(org.mwg.ml.common.matrix.MatrixOps.multiply(input, weights), biases);
              activation = this.activate(integration, layer == nbHiddenLayers);
              input = activation;
            }
            return input.data();
          }
          public learn(inputVec: Float64Array, outputVec: Float64Array): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let nbInput: number = <number>state.get(FlatNeuralNode.NB_INPUTS);
            let nbOutput: number = <number>state.get(FlatNeuralNode.NB_OUTPUTS);
            if (inputVec.length != nbInput || outputVec.length != nbOutput) {
              throw new Error("Please reconfigure the neuralnet before changing input or output dimensions");
            }
            let nbHiddenLayers: number = <number>state.get(FlatNeuralNode.NB_LAYERS);
            let learningRate: number = <number>state.get(FlatNeuralNode.LEARNING_RATE);
            let forwards: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let integrations: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let activations: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let derivations: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let input: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(1, nbInput);
            let weights: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let biases: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            for (let i: number = 0; i < nbInput; i++) {
              input.set(0, i, inputVec[i]);
            }
            for (let layer: number = 0; layer < nbHiddenLayers + 1; layer++) {
              forwards[layer] = input;
              weights[layer] = org.mwg.ml.neuralnet.FlatNeuralNode.layerWeights(state, layer);
              biases[layer] = org.mwg.ml.neuralnet.FlatNeuralNode.layerBias(state, layer);
              integrations[layer] = org.mwg.ml.common.matrix.MatrixOps.add(org.mwg.ml.common.matrix.MatrixOps.multiply(input, weights[layer]), biases[layer]);
              activations[layer] = this.activate(integrations[layer], layer == nbHiddenLayers);
              derivations[layer] = this.derivate(integrations[layer], activations[layer], layer == nbHiddenLayers);
              input = activations[layer];
            }
            let calculated: Float64Array = input.data();
            let derivativeErr: Float64Array = new Float64Array(calculated.length);
            for (let i: number = 0; i < calculated.length; i++) {
              derivativeErr[i] = -(outputVec[i] - calculated[i]);
            }
            let newWeights: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let newBiases: org.mwg.struct.DMatrix[] = new Array<org.mwg.struct.DMatrix>(nbHiddenLayers + 1);
            let previousErr: Float64Array;
            for (let layer: number = nbHiddenLayers; layer >= 0; layer--) {
              newWeights[layer] = org.mwg.ml.common.matrix.VolatileDMatrix.empty(weights[layer].rows(), weights[layer].columns());
              newBiases[layer] = org.mwg.ml.common.matrix.VolatileDMatrix.empty(biases[layer].rows(), biases[layer].columns());
              for (let i: number = 0; i < derivativeErr.length; i++) {
                derivativeErr[i] = derivativeErr[i] * derivations[layer].get(0, i);
              }
              if (layer > 0) {
                previousErr = new Float64Array(derivations[layer - 1].columns());
              } else {
                previousErr = null;
              }
              for (let column: number = 0; column < newWeights[layer].columns(); column++) {
                for (let row: number = 0; row < newWeights[layer].rows(); row++) {
                  newWeights[layer].set(row, column, weights[layer].get(row, column) - learningRate * derivativeErr[column] * forwards[layer].get(0, row));
                  if (layer > 0) {
                    previousErr[row] += derivativeErr[column] * weights[layer].get(row, column);
                  }
                }
                newBiases[layer].set(0, column, biases[layer].get(0, column) - learningRate * derivativeErr[column]);
              }
              derivativeErr = previousErr;
              org.mwg.ml.common.matrix.MatrixOps.copyMatrix(newWeights[layer], weights[layer]);
              org.mwg.ml.common.matrix.MatrixOps.copyMatrix(newBiases[layer], biases[layer]);
            }
          }
          private activate(integration: org.mwg.struct.DMatrix, linearActivation: boolean): org.mwg.struct.DMatrix {
            if (linearActivation) {
              return integration;
            } else {
              let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(integration.rows(), integration.columns());
              for (let i: number = 0; i < integration.rows(); i++) {
                for (let j: number = 0; j < integration.columns(); j++) {
                  result.set(i, j, 1 / (1 + Math.exp(-integration.get(i, j))));
                }
              }
              return result;
            }
          }
          private derivate(integration: org.mwg.struct.DMatrix, activation: org.mwg.struct.DMatrix, linearActivation: boolean): org.mwg.struct.DMatrix {
            let result: org.mwg.struct.DMatrix = org.mwg.ml.common.matrix.VolatileDMatrix.empty(1, activation.columns());
            if (linearActivation) {
              for (let j: number = 0; j < activation.columns(); j++) {
                result.set(0, j, 1);
              }
            } else {
              for (let j: number = 0; j < activation.columns(); j++) {
                result.set(0, j, activation.get(0, j) * (1 - activation.get(0, j)));
              }
            }
            return result;
          }
        }
        export interface NeuralFunction {
          activate(x: number): number;
          derivate(x: number, fct: number): number;
        }
        export class NeuralNode extends org.mwg.base.BaseNode {
          public static NAME: string = "NeuralNode";
          public static INPUTS: string = "inputs";
          public static INPUTS_MAP: string = "inputs_map";
          public static OUTPUTS: string = "outputs";
          public static OUTPUTS_MAP: string = "outputs_map";
          private static WEIGHTS: string = "weights";
          private static NODE_TYPE: string = "node_type";
          public static LEARNINGRATE: string = "learning";
          public static LEARNINGRATE_DEF: number = 0.01;
          private static random: java.util.Random = new java.util.Random();
          private static msgIdCounter: number = -1;
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          public configure(inputs: number, outputs: number, hiddenlayers: number, nodesPerLayer: number): org.mwg.ml.neuralnet.NeuralNode {
            let internalNodes: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode>();
            internalNodes.add(this);
            let previousLayer: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode>();
            let state: org.mwg.plugin.NodeState = this.phasedState();
            state.setFromKey(NeuralNode.NODE_TYPE, org.mwg.Type.INT, org.mwg.ml.neuralnet.NeuralNodeType.ROOT);
            for (let i: number = 0; i < inputs; i++) {
              let inputNode: org.mwg.ml.neuralnet.NeuralNode = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.INPUT);
              this.forwardConnect(inputNode);
              internalNodes.add(inputNode);
              previousLayer.add(inputNode);
            }
            let nextLayer: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode>();
            for (let i: number = 0; i < hiddenlayers; i++) {
              for (let j: number = 0; j < nodesPerLayer; j++) {
                let hidden: org.mwg.ml.neuralnet.NeuralNode = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN);
                nextLayer.add(hidden);
                internalNodes.add(hidden);
                for (let k: number = 0; k < previousLayer.size(); k++) {
                  previousLayer.get(k).forwardConnect(hidden);
                }
              }
              previousLayer = nextLayer;
              nextLayer = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNode>();
            }
            for (let i: number = 0; i < outputs; i++) {
              let output: org.mwg.ml.neuralnet.NeuralNode = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.OUTPUT);
              for (let k: number = 0; k < previousLayer.size(); k++) {
                previousLayer.get(k).forwardConnect(output);
              }
              internalNodes.add(output);
            }
            for (let i: number = 0; i < internalNodes.size(); i++) {
              if (internalNodes.get(i).id() != this.id()) {
                internalNodes.get(i).initWeightsRadomly(0.1);
                internalNodes.get(i).free();
              }
            }
            return this;
          }
          private initWeightsRadomly(maxValue: number): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let type: number = state.getFromKeyWithDefault(NeuralNode.NODE_TYPE, org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN);
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN || type == org.mwg.ml.neuralnet.NeuralNodeType.OUTPUT) {
              let input: org.mwg.struct.Relation = <org.mwg.struct.Relation>state.getFromKey(NeuralNode.INPUTS);
              let weights: Float64Array = new Float64Array(input.size() + 1);
              for (let i: number = 0; i < weights.length; i++) {
                weights[i] = NeuralNode.random.nextDouble() * maxValue;
              }
              state.setFromKey(NeuralNode.WEIGHTS, org.mwg.Type.DOUBLE_ARRAY, weights);
            }
          }
          private forwardConnect(inputNode: org.mwg.ml.neuralnet.NeuralNode): void {
            let outputs: org.mwg.struct.Relation = <org.mwg.struct.Relation>this.getOrCreate(NeuralNode.OUTPUTS, org.mwg.Type.RELATION);
            outputs.add(inputNode.id());
            let pos: number = outputs.size() - 1;
            let map: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.getOrCreate(NeuralNode.OUTPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
            map.put(inputNode.id(), pos);
            let inputs: org.mwg.struct.Relation = <org.mwg.struct.Relation>inputNode.getOrCreate(NeuralNode.INPUTS, org.mwg.Type.RELATION);
            inputs.add(this.id());
            let posint: number = inputs.size() - 1;
            let mapin: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>inputNode.getOrCreate(NeuralNode.INPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
            mapin.put(this.id(), posint);
          }
          private createNewNode(neuralNodeType: number): org.mwg.ml.neuralnet.NeuralNode {
            let temp: org.mwg.ml.neuralnet.NeuralNode = <org.mwg.ml.neuralnet.NeuralNode>this.graph().newTypedNode(this.world(), this.time(), NeuralNode.NAME);
            temp.set(NeuralNode.NODE_TYPE, org.mwg.Type.INT, neuralNodeType);
            return temp;
          }
          private static integrationFct(values: Float64Array, weights: Float64Array): number {
            let value: number = 0;
            for (let i: number = 0; i < values.length; i++) {
              value += weights[i] * values[i];
            }
            value += weights[values.length];
            return value;
          }
          private static activationFunction(x: number, type: number): number {
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN) {
              return 1 / (1 + Math.exp(-x));
            } else {
              return x;
            }
          }
          private static derivateActivationFunction(fctVal: number, x: number, type: number): number {
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN) {
              return fctVal * (1 - fctVal);
            } else {
              return 1;
            }
          }
          private static calculateErr(calculated: number, target: number): number {
            return (target - calculated) * (target - calculated) / 2;
          }
          private static calculateDerivativeErr(calculated: number, target: number): number {
            return -(target - calculated);
          }
          private generateMsgId(): number {
            NeuralNode.msgIdCounter++;
            return NeuralNode.msgIdCounter;
          }
          public predict(inputs: Float64Array, callback: org.mwg.Callback<Float64Array>): void {
            let prediction: Float64Array = null;
            callback(prediction);
          }
          public learn(inputs: Float64Array, outputs: Float64Array, callback: org.mwg.Callback<Float64Array>): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            this.predict(inputs, (predictions : Float64Array) => {
{
                let errors: Float64Array = new Float64Array(outputs.length);
                let derivatives: Float64Array = new Float64Array(outputs.length);
                for (let i: number = 0; i < errors.length; i++) {
                  errors[i] = org.mwg.ml.neuralnet.NeuralNode.calculateErr(predictions[i], outputs[i]);
                  derivatives[i] = org.mwg.ml.neuralnet.NeuralNode.calculateDerivativeErr(predictions[i], outputs[i]);
                }
                let learningRate: number = state.getFromKeyWithDefault(org.mwg.ml.neuralnet.NeuralNode.LEARNINGRATE, org.mwg.ml.neuralnet.NeuralNode.LEARNINGRATE_DEF);
                if (callback != null) {
                  callback(errors);
                }
              }            });
          }
        }
        export class NeuralNodeEmpty extends org.mwg.base.BaseNode {
          public static NAME: string = "NeuralNodeEmpty";
          public static INPUTS: string = "inputs";
          public static INPUTS_MAP: string = "inputs_map";
          public static OUTPUTS: string = "outputs";
          public static OUTPUTS_MAP: string = "outputs_map";
          private static WEIGHTS: string = "weights";
          private static NODE_TYPE: string = "node_type";
          public static LEARNINGRATE: string = "learning";
          public static LEARNINGRATE_DEF: number = 0.01;
          private static random: java.util.Random = new java.util.Random();
          private static predictTask: org.mwg.task.Task = NeuralNodeEmpty.createPredictTask();
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          public configure(inputs: number, outputs: number, hiddenlayers: number, nodesPerLayer: number): org.mwg.ml.neuralnet.NeuralNodeEmpty {
            let internalNodes: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty>();
            internalNodes.add(this);
            let previousLayer: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty>();
            let state: org.mwg.plugin.NodeState = this.phasedState();
            state.setFromKey(NeuralNodeEmpty.NODE_TYPE, org.mwg.Type.INT, org.mwg.ml.neuralnet.NeuralNodeType.ROOT);
            for (let i: number = 0; i < inputs; i++) {
              let inputNode: org.mwg.ml.neuralnet.NeuralNodeEmpty = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.INPUT);
              this.forwardConnect(inputNode);
              internalNodes.add(inputNode);
              previousLayer.add(inputNode);
            }
            let nextLayer: java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty> = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty>();
            for (let i: number = 0; i < hiddenlayers; i++) {
              for (let j: number = 0; j < nodesPerLayer; j++) {
                let hidden: org.mwg.ml.neuralnet.NeuralNodeEmpty = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN);
                nextLayer.add(hidden);
                internalNodes.add(hidden);
                for (let k: number = 0; k < previousLayer.size(); k++) {
                  previousLayer.get(k).forwardConnect(hidden);
                }
              }
              previousLayer = nextLayer;
              nextLayer = new java.util.ArrayList<org.mwg.ml.neuralnet.NeuralNodeEmpty>();
            }
            for (let i: number = 0; i < outputs; i++) {
              let output: org.mwg.ml.neuralnet.NeuralNodeEmpty = this.createNewNode(org.mwg.ml.neuralnet.NeuralNodeType.OUTPUT);
              for (let k: number = 0; k < previousLayer.size(); k++) {
                previousLayer.get(k).forwardConnect(output);
              }
              output.forwardConnect(this);
              internalNodes.add(output);
            }
            for (let i: number = 0; i < internalNodes.size(); i++) {
              if (internalNodes.get(i).id() != this.id()) {
                internalNodes.get(i).initWeightsRadomly(0.1);
                internalNodes.get(i).free();
              }
            }
            return this;
          }
          private initWeightsRadomly(maxValue: number): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            let type: number = state.getFromKeyWithDefault(NeuralNodeEmpty.NODE_TYPE, org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN);
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN || type == org.mwg.ml.neuralnet.NeuralNodeType.OUTPUT) {
              let input: org.mwg.struct.Relation = <org.mwg.struct.Relation>state.getFromKey(NeuralNodeEmpty.INPUTS);
              let weights: Float64Array = new Float64Array(input.size() + 1);
              for (let i: number = 0; i < weights.length; i++) {
                weights[i] = NeuralNodeEmpty.random.nextDouble() * maxValue;
              }
              state.setFromKey(NeuralNodeEmpty.WEIGHTS, org.mwg.Type.DOUBLE_ARRAY, weights);
            }
          }
          private forwardConnect(inputNode: org.mwg.ml.neuralnet.NeuralNodeEmpty): void {
            let outputs: org.mwg.struct.Relation = <org.mwg.struct.Relation>this.getOrCreate(NeuralNodeEmpty.OUTPUTS, org.mwg.Type.RELATION);
            outputs.add(inputNode.id());
            let pos: number = outputs.size() - 1;
            let map: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.getOrCreate(NeuralNodeEmpty.OUTPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
            map.put(inputNode.id(), pos);
            let inputs: org.mwg.struct.Relation = <org.mwg.struct.Relation>inputNode.getOrCreate(NeuralNodeEmpty.INPUTS, org.mwg.Type.RELATION);
            inputs.add(this.id());
            let posint: number = inputs.size() - 1;
            let mapin: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>inputNode.getOrCreate(NeuralNodeEmpty.INPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
            mapin.put(this.id(), posint);
          }
          private createNewNode(neuralNodeType: number): org.mwg.ml.neuralnet.NeuralNodeEmpty {
            let temp: org.mwg.ml.neuralnet.NeuralNodeEmpty = <org.mwg.ml.neuralnet.NeuralNodeEmpty>this.graph().newTypedNode(this.world(), this.time(), NeuralNodeEmpty.NAME);
            temp.set(NeuralNodeEmpty.NODE_TYPE, org.mwg.Type.INT, neuralNodeType);
            return temp;
          }
          private static integrationFct(values: Float64Array, weights: Float64Array): number {
            let value: number = 0;
            for (let i: number = 0; i < values.length; i++) {
              value += weights[i] * values[i];
            }
            value += weights[values.length];
            return value;
          }
          private static activationFunction(x: number, type: number): number {
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN) {
              return 1 / (1 + Math.exp(-x));
            } else {
              return x;
            }
          }
          private static derivateActivationFunction(fctVal: number, x: number, type: number): number {
            if (type == org.mwg.ml.neuralnet.NeuralNodeType.HIDDEN) {
              return fctVal * (1 - fctVal);
            } else {
              return 1;
            }
          }
          private static calculateErr(calculated: number, target: number): number {
            return (target - calculated) * (target - calculated) / 2;
          }
          private static calculateDerivativeErr(calculated: number, target: number): number {
            return -(target - calculated);
          }
          private static createPredictTask(): org.mwg.task.Task {
            let t: org.mwg.task.Task = org.mwg.task.Tasks.newTask();
            t.then(org.mwg.internal.task.CoreActions.setAsVar("parent")).then(org.mwg.internal.task.CoreActions.traverse(NeuralNodeEmpty.OUTPUTS)).forEach(org.mwg.task.Tasks.newTask().thenDo((ctx : org.mwg.task.TaskContext) => {
{
                ctx.continueTask();
              }            }).ifThen((ctx : org.mwg.task.TaskContext) => {
{
                return false;
              }            }, org.mwg.task.Tasks.newTask().thenDo((ctx : org.mwg.task.TaskContext) => {
{
                ctx.continueTask();
              }            }).ifThenElse((ctx : org.mwg.task.TaskContext) => {
{
                return false;
              }            }, t, org.mwg.task.Tasks.newTask().thenDo((ctx : org.mwg.task.TaskContext) => {
{
                ctx.continueTask();
              }            }))));
            return t;
          }
          public predict(inputs: Float64Array, callback: org.mwg.Callback<Float64Array>): void {
            let prediction: Float64Array = null;
            callback(prediction);
          }
          public learn(inputs: Float64Array, outputs: Float64Array, callback: org.mwg.Callback<Float64Array>): void {
            let state: org.mwg.plugin.NodeState = this.phasedState();
            this.predict(inputs, (predictions : Float64Array) => {
{
                let errors: Float64Array = new Float64Array(outputs.length);
                let derivatives: Float64Array = new Float64Array(outputs.length);
                for (let i: number = 0; i < errors.length; i++) {
                  errors[i] = org.mwg.ml.neuralnet.NeuralNodeEmpty.calculateErr(predictions[i], outputs[i]);
                  derivatives[i] = org.mwg.ml.neuralnet.NeuralNodeEmpty.calculateDerivativeErr(predictions[i], outputs[i]);
                }
                let learningRate: number = state.getFromKeyWithDefault(org.mwg.ml.neuralnet.NeuralNodeEmpty.LEARNINGRATE, org.mwg.ml.neuralnet.NeuralNodeEmpty.LEARNINGRATE_DEF);
                if (callback != null) {
                  callback(errors);
                }
              }            });
          }
        }
        export class NeuralNodeType {
          public static HIDDEN: number = 0;
          public static INPUT: number = 1;
          public static OUTPUT: number = 2;
          public static ROOT: number = 3;
        }
        export module bio {
          export class BioInputNeuralNode extends org.mwg.base.BaseNode {
            public static NAME: string = "BioInputNeuralNode";
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public learn2(value: number, spikeLimit: number, threshold: number, callback: org.mwg.Callback<any>): void {
              let job: org.mwg.task.Task[] = new Array<org.mwg.task.Task>(1);
              job[0] = org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.readVar("target")).forEach(org.mwg.task.Tasks.newTask().then(org.mwg.internal.task.CoreActions.lookup("{{result}}")).thenDo((context)=>{
                let neural: org.mwg.ml.neuralnet.bio.BioNeuralNode = <org.mwg.ml.neuralnet.bio.BioNeuralNode>context.result().get(0);
                let outputs: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>neural.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_OUTPUTS);
                outputs.each((key, value1)=>(context.addToVariable("target", key)));
                let signal: number = <number>context.variable("signal").get(0);
                let sender: number = <number>context.variable("sender").get(0);
                let nextSignal: number = neural.learn(sender, signal, spikeLimit, threshold);
                context.setVariable("signal", nextSignal);
                context.setVariable("sender", neural.id());
              }).ifThen((context)=>((<number>context.variable("signal").get(0)) > 0), org.mwg.task.Tasks.newTask().thenDo((context)=>{
                let ctx: org.mwg.task.TaskContext = job[0].prepare(this.graph(), null, (result)=>{
                  context.continueTask();
                });
                ctx.setVariable("signal", context.variable("signal"));
                ctx.setVariable("sender", context.variable("sender"));
                ctx.setVariable("target", context.variable("target"));
                job[0].executeUsing(ctx);
              })));
              let ctx: org.mwg.task.TaskContext = job[0].prepare(this.graph(), null, callback);
              ctx.setVariable("signal", value);
              ctx.setVariable("sender", this.id());
              let outputs: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_OUTPUTS);
              outputs.each((key, value1)=>{
                ctx.addToVariable("target", key);
              });
              job[0].executeUsing(ctx);
            }
            public learn(value: number, spikeLimit: number, threshold: number, callback: org.mwg.Callback<any>): void {
              let loop: java.util.Map<number, java.util.List<org.mwg.utility.Tuple<number, number>>> = new java.util.HashMap<number, java.util.List<org.mwg.utility.Tuple<number, number>>>();
              let firstHiddenLayer: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_OUTPUTS);
              firstHiddenLayer.each((key, value1)=>{
                let previous: java.util.List<org.mwg.utility.Tuple<number, number>> = loop.get(key);
                if (previous == null) {
                  previous = new java.util.ArrayList<org.mwg.utility.Tuple<number, number>>();
                  loop.put(key, previous);
                }
                previous.add(new org.mwg.utility.Tuple<number, number>(this.id(), value));
              });
              let job: org.mwg.plugin.Job[] = new Array<org.mwg.plugin.Job>(1);
              job[0] = () => {
{
                  let keys: Float64Array = new Float64Array(loop.size());
                  let index: number = 0;
                  loop.keySet().forEach((v: number) => {
{
                      keys[index] = v;
                      index++;
                    }                  });
                  this.graph().lookupAll(this.world(), this.time(), keys, (nextLayers)=>{
                    for (let i: number = 0; i < nextLayers.length; i++) {
                      if (nextLayers[i] instanceof org.mwg.ml.neuralnet.bio.BioOutputNeuralNode) {
                        let neural: org.mwg.ml.neuralnet.bio.BioOutputNeuralNode = <org.mwg.ml.neuralnet.bio.BioOutputNeuralNode>nextLayers[i];
                        let currentId: number = neural.id();
                        let signals: java.util.List<org.mwg.utility.Tuple<number, number>> = loop.get(currentId);
                        loop.remove(currentId);
                        for (let j: number = 0; j < signals.size(); j++) {
                          let nextSignal: number = neural.learn(signals.get(j).left(), signals.get(j).right(), spikeLimit, threshold);
                        }
                      } else {
                        let neural: org.mwg.ml.neuralnet.bio.BioNeuralNode = <org.mwg.ml.neuralnet.bio.BioNeuralNode>nextLayers[i];
                        let currentId: number = neural.id();
                        let signals: java.util.List<org.mwg.utility.Tuple<number, number>> = loop.get(currentId);
                        loop.remove(currentId);
                        let outputs: org.mwg.struct.LongLongArrayMap = <org.mwg.struct.LongLongArrayMap>neural.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_OUTPUTS);
                        for (let j: number = 0; j < signals.size(); j++) {
                          let nextSignal: number = neural.learn(signals.get(j).left(), signals.get(j).right(), spikeLimit, threshold);
                          if (nextSignal != 0) {
                            outputs.each((key, value1)=>{
                              let previous: java.util.List<org.mwg.utility.Tuple<number, number>> = loop.get(key);
                              if (previous == null) {
                                previous = new java.util.ArrayList<org.mwg.utility.Tuple<number, number>>();
                                loop.put(key, previous);
                              }
                              previous.add(new org.mwg.utility.Tuple<number, number>(currentId, nextSignal));
                            });
                          }
                        }
                      }
                      nextLayers[i].free();
                    }
                    if (loop.size() > 0) {
                      this.graph().scheduler().dispatch(org.mwg.plugin.SchedulerAffinity.SAME_THREAD, job[0]);
                    }
                  });
                }              };
              this.graph().scheduler().dispatch(org.mwg.plugin.SchedulerAffinity.SAME_THREAD, job[0]);
            }
          }
          export class BioNeuralNetwork extends org.mwg.base.BaseNode {
            public static NAME: string = "BioNeuralNetworkNode";
            public static RELATION_INPUTS: string = "inputs";
            private static RELATION_INPUTS_MAP: string = "inputs_map";
            public static RELATION_OUTPUTS: string = "outputs";
            private static RELATION_OUTPUTS_MAP: string = "outputs_map";
            public static BUFFER_SPIKE_SUM: string = "spike_sum";
            public static BUFFER_SPIKE_NB: string = "spike_nb";
            public static WEIGHTS: string = "weights";
            public static BIAS: string = "bias";
            public static SPIKE_LIMIT: string = "spikeLimit";
            public static THRESHOLD: string = "threshold";
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public configure(inputs: number, outputs: number, hiddenlayers: number, nodesPerLayer: number, spikeLimit: number, threshold: number): org.mwg.ml.neuralnet.bio.BioNeuralNetwork {
              this.set(BioNeuralNetwork.SPIKE_LIMIT, org.mwg.Type.INT, spikeLimit);
              this.set(BioNeuralNetwork.THRESHOLD, org.mwg.Type.DOUBLE, threshold);
              let inputNodes: org.mwg.Node[] = new Array<org.mwg.ml.neuralnet.bio.BioInputNeuralNode>(inputs);
              for (let i: number = 0; i < inputs; i++) {
                let input: org.mwg.ml.neuralnet.bio.BioInputNeuralNode = <org.mwg.ml.neuralnet.bio.BioInputNeuralNode>this.graph().newTypedNode(this.world(), this.time(), org.mwg.ml.neuralnet.bio.BioInputNeuralNode.NAME);
                inputNodes[i] = input;
                this.addToRelation(BioNeuralNetwork.RELATION_INPUTS, input);
              }
              let previousLayer: org.mwg.ml.neuralnet.bio.BioNeuralNode[] = new Array<org.mwg.ml.neuralnet.bio.BioNeuralNode>(nodesPerLayer);
              let rand: java.util.Random = new java.util.Random();
              rand.setSeed(1234);
              for (let i: number = 0; i < hiddenlayers; i++) {
                if (i == 0) {
                  for (let j: number = 0; j < nodesPerLayer; j++) {
                    let neuralNode: org.mwg.ml.neuralnet.bio.BioNeuralNode = <org.mwg.ml.neuralnet.bio.BioNeuralNode>this.graph().newTypedNode(this.world(), this.time(), org.mwg.ml.neuralnet.bio.BioNeuralNode.NAME);
                    let spikeSum: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_SUM, org.mwg.Type.DMATRIX);
                    spikeSum.init(1, inputs);
                    spikeSum.fill(0);
                    let spikeNb: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_NB, org.mwg.Type.DMATRIX);
                    spikeNb.init(1, inputs);
                    spikeNb.fill(0);
                    let weights: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.WEIGHTS, org.mwg.Type.DMATRIX);
                    weights.init(inputs, 1);
                    weights.fillWithRandom(rand, -1, 1);
                    previousLayer[j] = neuralNode;
                    for (let k: number = 0; k < inputs; k++) {
                      (<org.mwg.struct.LongLongMap>inputNodes[k].getOrCreate(BioNeuralNetwork.RELATION_OUTPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(neuralNode.id(), j);
                      (<org.mwg.struct.LongLongMap>neuralNode.getOrCreate(BioNeuralNetwork.RELATION_INPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(inputNodes[k].id(), k);
                    }
                  }
                  this.graph().freeNodes(inputNodes);
                  inputNodes = null;
                } else {
                  let tempLayer: org.mwg.ml.neuralnet.bio.BioNeuralNode[] = new Array<org.mwg.ml.neuralnet.bio.BioNeuralNode>(nodesPerLayer);
                  for (let j: number = 0; j < nodesPerLayer; j++) {
                    let neuralNode: org.mwg.ml.neuralnet.bio.BioNeuralNode = <org.mwg.ml.neuralnet.bio.BioNeuralNode>this.graph().newTypedNode(this.world(), this.time(), org.mwg.ml.neuralnet.bio.BioNeuralNode.NAME);
                    let spikeSum: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_SUM, org.mwg.Type.DMATRIX);
                    spikeSum.init(1, nodesPerLayer);
                    spikeSum.fill(0);
                    let spikeNb: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_NB, org.mwg.Type.DMATRIX);
                    spikeNb.init(1, nodesPerLayer);
                    spikeNb.fill(0);
                    let weights: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>neuralNode.getOrCreate(BioNeuralNetwork.WEIGHTS, org.mwg.Type.DMATRIX);
                    weights.init(nodesPerLayer, 1);
                    weights.fillWithRandom(rand, -1, 1);
                    tempLayer[j] = neuralNode;
                    for (let k: number = 0; k < nodesPerLayer; k++) {
                      (<org.mwg.struct.LongLongMap>previousLayer[k].getOrCreate(BioNeuralNetwork.RELATION_OUTPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(neuralNode.id(), j);
                      (<org.mwg.struct.LongLongMap>neuralNode.getOrCreate(BioNeuralNetwork.RELATION_INPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(previousLayer[k].id(), k);
                    }
                  }
                  this.graph().freeNodes(previousLayer);
                  previousLayer = tempLayer;
                }
              }
              for (let i: number = 0; i < outputs; i++) {
                let output: org.mwg.ml.neuralnet.bio.BioOutputNeuralNode = <org.mwg.ml.neuralnet.bio.BioOutputNeuralNode>this.graph().newTypedNode(this.world(), this.time(), org.mwg.ml.neuralnet.bio.BioOutputNeuralNode.NAME);
                let spikeSum: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>output.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_SUM, org.mwg.Type.DMATRIX);
                spikeSum.init(1, nodesPerLayer);
                spikeSum.fill(0);
                let spikeNb: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>output.getOrCreate(BioNeuralNetwork.BUFFER_SPIKE_NB, org.mwg.Type.DMATRIX);
                spikeNb.init(1, nodesPerLayer);
                spikeNb.fill(0);
                let weights: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>output.getOrCreate(BioNeuralNetwork.WEIGHTS, org.mwg.Type.DMATRIX);
                weights.init(nodesPerLayer, 1);
                weights.fillWithRandom(rand, -1, 1);
                for (let k: number = 0; k < nodesPerLayer; k++) {
                  (<org.mwg.struct.LongLongMap>previousLayer[k].getOrCreate(BioNeuralNetwork.RELATION_OUTPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(output.id(), i);
                  (<org.mwg.struct.LongLongMap>output.getOrCreate(BioNeuralNetwork.RELATION_INPUTS, org.mwg.Type.LONG_TO_LONG_MAP)).put(previousLayer[k].id(), k);
                }
                this.addToRelation(BioNeuralNetwork.RELATION_OUTPUTS, output);
              }
              this.graph().freeNodes(previousLayer);
              return this;
            }
            public learn(inputCode: number, isInput: boolean, value: number, callbacks: org.mwg.Callback<any>): void {
              let spikeLimit: number = <number>this.get(BioNeuralNetwork.SPIKE_LIMIT);
              let threshold: number = <number>this.get(BioNeuralNetwork.THRESHOLD);
              let mapping: org.mwg.struct.LongLongMap;
              if (isInput) {
                mapping = <org.mwg.struct.LongLongMap>this.getOrCreate(BioNeuralNetwork.RELATION_INPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
              } else {
                mapping = <org.mwg.struct.LongLongMap>this.getOrCreate(BioNeuralNetwork.RELATION_OUTPUTS_MAP, org.mwg.Type.LONG_TO_LONG_MAP);
              }
              let previousNode: number = mapping.get(inputCode);
              if (previousNode == org.mwg.Constants.NULL_LONG) {
                let relationship: org.mwg.struct.Relation;
                if (isInput) {
                  relationship = <org.mwg.struct.Relation>this.get(BioNeuralNetwork.RELATION_INPUTS);
                } else {
                  relationship = <org.mwg.struct.Relation>this.get(BioNeuralNetwork.RELATION_OUTPUTS);
                }
                if (relationship == null) {
                  throw new Error("Bad API usage, please call configure method first!");
                }
                if (relationship.size() == mapping.size()) {
                  throw new Error("All input/output has been consumed, please reset the " + BioNeuralNetwork.NAME);
                }
                for (let i: number = 0; i < relationship.size(); i++) {
                  let relationVal: number = relationship.get(i);
                  let isPresent: boolean[] = [false];
                  mapping.each((key : number, value : number) => {
{
                      if (value == relationVal) {
                        isPresent[0] = true;
                      }
                    }                  });
                  if (!isPresent[0]) {
                    previousNode = relationVal;
break;
                  }
                }
              }
              if (previousNode == org.mwg.Constants.NULL_LONG) {
                throw new Error("Internal error exception!");
              }
              this.graph().lookup(this.world(), this.time(), previousNode, (result : org.mwg.Node) => {
{
                  if (result == null) {
                    throw new Error("Internal error exception!");
                  } else {
                    if (isInput) {
                      (<org.mwg.ml.neuralnet.bio.BioInputNeuralNode>result).learn(value, spikeLimit, threshold, callbacks);
                    } else {
                    }
                  }
                }              });
            }
          }
          export class BioNeuralNode extends org.mwg.base.BaseNode {
            public static NAME: string = "BioNeuralNode";
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public init(): void {
              this.set(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BIAS, org.mwg.Type.DOUBLE, new java.util.Random().nextDouble() * 2 - 1);
            }
            public learn(sender: number, value: number, spikeLimit: number, threshold: number): number {
              let spikeSum: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BUFFER_SPIKE_SUM);
              let spikeNb: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BUFFER_SPIKE_NB);
              let weights: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.WEIGHTS);
              let reverse: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_INPUTS);
              let senderIndex: number = <number>reverse.get(sender);
              spikeSum.add(0, senderIndex, value);
              spikeNb.add(0, senderIndex, 1);
              let signal: number = org.mwg.ml.common.matrix.MatrixOps.multiply(spikeSum, weights).get(0, 0);
              let bias: number = <number>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BIAS);
              let sigmoid: number = 1 / (1 + Math.exp(-(signal + bias)));
              let spikeSumTot: number = 0;
              for (let i: number = 0; i < spikeNb.columns(); i++) {
                let loopSpikeNb: number = <number>spikeNb.get(0, i);
                if (loopSpikeNb >= spikeLimit) {
                  spikeSum.fill(0);
                  spikeNb.fill(0);
                  return sigmoid;
                }
                spikeSumTot = spikeSumTot + loopSpikeNb;
              }
              if ((spikeSumTot >= spikeSum.columns()) && (sigmoid > threshold)) {
                spikeSum.fill(0);
                spikeNb.fill(0);
                return sigmoid;
              }
              return 0;
            }
          }
          export class BioOutputNeuralNode extends org.mwg.base.BaseNode {
            public static NAME: string = "BioOutputNeuralNode";
            constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
              super(p_world, p_time, p_id, p_graph);
            }
            public init(): void {
              this.set(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BIAS, org.mwg.Type.DOUBLE, new java.util.Random().nextDouble() * 2 - 1);
            }
            public learn(sender: number, value: number, spikeLimit: number, threshold: number): number {
              let spikeSum: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BUFFER_SPIKE_SUM);
              let spikeNb: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BUFFER_SPIKE_NB);
              let weights: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.WEIGHTS);
              let reverse: org.mwg.struct.LongLongMap = <org.mwg.struct.LongLongMap>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.RELATION_INPUTS);
              let senderIndex: number = <number>reverse.get(sender);
              spikeSum.add(0, senderIndex, value);
              spikeNb.add(0, senderIndex, 1);
              let signal: number = org.mwg.ml.common.matrix.MatrixOps.multiply(spikeSum, weights).get(0, 0);
              let bias: number = <number>this.get(org.mwg.ml.neuralnet.bio.BioNeuralNetwork.BIAS);
              let integrated_signal: number = signal + bias;
              let maxSpikeNb: number = 0;
              for (let i: number = 0; i < spikeNb.columns(); i++) {
                let loopSpikeNb: number = <number>spikeNb.get(0, i);
                if (loopSpikeNb >= spikeLimit) {
                  spikeSum.fill(0);
                  spikeNb.fill(0);
                  return integrated_signal;
                }
                if (loopSpikeNb > maxSpikeNb) {
                  maxSpikeNb = loopSpikeNb;
                }
              }
              return 0;
            }
          }
        }
        export module functions {
          export class LinearFunction implements org.mwg.ml.neuralnet.NeuralFunction {
            public activate(x: number): number {
              return x;
            }
            public derivate(x: number, fct: number): number {
              return 1;
            }
          }
          export class SigmoidFunction implements org.mwg.ml.neuralnet.NeuralFunction {
            public activate(x: number): number {
              return 1 / (1 + Math.exp(-x));
            }
            public derivate(x: number, fct: number): number {
              return fct * (1 - fct);
            }
          }
        }
      }
      export module nn {
        export class ExMatrix implements org.mwg.struct.DMatrix {
          private static DW_KEY: string = "-dw";
          private static STEPCACHE_KEY: string = "-sc";
          private w: org.mwg.struct.DMatrix;
          private dw: org.mwg.struct.DMatrix;
          private stepCache: org.mwg.struct.DMatrix;
          constructor(node: org.mwg.struct.ENode, attribute: string) {
            if (node != null) {
              this.w = <org.mwg.struct.DMatrix>node.getOrCreate(attribute, org.mwg.Type.DMATRIX);
              this.dw = <org.mwg.struct.DMatrix>node.getOrCreate(attribute + ExMatrix.DW_KEY, org.mwg.Type.DMATRIX);
              this.stepCache = <org.mwg.struct.DMatrix>node.getOrCreate(attribute + ExMatrix.STEPCACHE_KEY, org.mwg.Type.DMATRIX);
            }
          }
          public init(rows: number, columns: number): org.mwg.struct.DMatrix {
            if (this.w == null) {
              this.w = org.mwg.ml.common.matrix.VolatileDMatrix.empty(rows, columns);
              this.dw = org.mwg.ml.common.matrix.VolatileDMatrix.empty(rows, columns);
              this.stepCache = org.mwg.ml.common.matrix.VolatileDMatrix.empty(rows, columns);
            } else {
              this.w.init(rows, columns);
              this.dw.init(rows, columns);
              this.stepCache.init(rows, columns);
            }
            return this;
          }
          public fill(value: number): org.mwg.struct.DMatrix {
            this.w.fill(value);
            return this;
          }
          public fillWith(values: Float64Array): org.mwg.struct.DMatrix {
            this.w.fillWith(values);
            return this;
          }
          public fillWithRandom(random: java.util.Random, min: number, max: number): org.mwg.struct.DMatrix {
            this.w.fillWithRandom(random, min, max);
            return this;
          }
          public fillWithRandomStd(random: java.util.Random, std: number): org.mwg.struct.DMatrix {
            this.w.fillWithRandomStd(random, std);
            return this;
          }
          public rows(): number {
            return this.w.rows();
          }
          public columns(): number {
            return this.w.columns();
          }
          public column(i: number): Float64Array {
            return this.w.column(i);
          }
          public get(rowIndex: number, columnIndex: number): number {
            return this.w.get(rowIndex, columnIndex);
          }
          public set(rowIndex: number, columnIndex: number, value: number): org.mwg.struct.DMatrix {
            this.w.set(rowIndex, columnIndex, value);
            return this;
          }
          public add(rowIndex: number, columnIndex: number, value: number): org.mwg.struct.DMatrix {
            this.w.add(rowIndex, columnIndex, value);
            return this;
          }
          public appendColumn(newColumn: Float64Array): org.mwg.struct.DMatrix {
            this.w.appendColumn(newColumn);
            this.dw.appendColumn(new Float64Array(newColumn.length));
            this.stepCache.appendColumn(new Float64Array(newColumn.length));
            return null;
          }
          public data(): Float64Array {
            return this.w.data();
          }
          public leadingDimension(): number {
            return this.w.leadingDimension();
          }
          public unsafeGet(index: number): number {
            return this.w.unsafeGet(index);
          }
          public unsafeSet(index: number, value: number): org.mwg.struct.DMatrix {
            this.w.unsafeSet(index, value);
            return this;
          }
          public getW(): org.mwg.struct.DMatrix {
            return this.w;
          }
          public getDw(): org.mwg.struct.DMatrix {
            return this.dw;
          }
          public getStepCache(): org.mwg.struct.DMatrix {
            return this.stepCache;
          }
        }
      }
    }
  }
}
